/*=============================================================================
	OpenGL3.cpp: Unreal OpenGL 3.3 support.
	Portions copyright 1999 Epic Games, Inc. All Rights Reserved.

	OpenGL renderer by Daniel Vogel <vogel@lokigames.com>
	Loki Software, Inc.

	Other URenderDevice subclasses include:
	* USoftwareRenderDevice: Software renderer.
	* UGlideRenderDevice: 3dfx Glide renderer.
	* UDirect3DRenderDevice: Direct3D renderer.
	* UOpenGLRenderDevice: OpenGL renderer.

	Revision history:
	* A few parts remaining from OpenGL renderer by Daniel Vogel
	* Various modifications and additions by Chris Dohnal
	* Various modifications and additions by Smirftsch / Oldunreal

=============================================================================*/

#include "OpenGL3Drv.h"
#include "OpenGL3.h"

#ifdef _WIN32
#include <mmsystem.h>
#endif


/*-----------------------------------------------------------------------------
	Globals.
-----------------------------------------------------------------------------*/

static const TCHAR *g_pSection = TEXT("OpenGL3Drv.OpenGLRenderDevice");


/*-----------------------------------------------------------------------------
	Vertex programs.
-----------------------------------------------------------------------------*/

//Vertex shader attribute locations
// 0  - Vertex position
// 1  - Primary color
// 2  - Secondary color
// 3  - Normal
// 4  - TexCoord 0
// 5  - TexCoord 1
// 6  - TexCoord 2
// 7  - TexCoord 3

// csVertInfo
//   fxInfo -> XAxis, csUDot
//   fyInfo -> YAxis, csVDot
//   texInfo -> Up to 4 texture [UPan, VPan, UMult, VMult]

// fragInfo
//   blendInfo -> alphaRef, baseTexLodBias, rsvd, Lightmap blend scale factor
//   fogColor -> Fog color
//   fogInfo -> Linear fog [1.0 / (end - start)], Linear fog [end / (end - start)], rsvd, rsvd

static const char *g_vsDefaultRenderingState =
	"#version 330\n"

	"layout(std140) uniform trInfo {\n"
		"mat4 trMat;\n"
	"};\n"

	"in vec4 viPos;\n"
	"in vec4 viPriCol;\n"
	"in vec2 viTC0;\n"

	"out vec4 fiPriCol;\n"
	"out vec2 fiTC[1];\n"

	"void main() {\n"
		"fiPriCol = viPriCol;\n"
		"fiTC[0] = viTC0;\n"
		"gl_Position = trMat * viPos;\n"
	"}\n"
;

static const char *g_vsDefaultRenderingStateWithFog =
	"#version 330\n"

	"layout(std140) uniform trInfo {\n"
		"mat4 trMat;\n"
	"};\n"

	"in vec4 viPos;\n"
	"in vec4 viPriCol;\n"
	"in vec4 viSecCol;\n"
	"in vec2 viTC0;\n"

	"out vec4 fiPriCol;\n"
	"out vec4 oSecCol;\n"
	"out vec2 fiTC[1];\n"

	"void main() {\n"
		"fiPriCol = viPriCol;\n"
		"oSecCol = viSecCol;\n"
		"fiTC[0] = viTC0;\n"
		"gl_Position = trMat * viPos;\n"
	"}\n"
;

static const char *g_vsDefaultRenderingStateWithLinearFog =
	"#version 330\n"

	"layout(std140) uniform trInfo {\n"
		"mat4 trMat;\n"
	"};\n"

	"in vec4 viPos;\n"
	"in vec4 viPriCol;\n"
	"in vec2 viTC0;\n"

	"out vec4 fiPriCol;\n"
	"out vec2 fiTC[1];\n"
	"out float fiPosZ;\n"

	"void main() {\n"
		"fiPosZ = viPos.z;\n"
		"fiPriCol = viPriCol;\n"
		"fiTC[0] = viTC0;\n"
		"gl_Position = trMat * viPos;\n"
	"}\n"
;

static const char *g_vsComplexSurfaceSingleTexture =
	"#version 330\n"

	"layout(std140) uniform trInfo {\n"
		"mat4 trMat;\n"
	"};\n"

	"layout(std140) uniform csVertInfo {\n"
		"vec4 fxInfo;\n"
		"vec4 fyInfo;\n"
		"vec4 texInfo[4];\n"
	"};\n"

	"in vec4 viPos;\n"
	"in vec4 viPriCol;\n"

	"out vec4 fiPriCol;\n"
	"out vec2 fiTC[1];\n"

	"vec2 t0;\n"

	"void main() {\n"
		"t0.x = dot(viPos.xyz, fxInfo.xyz) - fxInfo.w;\n"
		"t0.y = dot(viPos.xyz, fyInfo.xyz) - fyInfo.w;\n"

		"fiTC[0].xy = (t0.xy - texInfo[0].xy) * texInfo[0].zw;\n"

		"fiPriCol = viPriCol;\n"
		"gl_Position = trMat * viPos;\n"
	"}\n"
;

static const char *g_vsComplexSurfaceDualTexture =
	"#version 330\n"

	"layout(std140) uniform trInfo {\n"
		"mat4 trMat;\n"
	"};\n"

	"layout(std140) uniform csVertInfo {\n"
		"vec4 fxInfo;\n"
		"vec4 fyInfo;\n"
		"vec4 texInfo[4];\n"
	"};\n"

	"in vec4 viPos;\n"
	"in vec4 viPriCol;\n"

	"out vec4 fiPriCol;\n"
	"out vec2 fiTC[2];\n"

	"vec2 t0;\n"

	"void main() {\n"
		"t0.x = dot(viPos.xyz, fxInfo.xyz) - fxInfo.w;\n"
		"t0.y = dot(viPos.xyz, fyInfo.xyz) - fyInfo.w;\n"

		"fiTC[0].xy = (t0.xy - texInfo[0].xy) * texInfo[0].zw;\n"
		"fiTC[1].xy = (t0.xy - texInfo[1].xy) * texInfo[1].zw;\n"

		"fiPriCol = viPriCol;\n"
		"gl_Position = trMat * viPos;\n"
	"}\n"
;

static const char *g_vsComplexSurfaceTripleTexture =
	"#version 330\n"

	"layout(std140) uniform trInfo {\n"
		"mat4 trMat;\n"
	"};\n"

	"layout(std140) uniform csVertInfo {\n"
		"vec4 fxInfo;\n"
		"vec4 fyInfo;\n"
		"vec4 texInfo[4];\n"
	"};\n"

	"in vec4 viPos;\n"
	"in vec4 viPriCol;\n"

	"out vec4 fiPriCol;\n"
	"out vec2 fiTC[3];\n"

	"vec2 t0;\n"

	"void main() {\n"
		"t0.x = dot(viPos.xyz, fxInfo.xyz) - fxInfo.w;\n"
		"t0.y = dot(viPos.xyz, fyInfo.xyz) - fyInfo.w;\n"

		"fiTC[0].xy = (t0.xy - texInfo[0].xy) * texInfo[0].zw;\n"
		"fiTC[1].xy = (t0.xy - texInfo[1].xy) * texInfo[1].zw;\n"
		"fiTC[2].xy = (t0.xy - texInfo[2].xy) * texInfo[2].zw;\n"

		"fiPriCol = viPriCol;\n"
		"gl_Position = trMat * viPos;\n"
	"}\n"
;

static const char *g_vsComplexSurfaceQuadTexture =
	"#version 330\n"

	"layout(std140) uniform trInfo {\n"
		"mat4 trMat;\n"
	"};\n"

	"layout(std140) uniform csVertInfo {\n"
		"vec4 fxInfo;\n"
		"vec4 fyInfo;\n"
		"vec4 texInfo[4];\n"
	"};\n"

	"in vec4 viPos;\n"
	"in vec4 viPriCol;\n"

	"out vec4 fiPriCol;\n"
	"out vec2 fiTC[4];\n"

	"vec2 t0;\n"

	"void main() {\n"
		"t0.x = dot(viPos.xyz, fxInfo.xyz) - fxInfo.w;\n"
		"t0.y = dot(viPos.xyz, fyInfo.xyz) - fyInfo.w;\n"

		"fiTC[0].xy = (t0.xy - texInfo[0].xy) * texInfo[0].zw;\n"
		"fiTC[1].xy = (t0.xy - texInfo[1].xy) * texInfo[1].zw;\n"
		"fiTC[2].xy = (t0.xy - texInfo[2].xy) * texInfo[2].zw;\n"
		"fiTC[3].xy = (t0.xy - texInfo[3].xy) * texInfo[3].zw;\n"

		"fiPriCol = viPriCol;\n"
		"gl_Position = trMat * viPos;\n"
	"}\n"
;

static const char *g_vsDetailTexture =
	"#version 330\n"

	"layout(std140) uniform trInfo {\n"
		"mat4 trMat;\n"
	"};\n"

	"layout(std140) uniform csVertInfo {\n"
		"vec4 fxInfo;\n"
		"vec4 fyInfo;\n"
		"vec4 texInfo[4];\n"
	"};\n"

	"in vec4 viPos;\n"
	"in vec4 viPriCol;\n"

	"out vec4 fiPriCol;\n"
	"out vec4 fiDetailTC;\n"
	"out float fiPosZ;\n"

	"const float detailScale = 4.223;\n"

	"vec2 t0;\n"
	"vec4 tDetail;\n"

	"void main() {\n"
		"t0.x = dot(viPos.xyz, fxInfo.xyz) - fxInfo.w;\n"
		"t0.y = dot(viPos.xyz, fyInfo.xyz) - fyInfo.w;\n"

		"tDetail.xy = (t0.xy - texInfo[0].xy) * texInfo[0].zw;\n"
		"tDetail.zw = tDetail.xy * detailScale;\n"
		"fiDetailTC = tDetail;\n"

		"fiPosZ = viPos.z;\n"
		"fiPriCol = viPriCol;\n"
		"gl_Position = trMat * viPos;\n"
	"}\n"
;

static const char *g_vsComplexSurfaceSingleTextureAndDetailTexture =
	"#version 330\n"

	"layout(std140) uniform trInfo {\n"
		"mat4 trMat;\n"
	"};\n"

	"layout(std140) uniform csVertInfo {\n"
		"vec4 fxInfo;\n"
		"vec4 fyInfo;\n"
		"vec4 texInfo[4];\n"
	"};\n"

	"in vec4 viPos;\n"
	"in vec4 viPriCol;\n"

	"out vec4 fiPriCol;\n"
	"out vec2 fiTC[1];\n"
	"out vec4 fiDetailTC;\n"
	"out float fiPosZ;\n"

	"const float detailScale = 4.223;\n"

	"vec2 t0;\n"
	"vec4 tDetail;\n"

	"void main() {\n"
		"t0.x = dot(viPos.xyz, fxInfo.xyz) - fxInfo.w;\n"
		"t0.y = dot(viPos.xyz, fyInfo.xyz) - fyInfo.w;\n"

		"fiTC[0].xy = (t0.xy - texInfo[0].xy) * texInfo[0].zw;\n"

		"tDetail.xy = (t0.xy - texInfo[1].xy) * texInfo[1].zw;\n"
		"tDetail.zw = tDetail.xy * detailScale;\n"
		"fiDetailTC = tDetail;\n"

		"fiPosZ = viPos.z;\n"
		"fiPriCol = viPriCol;\n"
		"gl_Position = trMat * viPos;\n"
	"}\n"
;

static const char *g_vsComplexSurfaceDualTextureAndDetailTexture =
	"#version 330\n"

	"layout(std140) uniform trInfo {\n"
		"mat4 trMat;\n"
	"};\n"

	"layout(std140) uniform csVertInfo {\n"
		"vec4 fxInfo;\n"
		"vec4 fyInfo;\n"
		"vec4 texInfo[4];\n"
	"};\n"

	"in vec4 viPos;\n"
	"in vec4 viPriCol;\n"

	"out vec4 fiPriCol;\n"
	"out vec2 fiTC[2];\n"
	"out vec4 fiDetailTC;\n"
	"out float fiPosZ;\n"

	"const float detailScale = 4.223;\n"

	"vec2 t0;\n"
	"vec4 tDetail;\n"

	"void main() {\n"
		"t0.x = dot(viPos.xyz, fxInfo.xyz) - fxInfo.w;\n"
		"t0.y = dot(viPos.xyz, fyInfo.xyz) - fyInfo.w;\n"

		"fiTC[0].xy = (t0.xy - texInfo[0].xy) * texInfo[0].zw;\n"
		"fiTC[1].xy = (t0.xy - texInfo[1].xy) * texInfo[1].zw;\n"

		"tDetail.xy = (t0.xy - texInfo[2].xy) * texInfo[2].zw;\n"
		"tDetail.zw = tDetail.xy * detailScale;\n"
		"fiDetailTC = tDetail;\n"

		"fiPosZ = viPos.z;\n"
		"fiPriCol = viPriCol;\n"
		"gl_Position = trMat * viPos;\n"
	"}\n"
;


/*-----------------------------------------------------------------------------
	Fragment programs.
-----------------------------------------------------------------------------*/

static const char *g_fsDefaultRenderingState =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec4 fiPriCol;\n"
	"in vec2 fiTC[1];\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"

	"vec4 t0;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"
		"t0 = t0 * fiPriCol;\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsDefaultRenderingStateWithFog =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec4 fiPriCol;\n"
	"in vec4 oSecCol;\n"
	"in vec2 fiTC[1];\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"

	"vec4 t0;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"
		"t0.a = t0.a * fiPriCol.a;\n"
		"t0.rgb = (t0.rgb * fiPriCol.rgb) + oSecCol.rgb;\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsDefaultRenderingStateWithLinearFog =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec4 fiPriCol;\n"
	"in vec2 fiTC[1];\n"
	"in float fiPosZ;\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"

	"vec4 t0;\n"
	"float fogBlend;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"
		"t0 = t0 * fiPriCol;\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"fogBlend = clamp((-fiPosZ * fogInfo.x) + fogInfo.y, 0.0, 1.0);\n"
		"t0.rgb = mix(fogColor.rgb, t0.rgb, fogBlend);\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsComplexSurfaceSingleTexture =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec2 fiTC[1];\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"

	"vec4 t0;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsComplexSurfaceSingleTextureNoLod =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec2 fiTC[1];\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"

	"vec4 t0;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0]);\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsComplexSurfaceDualTextureModulated =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec2 fiTC[2];\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"
	"uniform sampler2D tex1;\n"

	"vec4 t0;\n"
	"vec4 t1;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"
		"t1 = texture(tex1, fiTC[1]);\n"

		"t0 = t0 * t1;\n"
		"t0.rgb = t0.rgb * blendInfo.aaa;\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsComplexSurfaceTripleTextureModulated =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec2 fiTC[3];\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"
	"uniform sampler2D tex1;\n"
	"uniform sampler2D tex2;\n"

	"vec4 t0;\n"
	"vec4 t1;\n"
	"vec4 t2;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"
		"t1 = texture(tex1, fiTC[1]);\n"
		"t2 = texture(tex2, fiTC[2]);\n"

		"t0 = t0 * t1;\n"
		"t0.rgb = t0.rgb * blendInfo.aaa;\n"

		"t0 = t0 * t2;\n"
		"t0.rgb = t0.rgb * blendInfo.aaa;\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsComplexSurfaceSingleTextureWithFog =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec2 fiTC[2];\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"
	"uniform sampler2D tex1;\n"

	"vec4 t0;\n"
	"vec4 tFog;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"
		"tFog = texture(tex1, fiTC[1]);\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"tFog.a = 1.0 - tFog.a;\n"
		"t0.rgb = (t0.rgb * tFog.aaa) + tFog.rgb;\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsComplexSurfaceDualTextureModulatedWithFog =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec2 fiTC[3];\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"
	"uniform sampler2D tex1;\n"
	"uniform sampler2D tex2;\n"

	"vec4 t0;\n"
	"vec4 t1;\n"
	"vec4 tFog;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"
		"t1 = texture(tex1, fiTC[1]);\n"
		"tFog = texture(tex2, fiTC[2]);\n"

		"t0 = t0 * t1;\n"
		"t0.rgb = t0.rgb * blendInfo.aaa;\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"tFog.a = 1.0 - tFog.a;\n"
		"t0.rgb = (t0.rgb * tFog.aaa) + tFog.rgb;\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsComplexSurfaceTripleTextureModulatedWithFog =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec2 fiTC[4];\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"
	"uniform sampler2D tex1;\n"
	"uniform sampler2D tex2;\n"
	"uniform sampler2D tex3;\n"

	"vec4 t0;\n"
	"vec4 t1;\n"
	"vec4 t2;\n"
	"vec4 tFog;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"
		"t1 = texture(tex1, fiTC[1]);\n"
		"t2 = texture(tex2, fiTC[2]);\n"
		"tFog = texture(tex3, fiTC[3]);\n"

		"t0 = t0 * t1;\n"
		"t0.rgb = t0.rgb * blendInfo.aaa;\n"

		"t0 = t0 * t2;\n"
		"t0.rgb = t0.rgb * blendInfo.aaa;\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"tFog.a = 1.0 - tFog.a;\n"
		"t0.rgb = (t0.rgb * tFog.aaa) + tFog.rgb;\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsDetailTexture =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec4 fiPriCol;\n"
	"in vec4 fiDetailTC;\n"
	"in float fiPosZ;\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"

	"const float rcpDetailNearZ = 0.002631578947;\n"

	"vec4 tDetail0v;\n"
	"float tBlend;\n"

	"void main() {\n"
		"tBlend = clamp(fiPosZ * rcpDetailNearZ, 0.0, 1.0);\n"
		"if (tBlend > 0.999) discard;\n"

		"tDetail0v = texture(tex0, fiDetailTC.xy);\n"
		"tDetail0v = mix(tDetail0v, fiPriCol, tBlend);\n"

		"fragColor = tDetail0v;\n"
	"}\n"
;

static const char *g_fsDetailTextureTwoLayer =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec4 fiPriCol;\n"
	"in vec4 fiDetailTC;\n"
	"in float fiPosZ;\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"

	"const float rcpDetailNearZ = 0.002631578947;\n"
	"const float detailScale = 4.223;\n"

	"vec4 tDetail0v;\n"
	"vec4 tDetail1v;\n"
	"float tBlend;\n"

	"void main() {\n"
		"tBlend = clamp(fiPosZ * rcpDetailNearZ, 0.0, 1.0);\n"
		"if (tBlend > 0.999) discard;\n"

		"tDetail0v = texture(tex0, fiDetailTC.xy);\n"
		"tDetail0v = mix(tDetail0v, fiPriCol, tBlend);\n"

		"tBlend = clamp(tBlend * detailScale, 0.0, 1.0);\n"

		"tDetail1v = texture(tex0, fiDetailTC.zw);\n"
		"tDetail1v = mix(tDetail1v, fiPriCol, tBlend);\n"

		"fragColor = (tDetail0v * tDetail1v) * 2.0;\n"
	"}\n"
;

static const char *g_fsSingleTextureAndDetailTexture =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec4 fiPriCol;\n"
	"in vec2 fiTC[1];\n"
	"in vec4 fiDetailTC;\n"
	"in float fiPosZ;\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"
	"uniform sampler2D tex1;\n"

	"const float rcpDetailNearZ = 0.002631578947;\n"

	"vec4 t0;\n"
	"vec4 tDetail0v;\n"
	"float tBlend;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"tBlend = clamp(fiPosZ * rcpDetailNearZ, 0.0, 1.0);\n"

		"if (tBlend <= 0.999) {\n"
			"tDetail0v = texture(tex1, fiDetailTC.xy);\n"
			"tDetail0v = mix(tDetail0v, fiPriCol, tBlend);\n"

			"t0.rgb = (t0.rgb * tDetail0v.rgb) * 2.0;\n"
		"}\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsSingleTextureAndDetailTextureTwoLayer =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec4 fiPriCol;\n"
	"in vec2 fiTC[1];\n"
	"in vec4 fiDetailTC;\n"
	"in float fiPosZ;\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"
	"uniform sampler2D tex1;\n"

	"const float rcpDetailNearZ = 0.002631578947;\n"
	"const float detailScale = 4.223;\n"

	"vec4 t0;\n"
	"vec4 tDetail0v;\n"
	"vec4 tDetail1v;\n"
	"float tBlend;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"tBlend = clamp(fiPosZ * rcpDetailNearZ, 0.0, 1.0);\n"

		"if (tBlend <= 0.999) {\n"
			"tDetail0v = texture(tex1, fiDetailTC.xy);\n"
			"tDetail0v = mix(tDetail0v, fiPriCol, tBlend);\n"

			"t0.rgb = (t0.rgb * tDetail0v.rgb) * 2.0;\n"

			"tBlend = clamp(tBlend * detailScale, 0.0, 1.0);\n"

			"tDetail1v = texture(tex1, fiDetailTC.zw);\n"
			"tDetail1v = mix(tDetail1v, fiPriCol, tBlend);\n"

			"t0.rgb = (t0.rgb * tDetail1v.rgb) * 2.0;\n"
		"}\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsDualTextureAndDetailTexture =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec4 fiPriCol;\n"
	"in vec2 fiTC[2];\n"
	"in vec4 fiDetailTC;\n"
	"in float fiPosZ;\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"
	"uniform sampler2D tex1;\n"
	"uniform sampler2D tex2;\n"

	"const float rcpDetailNearZ = 0.002631578947;\n"

	"vec4 t0;\n"
	"vec4 t1;\n"
	"vec4 tDetail0v;\n"
	"float tBlend;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"
		"t1 = texture(tex1, fiTC[1]);\n"

		"t0 = t0 * t1;\n"
		"t0.rgb = t0.rgb * blendInfo.aaa;\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"tBlend = clamp(fiPosZ * rcpDetailNearZ, 0.0, 1.0);\n"

		"if (tBlend <= 0.999) {\n"
			"tDetail0v = texture(tex2, fiDetailTC.xy);\n"
			"tDetail0v = mix(tDetail0v, fiPriCol, tBlend);\n"

			"t0.rgb = (t0.rgb * tDetail0v.rgb) * 2.0;\n"
		"}\n"

		"fragColor = t0;\n"
	"}\n"
;

static const char *g_fsDualTextureAndDetailTextureTwoLayer =
	"#version 330\n"
	"precision highp float;\n"

	"layout(std140) uniform fragInfo {\n"
		"vec4 blendInfo;\n"
		"vec4 fogColor;\n"
		"vec4 fogInfo;\n"
	"};\n"

	"in vec4 fiPriCol;\n"
	"in vec2 fiTC[2];\n"
	"in vec4 fiDetailTC;\n"
	"in float fiPosZ;\n"

	"layout(location = 0) out vec4 fragColor;\n"

	"uniform sampler2D tex0;\n"
	"uniform sampler2D tex1;\n"
	"uniform sampler2D tex2;\n"

	"const float rcpDetailNearZ = 0.002631578947;\n"
	"const float detailScale = 4.223;\n"

	"vec4 t0;\n"
	"vec4 t1;\n"
	"vec4 tDetail0v;\n"
	"vec4 tDetail1v;\n"
	"float tBlend;\n"

	"void main() {\n"
		"t0 = texture(tex0, fiTC[0], blendInfo.y);\n"
		"t1 = texture(tex1, fiTC[1]);\n"

		"t0 = t0 * t1;\n"
		"t0.rgb = t0.rgb * blendInfo.aaa;\n"

		"if (t0.a <= blendInfo.x) discard;\n"

		"tBlend = clamp(fiPosZ * rcpDetailNearZ, 0.0, 1.0);\n"

		"if (tBlend <= 0.999) {\n"
			"tDetail0v = texture(tex2, fiDetailTC.xy);\n"
			"tDetail0v = mix(tDetail0v, fiPriCol, tBlend);\n"

			"t0.rgb = (t0.rgb * tDetail0v.rgb) * 2.0;\n"

			"tBlend = clamp(tBlend * detailScale, 0.0, 1.0);\n"

			"tDetail1v = texture(tex2, fiDetailTC.zw);\n"
			"tDetail1v = mix(tDetail1v, fiPriCol, tBlend);\n"

			"t0.rgb = (t0.rgb * tDetail1v.rgb) * 2.0;\n"
		"}\n"

		"fragColor = t0;\n"
	"}\n"
;


/*-----------------------------------------------------------------------------
	OpenGL3Drv.
-----------------------------------------------------------------------------*/

IMPLEMENT_CLASS(UOpenGLRenderDevice);


int UOpenGLRenderDevice::dbgPrintf(const char *format, ...) {
	const unsigned int DBG_PRINT_BUF_SIZE = 1024;
	char dbgPrintBuf[DBG_PRINT_BUF_SIZE];
	va_list vaArgs;
	int iRet = 0;

	va_start(vaArgs, format);

#ifdef _WIN32
#pragma warning(push)
#pragma warning(disable : 4996)
	iRet = vsnprintf(dbgPrintBuf, DBG_PRINT_BUF_SIZE, format, vaArgs);
	dbgPrintBuf[DBG_PRINT_BUF_SIZE - 1] = '\0';
#pragma warning(pop)

	TCHAR_CALL_OS(OutputDebugStringW(appFromAnsi(dbgPrintBuf)), OutputDebugStringA(dbgPrintBuf));
#endif

	va_end(vaArgs);

	return iRet;
}


void UOpenGLRenderDevice::StaticConstructor() {
	guard(UOpenGLRenderDevice::StaticConstructor);

#ifdef UTGLR_DX_BUILD
	const UBOOL UTGLR_DEFAULT_OneXBlending = 1;
#else
	const UBOOL UTGLR_DEFAULT_OneXBlending = 0;
#endif

#if defined UTGLR_DX_BUILD || defined UTGLR_RUNE_BUILD
	const UBOOL UTGLR_DEFAULT_UseS3TC = 0;
#else
	const UBOOL UTGLR_DEFAULT_UseS3TC = 1;
#endif

#if defined UTGLR_DX_BUILD || defined UTGLR_RUNE_BUILD
	const UBOOL UTGLR_DEFAULT_ZRangeHack = 0;
#else
	const UBOOL UTGLR_DEFAULT_ZRangeHack = 1;
#endif

#define CPP_PROPERTY_LOCAL(_name) _name, CPP_PROPERTY(_name)
#define CPP_PROPERTY_LOCAL_DCV(_name) DCV._name, CPP_PROPERTY(DCV._name)

	//Set parameter defaults and add parameters
	SC_AddFloatConfigParam(TEXT("LODBias"), CPP_PROPERTY_LOCAL(LODBias), 0.0f);
	SC_AddFloatConfigParam(TEXT("GammaOffset"), CPP_PROPERTY_LOCAL(GammaOffset), 0.0f);
	SC_AddFloatConfigParam(TEXT("GammaOffsetRed"), CPP_PROPERTY_LOCAL(GammaOffsetRed), 0.0f);
	SC_AddFloatConfigParam(TEXT("GammaOffsetGreen"), CPP_PROPERTY_LOCAL(GammaOffsetGreen), 0.0f);
	SC_AddFloatConfigParam(TEXT("GammaOffsetBlue"), CPP_PROPERTY_LOCAL(GammaOffsetBlue), 0.0f);
	SC_AddBoolConfigParam(1,  TEXT("GammaCorrectScreenshots"), CPP_PROPERTY_LOCAL(GammaCorrectScreenshots), 0);
	SC_AddBoolConfigParam(0,  TEXT("OneXBlending"), CPP_PROPERTY_LOCAL(OneXBlending), UTGLR_DEFAULT_OneXBlending);
	SC_AddIntConfigParam(TEXT("MinLogTextureSize"), CPP_PROPERTY_LOCAL(MinLogTextureSize), 0);
	SC_AddIntConfigParam(TEXT("MaxLogTextureSize"), CPP_PROPERTY_LOCAL(MaxLogTextureSize), 12);
	SC_AddBoolConfigParam(4,  TEXT("ShareLists"), CPP_PROPERTY_LOCAL(ShareLists), 0);
	SC_AddBoolConfigParam(3,  TEXT("UsePrecache"), CPP_PROPERTY_LOCAL(UsePrecache), 0);
	SC_AddBoolConfigParam(2,  TEXT("UseTrilinear"), CPP_PROPERTY_LOCAL(UseTrilinear), 0);
	SC_AddBoolConfigParam(1,  TEXT("UseS3TC"), CPP_PROPERTY_LOCAL(UseS3TC), UTGLR_DEFAULT_UseS3TC);
	SC_AddBoolConfigParam(0,  TEXT("Use16BitTextures"), CPP_PROPERTY_LOCAL(Use16BitTextures), 0);
	SC_AddIntConfigParam(TEXT("MaxAnisotropy"), CPP_PROPERTY_LOCAL(MaxAnisotropy), 0);
	SC_AddBoolConfigParam(0,  TEXT("NoFiltering"), CPP_PROPERTY_LOCAL(NoFiltering), 0);
	SC_AddIntConfigParam(TEXT("MaxTMUnits"), CPP_PROPERTY_LOCAL(MaxTMUnits), 0);
	SC_AddIntConfigParam(TEXT("RefreshRate"), CPP_PROPERTY_LOCAL(RefreshRate), 0);
	SC_AddIntConfigParam(TEXT("DetailMax"), CPP_PROPERTY_LOCAL(DetailMax), 0);
	SC_AddBoolConfigParam(7,  TEXT("ColorizeDetailTextures"), CPP_PROPERTY_LOCAL(ColorizeDetailTextures), 0);
	SC_AddBoolConfigParam(6,  TEXT("SinglePassFog"), CPP_PROPERTY_LOCAL(SinglePassFog), 1);
	SC_AddBoolConfigParam(5,  TEXT("SinglePassDetail"), CPP_PROPERTY_LOCAL_DCV(SinglePassDetail), 0);
	SC_AddBoolConfigParam(4,  TEXT("UseSSE"), CPP_PROPERTY_LOCAL(UseSSE), 1);
	SC_AddBoolConfigParam(3,  TEXT("UseSSE2"), CPP_PROPERTY_LOCAL(UseSSE2), 1);
	SC_AddBoolConfigParam(2,  TEXT("UseTexIdPool"), CPP_PROPERTY_LOCAL(UseTexIdPool), 1);
	SC_AddBoolConfigParam(1,  TEXT("UseTexPool"), CPP_PROPERTY_LOCAL(UseTexPool), 1);
	SC_AddBoolConfigParam(0,  TEXT("CacheStaticMaps"), CPP_PROPERTY_LOCAL(CacheStaticMaps), 1);
	SC_AddIntConfigParam(TEXT("DynamicTexIdRecycleLevel"), CPP_PROPERTY_LOCAL(DynamicTexIdRecycleLevel), 100);
	SC_AddBoolConfigParam(0,  TEXT("TexDXT1ToDXT3"), CPP_PROPERTY_LOCAL(TexDXT1ToDXT3), 0);
	SC_AddIntConfigParam(TEXT("SwapInterval"), CPP_PROPERTY_LOCAL(SwapInterval), -1);
	SC_AddIntConfigParam(TEXT("FrameRateLimit"), CPP_PROPERTY_LOCAL(FrameRateLimit), 0);
	SC_AddBoolConfigParam(3,  TEXT("SceneNodeHack"), CPP_PROPERTY_LOCAL(SceneNodeHack), 1);
	SC_AddBoolConfigParam(2,  TEXT("SmoothMaskedTextures"), CPP_PROPERTY_LOCAL(SmoothMaskedTextures), 0);
	SC_AddBoolConfigParam(1,  TEXT("MaskedTextureHack"), CPP_PROPERTY_LOCAL(MaskedTextureHack), 1);
	SC_AddBoolConfigParam(0,  TEXT("UseAA"), CPP_PROPERTY_LOCAL(UseAA), 0);
	SC_AddIntConfigParam(TEXT("NumAASamples"), CPP_PROPERTY_LOCAL(NumAASamples), 4);
	SC_AddBoolConfigParam(1,  TEXT("NoAATiles"), CPP_PROPERTY_LOCAL(NoAATiles), 1);
	SC_AddBoolConfigParam(0,  TEXT("ZRangeHack"), CPP_PROPERTY_LOCAL(ZRangeHack), UTGLR_DEFAULT_ZRangeHack);

#undef CPP_PROPERTY_LOCAL
#undef CPP_PROPERTY_LOCAL_DCV

	//Driver flags
	SpanBased				= 0;
	SupportsFogMaps			= 1;
#ifdef UTGLR_RUNE_BUILD
	SupportsDistanceFog		= 1;
#else
	SupportsDistanceFog		= 0;
#endif
	FullscreenOnly			= 0;

	SupportsLazyTextures	= 0;
	PrefersDeferredLoad		= 0;

	//Invalidate fixed texture ids
	m_noTextureId = 0;

	//Mark vertex buffer names as not allocated
	m_allocatedVertexBufferNames = false;

	//Mark shader names as not allocated
	m_allocatedShaderNames = false;

	//Frame rate limit timer not yet initialized
	m_frameRateLimitTimerInitialized = false;

	unguard;
}


void UOpenGLRenderDevice::SC_AddBoolConfigParam(DWORD BitMaskOffset, const TCHAR *pName, UBOOL &param, ECppProperty EC_CppProperty, INT InOffset, UBOOL defaultValue) {
	param = (((defaultValue) != 0) ? 1 : 0) << BitMaskOffset; //Doesn't exactly work like a UBOOL "// Boolean 0 (false) or 1 (true)."
	new(GetClass(), pName, RF_Public)UBoolProperty(EC_CppProperty, InOffset, TEXT("Options"), CPF_Config);
}

void UOpenGLRenderDevice::SC_AddIntConfigParam(const TCHAR *pName, INT &param, ECppProperty EC_CppProperty, INT InOffset, INT defaultValue) {
	param = defaultValue;
	new(GetClass(), pName, RF_Public)UIntProperty(EC_CppProperty, InOffset, TEXT("Options"), CPF_Config);
}

void UOpenGLRenderDevice::SC_AddFloatConfigParam(const TCHAR *pName, FLOAT &param, ECppProperty EC_CppProperty, INT InOffset, FLOAT defaultValue) {
	param = defaultValue;
	new(GetClass(), pName, RF_Public)UFloatProperty(EC_CppProperty, InOffset, TEXT("Options"), CPF_Config);
}


void UOpenGLRenderDevice::DbgPrintInitParam(const char *pName, INT value) {
	dbgPrintf("utglr: %s = %d\n", pName, value);
	return;
}

void UOpenGLRenderDevice::DbgPrintInitParam(const char *pName, FLOAT value) {
	dbgPrintf("utglr: %s = %f\n", pName, value);
	return;
}


#ifdef UTGLR_INCLUDE_SSE_CODE
bool UOpenGLRenderDevice::CPU_DetectCPUID(void) {
	//Check for cpuid instruction support
	__try {
		__asm {
			//CPUID function 0
			xor eax, eax
			cpuid
		}
	}
	__except (EXCEPTION_EXECUTE_HANDLER) {
		return false;
	}

	return true;
}

bool UOpenGLRenderDevice::CPU_DetectSSE(void) {
	bool bSupportsSSE;

	//Check for cpuid instruction support
	if (CPU_DetectCPUID() != true) {
		return false;
	}

	//Check for SSE support
	bSupportsSSE = false;
	__asm {
		//CPUID function 1
		mov eax, 1
		cpuid

		//Check the SSE bit
		test edx, 0x02000000
		jz l_no_sse

		//Set bSupportsSSE to true
		mov bSupportsSSE, 1

l_no_sse:
	}

	//Return if no CPU SSE support
	if (bSupportsSSE == false) {
		return bSupportsSSE;
	}

	//Check for SSE OS support
	__try {
		__asm {
			//Execute SSE instruction
			xorps xmm0, xmm0
		}
	}
	__except (EXCEPTION_EXECUTE_HANDLER) {
		//Clear SSE support flag
		bSupportsSSE = false;
	}

	return bSupportsSSE;
}

bool UOpenGLRenderDevice::CPU_DetectSSE2(void) {
	bool bSupportsSSE2;

	//Check for cpuid instruction support
	if (CPU_DetectCPUID() != true) {
		return false;
	}

	//Check for SSE2 support
	bSupportsSSE2 = false;
	__asm {
		//CPUID function 1
		mov eax, 1
		cpuid

		//Check the SSE2 bit
		test edx, 0x04000000
		jz l_no_sse2

		//Set bSupportsSSE2 to true
		mov bSupportsSSE2, 1

l_no_sse2:
	}

	//Return if no CPU SSE2 support
	if (bSupportsSSE2 == false) {
		return bSupportsSSE2;
	}

	//Check for SSE2 OS support
	__try {
		__asm {
			//Execute SSE2 instruction
			xorpd xmm0, xmm0
		}
	}
	__except (EXCEPTION_EXECUTE_HANDLER) {
		//Clear SSE2 support flag
		bSupportsSSE2 = false;
	}

	return bSupportsSSE2;
}
#endif //UTGLR_INCLUDE_SSE_CODE


void UOpenGLRenderDevice::InitFrameRateLimitTimerSafe(void) {
	//Only initialize once
	if (m_frameRateLimitTimerInitialized) {
		return;
	}
	m_frameRateLimitTimerInitialized = true;

#ifdef _WIN32
	//Request high resolution timer
	timeBeginPeriod(1);
#endif

	return;
}

void UOpenGLRenderDevice::ShutdownFrameRateLimitTimer(void) {
	//Only shutdown once
	if (!m_frameRateLimitTimerInitialized) {
		return;
	}
	m_frameRateLimitTimerInitialized = false;

#ifdef _WIN32
	//Release high resolution timer
	timeEndPeriod(1);
#endif

	return;
}


static void FASTCALL Buffer3Verts(UOpenGLRenderDevice *pRD, FTransTexture** Pts) {
	FGLTexCoord *pTexCoordArray = &pRD->m_pTexCoordArray[0][pRD->m_bufferedVerts];
	FGLNormal *pNormalArray = &pRD->m_pNormalArray[pRD->m_bufferedVerts];
	FGLVertexColor *pVertexColorArray = &pRD->m_pVertexColorArray[pRD->m_bufferedVerts];
	FGLSecondaryColor *pSecondaryColorArray = &pRD->m_pSecondaryColorArray[pRD->m_bufferedVerts];
	pRD->m_bufferedVerts += 3;
	for (INT i = 0; i < 3; i++) {
		const FTransTexture* P = *Pts++;

		pTexCoordArray->u = P->U * pRD->TexInfo[0].UMult;
		pTexCoordArray->v = P->V * pRD->TexInfo[0].VMult;
		pTexCoordArray++;

		if (pRD->m_requestedColorFlags & UOpenGLRenderDevice::CF_NORMAL_ARRAY) {
			pNormalArray->x = P->Normal.X;
			pNormalArray->y = P->Normal.Y;
			pNormalArray->z = P->Normal.Z;
			pNormalArray++;
		}

		pVertexColorArray->x = P->Point.X;
		pVertexColorArray->y = P->Point.Y;
		pVertexColorArray->z = P->Point.Z;

		if (pRD->m_requestedColorFlags & UOpenGLRenderDevice::CF_FOG_MODE) {
			FLOAT f255_Times_One_Minus_FogW = 255.0f * (1.0f - P->Fog.W);

			pVertexColorArray->color = UOpenGLRenderDevice::FPlaneTo_RGBScaled_A255(&P->Light, f255_Times_One_Minus_FogW);

			pSecondaryColorArray->specular = UOpenGLRenderDevice::FPlaneTo_RGB_A0(&P->Fog);
			pSecondaryColorArray++;
		}
		else if (pRD->m_requestedColorFlags & UOpenGLRenderDevice::CF_COLOR_ARRAY) {
#ifdef UTGLR_RUNE_BUILD
			pVertexColorArray->color = UOpenGLRenderDevice::FPlaneTo_RGB_Aub(&P->Light, pRD->m_gpAlpha);
#else
			pVertexColorArray->color = UOpenGLRenderDevice::FPlaneTo_RGB_A255(&P->Light);
#endif
		}
		else {
			pVertexColorArray->color = 0xFFFFFFFF;
		}
		pVertexColorArray++;
	}
}

static void FASTCALL Buffer3BasicVerts(UOpenGLRenderDevice *pRD, FTransTexture** Pts) {
	FGLTexCoord *pTexCoordArray = &pRD->m_pTexCoordArray[0][pRD->m_bufferedVerts];
	FGLVertexColor *pVertexColorArray = &pRD->m_pVertexColorArray[pRD->m_bufferedVerts];
	pRD->m_bufferedVerts += 3;
	FLOAT UMult = pRD->TexInfo[0].UMult;
	FLOAT VMult = pRD->TexInfo[0].VMult;
	for (INT i = 0; i < 3; i++) {
		const FTransTexture* P = *Pts++;

		pTexCoordArray->u = P->U * UMult;
		pTexCoordArray->v = P->V * VMult;
		pTexCoordArray++;

		pVertexColorArray->x = P->Point.X;
		pVertexColorArray->y = P->Point.Y;
		pVertexColorArray->z = P->Point.Z;
		pVertexColorArray->color = 0xFFFFFFFF;
		pVertexColorArray++;
	}
}

static void FASTCALL Buffer3ColoredVerts(UOpenGLRenderDevice *pRD, FTransTexture** Pts) {
	FGLTexCoord *pTexCoordArray = &pRD->m_pTexCoordArray[0][pRD->m_bufferedVerts];
	FGLVertexColor *pVertexColorArray = &pRD->m_pVertexColorArray[pRD->m_bufferedVerts];
	pRD->m_bufferedVerts += 3;
	for (INT i = 0; i < 3; i++) {
		const FTransTexture* P = *Pts++;

		pTexCoordArray->u = P->U * pRD->TexInfo[0].UMult;
		pTexCoordArray->v = P->V * pRD->TexInfo[0].VMult;
		pTexCoordArray++;

		pVertexColorArray->x = P->Point.X;
		pVertexColorArray->y = P->Point.Y;
		pVertexColorArray->z = P->Point.Z;
		pVertexColorArray->color = UOpenGLRenderDevice::FPlaneTo_RGB_A255(&P->Light);
		pVertexColorArray++;
	}
}

#ifdef UTGLR_INCLUDE_SSE_CODE
__declspec(naked) static void FASTCALL Buffer3ColoredVerts_SSE(UOpenGLRenderDevice *pRD, FTransTexture** Pts) {
	static float f255 = 255.0f;
	__asm {
		//pRD is in ecx
		//Pts is in edx

		push ebx
		push esi
		push edi

		mov eax, [ecx]UOpenGLRenderDevice.m_bufferedVerts

		lea ebx, [eax*8]
		add ebx, [ecx]UOpenGLRenderDevice.m_pTexCoordArray[0]

		mov edi, eax
		shl edi, 4
		add edi, [ecx]UOpenGLRenderDevice.m_pVertexColorArray

		//m_bufferedVerts += 3
		add eax, 3
		mov [ecx]UOpenGLRenderDevice.m_bufferedVerts, eax

		lea eax, [ecx]UOpenGLRenderDevice.TexInfo
		movss xmm0, [eax]FTexInfo.UMult
		movss xmm1, [eax]FTexInfo.VMult
		movss xmm2, f255

		//Pts in edx
		//Get PtsPlus12B
		lea esi, [edx + 12]

v_loop:
			mov eax, [edx]
			add edx, 4

			movss xmm3, [eax]FTransTexture.U
			mulss xmm3, xmm0
			movss [ebx]FGLTexCoord.u, xmm3
			movss xmm3, [eax]FTransTexture.V
			mulss xmm3, xmm1
			movss [ebx]FGLTexCoord.v, xmm3
			add ebx, TYPE FGLTexCoord

			mov ecx, [eax]FOutVector.Point.X
			mov [edi]FGLVertexColor.x, ecx
			mov ecx, [eax]FOutVector.Point.Y
			mov [edi]FGLVertexColor.y, ecx
			mov ecx, [eax]FOutVector.Point.Z
			mov [edi]FGLVertexColor.z, ecx

			movss xmm3, [eax]FTransSample.Light + 0
			mulss xmm3, xmm2
			movss xmm4, [eax]FTransSample.Light + 4
			mulss xmm4, xmm2
			movss xmm5, [eax]FTransSample.Light + 8
			mulss xmm5, xmm2
			cvtss2si eax, xmm3
			and eax, 255
			cvtss2si ecx, xmm4
			and ecx, 255
			shl ecx, 8
			or eax, ecx
			cvtss2si ecx, xmm5
			shl ecx, 16
			or ecx, 0xFF000000
			or eax, ecx
			mov [edi]FGLVertexColor.color, eax
			add edi, TYPE FGLVertexColor

			cmp edx, esi
			jne v_loop

		pop edi
		pop esi
		pop ebx

		ret
	}
}

__declspec(naked) static void FASTCALL Buffer3ColoredVerts_SSE2(UOpenGLRenderDevice *pRD, FTransTexture** Pts) {
	static __m128 fColorMul = { 255.0f, 255.0f, 255.0f, 0.0f };
	static DWORD alphaOr = 0xFF000000;
	__asm {
		//pRD is in ecx
		//Pts is in edx

		push ebx
		push esi
		push edi

		mov eax, [ecx]UOpenGLRenderDevice.m_bufferedVerts

		lea ebx, [eax*8]
		add ebx, [ecx]UOpenGLRenderDevice.m_pTexCoordArray[0]

		mov edi, eax
		shl edi, 4
		add edi, [ecx]UOpenGLRenderDevice.m_pVertexColorArray

		//m_bufferedVerts += 3
		add eax, 3
		mov [ecx]UOpenGLRenderDevice.m_bufferedVerts, eax

		lea eax, [ecx]UOpenGLRenderDevice.TexInfo
		movss xmm0, [eax]FTexInfo.UMult
		movss xmm1, [eax]FTexInfo.VMult
		movaps xmm2, fColorMul
		movd xmm3, alphaOr

		//Pts in edx
		//Get PtsPlus12B
		lea esi, [edx + 12]

v_loop:
			mov eax, [edx]
			add edx, 4

			movss xmm4, [eax]FTransTexture.U
			mulss xmm4, xmm0
			movss [ebx]FGLTexCoord.u, xmm4
			movss xmm4, [eax]FTransTexture.V
			mulss xmm4, xmm1
			movss [ebx]FGLTexCoord.v, xmm4
			add ebx, TYPE FGLTexCoord

			mov ecx, [eax]FOutVector.Point.X
			mov [edi]FGLVertexColor.x, ecx
			mov ecx, [eax]FOutVector.Point.Y
			mov [edi]FGLVertexColor.y, ecx
			mov ecx, [eax]FOutVector.Point.Z
			mov [edi]FGLVertexColor.z, ecx

			movups xmm4, [eax]FTransSample.Light
			mulps xmm4, xmm2
			cvtps2dq xmm4, xmm4
			packssdw xmm4, xmm4
			packuswb xmm4, xmm4
			por xmm4, xmm3
			movd [edi]FGLVertexColor.color, xmm4
			add edi, TYPE FGLVertexColor

			cmp edx, esi
			jne v_loop

		pop edi
		pop esi
		pop ebx

		ret
	}
}
#endif

static void FASTCALL Buffer3FoggedVerts(UOpenGLRenderDevice *pRD, FTransTexture** Pts) {
	FGLTexCoord *pTexCoordArray = &pRD->m_pTexCoordArray[0][pRD->m_bufferedVerts];
	FGLVertexColor *pVertexColorArray = &pRD->m_pVertexColorArray[pRD->m_bufferedVerts];
	FGLSecondaryColor *pSecondaryColorArray = &pRD->m_pSecondaryColorArray[pRD->m_bufferedVerts];
	pRD->m_bufferedVerts += 3;
	for (INT i = 0; i < 3; i++) {
		const FTransTexture* P = *Pts++;

		pTexCoordArray->u = P->U * pRD->TexInfo[0].UMult;
		pTexCoordArray->v = P->V * pRD->TexInfo[0].VMult;
		pTexCoordArray++;

		pVertexColorArray->x = P->Point.X;
		pVertexColorArray->y = P->Point.Y;
		pVertexColorArray->z = P->Point.Z;
		FLOAT f255_Times_One_Minus_FogW = 255.0f * (1.0f - P->Fog.W);
		pVertexColorArray->color = UOpenGLRenderDevice::FPlaneTo_RGBScaled_A255(&P->Light, f255_Times_One_Minus_FogW);
		pVertexColorArray++;

		pSecondaryColorArray->specular = UOpenGLRenderDevice::FPlaneTo_RGB_A0(&P->Fog);
		pSecondaryColorArray++;
	}
}

#ifdef UTGLR_INCLUDE_SSE_CODE
__declspec(naked) static void FASTCALL Buffer3FoggedVerts_SSE(UOpenGLRenderDevice *pRD, FTransTexture** Pts) {
	static float f255 = 255.0f;
	static float f1 = 1.0f;
	__asm {
		//pRD is in ecx
		//Pts is in edx

		push ebx
		push esi
		push edi
		push ebp
		sub esp, 4

		mov eax, [ecx]UOpenGLRenderDevice.m_bufferedVerts

		lea ebx, [eax*8]
		add ebx, [ecx]UOpenGLRenderDevice.m_pTexCoordArray[0]

		mov edi, eax
		shl edi, 4
		add edi, [ecx]UOpenGLRenderDevice.m_pVertexColorArray

		lea esi, [eax*4]
		add esi, [ecx]UOpenGLRenderDevice.m_pSecondaryColorArray

		//m_bufferedVerts += 3
		add eax, 3
		mov [ecx]UOpenGLRenderDevice.m_bufferedVerts, eax

		lea eax, [ecx]UOpenGLRenderDevice.TexInfo
		movss xmm0, [eax]FTexInfo.UMult
		movss xmm1, [eax]FTexInfo.VMult
		movss xmm2, f255

		//Pts in edx
		//Get PtsPlus12B
		lea ebp, [edx + 12]

v_loop:
			mov eax, [edx]
			add edx, 4

			movss xmm3, [eax]FTransTexture.U
			mulss xmm3, xmm0
			movss [ebx]FGLTexCoord.u, xmm3
			movss xmm3, [eax]FTransTexture.V
			mulss xmm3, xmm1
			movss [ebx]FGLTexCoord.v, xmm3
			add ebx, TYPE FGLTexCoord

			mov [esp], ebx

			movss xmm6, f1
			subss xmm6, [eax]FTransSample.Fog + 12
			mulss xmm6, xmm2

			mov ecx, [eax]FOutVector.Point.X
			mov [edi]FGLVertexColor.x, ecx
			mov ecx, [eax]FOutVector.Point.Y
			mov [edi]FGLVertexColor.y, ecx
			mov ecx, [eax]FOutVector.Point.Z
			mov [edi]FGLVertexColor.z, ecx

			movss xmm3, [eax]FTransSample.Light + 0
			mulss xmm3, xmm6
			movss xmm4, [eax]FTransSample.Light + 4
			mulss xmm4, xmm6
			movss xmm5, [eax]FTransSample.Light + 8
			mulss xmm5, xmm6
			cvtss2si ebx, xmm3
			and ebx, 255
			cvtss2si ecx, xmm4
			and ecx, 255
			shl ecx, 8
			or ebx, ecx
			cvtss2si ecx, xmm5
			shl ecx, 16
			or ecx, 0xFF000000
			or ebx, ecx
			mov [edi]FGLVertexColor.color, ebx
			add edi, TYPE FGLVertexColor

			mov ebx, [esp]

			movss xmm3, [eax]FTransSample.Fog + 0
			mulss xmm3, xmm2
			movss xmm4, [eax]FTransSample.Fog + 4
			mulss xmm4, xmm2
			movss xmm5, [eax]FTransSample.Fog + 8
			mulss xmm5, xmm2
			cvtss2si eax, xmm3
			and eax, 255
			cvtss2si ecx, xmm4
			and ecx, 255
			shl ecx, 8
			or eax, ecx
			cvtss2si ecx, xmm5
			and ecx, 255
			shl ecx, 16
			or eax, ecx
			mov [esi]FGLSecondaryColor.specular, eax
			add esi, TYPE FGLSecondaryColor

			cmp edx, ebp
			jne v_loop

		add esp, 4
		pop ebp
		pop edi
		pop esi
		pop ebx

		ret
	}
}

__declspec(naked) static void FASTCALL Buffer3FoggedVerts_SSE2(UOpenGLRenderDevice *pRD, FTransTexture** Pts) {
	static __m128 fColorMul = { 255.0f, 255.0f, 255.0f, 0.0f };
	static DWORD alphaOr = 0xFF000000;
	static float f1 = 1.0f;
	__asm {
		//pRD is in ecx
		//Pts is in edx

		push ebx
		push esi
		push edi
		push ebp

		mov eax, [ecx]UOpenGLRenderDevice.m_bufferedVerts

		lea ebx, [eax*8]
		add ebx, [ecx]UOpenGLRenderDevice.m_pTexCoordArray[0]

		mov edi, eax
		shl edi, 4
		add edi, [ecx]UOpenGLRenderDevice.m_pVertexColorArray

		lea esi, [eax*4]
		add esi, [ecx]UOpenGLRenderDevice.m_pSecondaryColorArray

		//m_bufferedVerts += 3
		add eax, 3
		mov [ecx]UOpenGLRenderDevice.m_bufferedVerts, eax

		lea eax, [ecx]UOpenGLRenderDevice.TexInfo
		movss xmm0, [eax]FTexInfo.UMult
		movss xmm1, [eax]FTexInfo.VMult
		movaps xmm2, fColorMul
		movd xmm3, alphaOr

		//Pts in edx
		//Get PtsPlus12B
		lea ebp, [edx + 12]

v_loop:
			mov eax, [edx]
			add edx, 4

			movss xmm4, [eax]FTransTexture.U
			mulss xmm4, xmm0
			movss [ebx]FGLTexCoord.u, xmm4
			movss xmm4, [eax]FTransTexture.V
			mulss xmm4, xmm1
			movss [ebx]FGLTexCoord.v, xmm4
			add ebx, TYPE FGLTexCoord

			movss xmm6, f1
			subss xmm6, [eax]FTransSample.Fog + 12
			mulss xmm6, xmm2
			shufps xmm6, xmm6, 0x00

			mov ecx, [eax]FOutVector.Point.X
			mov [edi]FGLVertexColor.x, ecx
			mov ecx, [eax]FOutVector.Point.Y
			mov [edi]FGLVertexColor.y, ecx
			mov ecx, [eax]FOutVector.Point.Z
			mov [edi]FGLVertexColor.z, ecx

			movups xmm4, [eax]FTransSample.Light
			mulps xmm4, xmm6
			cvtps2dq xmm4, xmm4
			packssdw xmm4, xmm4
			packuswb xmm4, xmm4
			por xmm4, xmm3
			movd [edi]FGLVertexColor.color, xmm4
			add edi, TYPE FGLVertexColor

			movups xmm4, [eax]FTransSample.Fog
			mulps xmm4, xmm2
			cvtps2dq xmm4, xmm4
			packssdw xmm4, xmm4
			packuswb xmm4, xmm4
			movd [esi]FGLSecondaryColor.specular, xmm4
			add esi, TYPE FGLSecondaryColor

			cmp edx, ebp
			jne v_loop

		pop ebp
		pop edi
		pop esi
		pop ebx

		ret
	}
}
#endif


//Must be called with (NumPts > 3)
void UOpenGLRenderDevice::BufferAdditionalClippedVerts(FTransTexture** Pts, INT NumPts) {
	INT i;

	i = 3;
	do {
		const FTransTexture* P;
		FGLTexCoord *pTexCoordArray;
		FGLVertexColor *pVertexColorArray;

		P = Pts[0];
		pTexCoordArray = &m_pTexCoordArray[0][m_bufferedVerts];
		pTexCoordArray->u = P->U * TexInfo[0].UMult;
		pTexCoordArray->v = P->V * TexInfo[0].VMult;
		pVertexColorArray = &m_pVertexColorArray[m_bufferedVerts];
		pVertexColorArray->x = P->Point.X;
		pVertexColorArray->y = P->Point.Y;
		pVertexColorArray->z = P->Point.Z;
		if (m_requestedColorFlags & CF_FOG_MODE) {
			FLOAT f255_Times_One_Minus_FogW = 255.0f * (1.0f - P->Fog.W);
			pVertexColorArray->color = FPlaneTo_RGBScaled_A255(&P->Light, f255_Times_One_Minus_FogW);

			m_pSecondaryColorArray[m_bufferedVerts].specular = FPlaneTo_RGB_A0(&P->Fog);
		}
		else if (m_requestedColorFlags & CF_COLOR_ARRAY) {
#ifdef UTGLR_RUNE_BUILD
			pVertexColorArray->color = FPlaneTo_RGB_Aub(&P->Light, m_gpAlpha);
#else
			pVertexColorArray->color = FPlaneTo_RGB_A255(&P->Light);
#endif
		}
		else {
			pVertexColorArray->color = 0xFFFFFFFF;
		}
		m_bufferedVerts++;

		P = Pts[i - 1];
		pTexCoordArray = &m_pTexCoordArray[0][m_bufferedVerts];
		pTexCoordArray->u = P->U * TexInfo[0].UMult;
		pTexCoordArray->v = P->V * TexInfo[0].VMult;
		pVertexColorArray = &m_pVertexColorArray[m_bufferedVerts];
		pVertexColorArray->x = P->Point.X;
		pVertexColorArray->y = P->Point.Y;
		pVertexColorArray->z = P->Point.Z;
		if (m_requestedColorFlags & CF_FOG_MODE) {
			FLOAT f255_Times_One_Minus_FogW = 255.0f * (1.0f - P->Fog.W);
			pVertexColorArray->color = FPlaneTo_RGBScaled_A255(&P->Light, f255_Times_One_Minus_FogW);

			m_pSecondaryColorArray[m_bufferedVerts].specular = FPlaneTo_RGB_A0(&P->Fog);
		}
		else if (m_requestedColorFlags & CF_COLOR_ARRAY) {
#ifdef UTGLR_RUNE_BUILD
			pVertexColorArray->color = FPlaneTo_RGB_Aub(&P->Light, m_gpAlpha);
#else
			pVertexColorArray->color = FPlaneTo_RGB_A255(&P->Light);
#endif
		}
		else {
			pVertexColorArray->color = 0xFFFFFFFF;
		}
		m_bufferedVerts++;

		P = Pts[i];
		pTexCoordArray = &m_pTexCoordArray[0][m_bufferedVerts];
		pTexCoordArray->u = P->U * TexInfo[0].UMult;
		pTexCoordArray->v = P->V * TexInfo[0].VMult;
		pVertexColorArray = &m_pVertexColorArray[m_bufferedVerts];
		pVertexColorArray->x = P->Point.X;
		pVertexColorArray->y = P->Point.Y;
		pVertexColorArray->z = P->Point.Z;
		if (m_requestedColorFlags & CF_FOG_MODE) {
			FLOAT f255_Times_One_Minus_FogW = 255.0f * (1.0f - P->Fog.W);
			pVertexColorArray->color = FPlaneTo_RGBScaled_A255(&P->Light, f255_Times_One_Minus_FogW);

			m_pSecondaryColorArray[m_bufferedVerts].specular = FPlaneTo_RGB_A0(&P->Fog);
		}
		else if (m_requestedColorFlags & CF_COLOR_ARRAY) {
#ifdef UTGLR_RUNE_BUILD
			pVertexColorArray->color = FPlaneTo_RGB_Aub(&P->Light, m_gpAlpha);
#else
			pVertexColorArray->color = FPlaneTo_RGB_A255(&P->Light);
#endif
		}
		else {
			pVertexColorArray->color = 0xFFFFFFFF;
		}
		m_bufferedVerts++;
	} while (++i < NumPts);

	return;
}


void UOpenGLRenderDevice::BuildGammaRamp(float redGamma, float greenGamma, float blueGamma, int brightness, FGammaRamp &ramp) {
	unsigned int u;

	//Parameter clamping
	if (brightness < -50) brightness = -50;
	if (brightness > 50) brightness = 50;

	float rcpRedGamma = 1.0f / (2.5f * redGamma);
	float rcpGreenGamma = 1.0f / (2.5f * greenGamma);
	float rcpBlueGamma = 1.0f / (2.5f * blueGamma);
	for (u = 0; u < 256; u++) {
		int iVal;
		int iValRed, iValGreen, iValBlue;

		//Initial value
		iVal = u;

		//Brightness
		iVal += brightness;
		//Clamping
		if (iVal < 0) iVal = 0;
		if (iVal > 255) iVal = 255;

		//Gamma
		iValRed = (int)appRound((float)appPow(iVal / 255.0f, rcpRedGamma) * 65535.0f);
		iValGreen = (int)appRound((float)appPow(iVal / 255.0f, rcpGreenGamma) * 65535.0f);
		iValBlue = (int)appRound((float)appPow(iVal / 255.0f, rcpBlueGamma) * 65535.0f);

		//Save results
		ramp.red[u] = (_WORD)iValRed;
		ramp.green[u] = (_WORD)iValGreen;
		ramp.blue[u] = (_WORD)iValBlue;
	}

	return;
}

void UOpenGLRenderDevice::BuildGammaRamp(float redGamma, float greenGamma, float blueGamma, int brightness, FByteGammaRamp &ramp) {
	unsigned int u;

	//Parameter clamping
	if (brightness < -50) brightness = -50;
	if (brightness > 50) brightness = 50;

	float rcpRedGamma = 1.0f / (2.5f * redGamma);
	float rcpGreenGamma = 1.0f / (2.5f * greenGamma);
	float rcpBlueGamma = 1.0f / (2.5f * blueGamma);
	for (u = 0; u < 256; u++) {
		int iVal;
		int iValRed, iValGreen, iValBlue;

		//Initial value
		iVal = u;

		//Brightness
		iVal += brightness;
		//Clamping
		if (iVal < 0) iVal = 0;
		if (iVal > 255) iVal = 255;

		//Gamma
		iValRed = (int)appRound((float)appPow(iVal / 255.0f, rcpRedGamma) * 255.0f);
		iValGreen = (int)appRound((float)appPow(iVal / 255.0f, rcpGreenGamma) * 255.0f);
		iValBlue = (int)appRound((float)appPow(iVal / 255.0f, rcpBlueGamma) * 255.0f);

		//Save results
		ramp.red[u] = (BYTE)iValRed;
		ramp.green[u] = (BYTE)iValGreen;
		ramp.blue[u] = (BYTE)iValBlue;
	}

	return;
}

void UOpenGLRenderDevice::SetGamma(FLOAT GammaCorrection) {
	FGammaRamp gammaRamp;

	GammaCorrection += GammaOffset;

	//Do not attempt to set gamma if <= zero
	if (GammaCorrection <= 0.0f) {
		return;
	}

	BuildGammaRamp(GammaCorrection + GammaOffsetRed, GammaCorrection + GammaOffsetGreen, GammaCorrection + GammaOffsetBlue, Brightness, gammaRamp);

	//Remember previous gamma ramp setting
	if (g_gammaFirstTime) {
		if (GetDeviceGammaRamp(m_hDC, &g_originalGammaRamp)) {
			g_haveOriginalGammaRamp = true;
		}
		g_gammaFirstTime = false;
	}

	//Set gamma ramp
	m_setGammaRampSucceeded = false;
	if (SetDeviceGammaRamp(m_hDC, &gammaRamp)) {
		m_setGammaRampSucceeded = true;
		SavedGammaCorrection = GammaCorrection;
	}

	return;
}

void UOpenGLRenderDevice::ResetGamma(void) {
	//Restore gamma ramp if original was successfully saved
	if (g_haveOriginalGammaRamp) {
		HWND hDesktopWnd;
		HDC hDC;

		hDesktopWnd = GetDesktopWindow();
		hDC = GetDC(hDesktopWnd);

		// vogel: grrr, UClient::destroy is called before this gets called so hDC is invalid
		SetDeviceGammaRamp(hDC, &g_originalGammaRamp);

		ReleaseDC(hDesktopWnd, hDC);
	}

	return;
}


bool UOpenGLRenderDevice::IsGLExtensionSupported(const char *pExtensionsString, const char *pExtensionName) {
	const char *pStart;
	const char *pWhere, *pTerminator;

	pStart = pExtensionsString;
	while (1) {
		pWhere = strstr(pStart, pExtensionName);
		if (pWhere == NULL) {
			break;
		}
		pTerminator = pWhere + strlen(pExtensionName);
		if ((pWhere == pStart) || (*(pWhere - 1) == ' ')) {
			if ((*pTerminator == ' ') || (*pTerminator == '\0')) {
				return true;
			}
		}
		pStart = pTerminator;
	}

	return false;
}

bool UOpenGLRenderDevice::IsIndexedGLExtensionSupported(const char *pExtensionName) {
	GLint numExtensions = 0;
	GLint num;
	const char *pExtStr;

	glGetIntegerv(GL_NUM_EXTENSIONS, &numExtensions);
	for (num = 0; num < numExtensions; num++) {
		pExtStr = (const char *)glGetStringi(GL_EXTENSIONS, num);
		if (!strcmp(pExtensionName, pExtStr)) {
			return true;
		}
	}

	return false;
}

bool UOpenGLRenderDevice::GetGL1Proc(void*& ProcAddress, const char *pName) {
	guard(UOpenGLRenderDevice::GetGL1Proc);

	ProcAddress = GetProcAddress(hModuleGlMain, pName);
	if (!ProcAddress) {
		debugf(TEXT("   Missing function '%s' for '%s' support"), appFromAnsi(pName), TEXT("OpenGL 1.x"));
		dbgPrintf("Missing function '%s' for '%s' support\n", pName, "OpenGL 1.x");
		return false;
	}

	return true;

	unguard;
}

bool UOpenGLRenderDevice::GetGL1Procs(void) {
	guard(UOpenGLRenderDevice::GetGL1Procs);

	bool loadOk = true;

	#define GL1_PROC(ret, func, params) loadOk &= GetGL1Proc(*(void**)&func, #func);
	#include "OpenGL1Funcs.h"
	#undef GL1_PROC

	return loadOk;

	unguard;
}

bool UOpenGLRenderDevice::GetGL3Proc(void*& ProcAddress, const char *pName) {
	guard(UOpenGLRenderDevice::GetGL3Proc);

	ProcAddress = wglGetProcAddress(pName);
	if (!ProcAddress) {
		debugf(TEXT("   Missing function '%s' for '%s' support"), appFromAnsi(pName), TEXT("OpenGL 3.3"));
		dbgPrintf("Missing function '%s' for '%s' support\n", pName, "OpenGL 3.3");
		return false;
	}

	return true;

	unguard;
}

bool UOpenGLRenderDevice::GetGL3Procs(void) {
	guard(UOpenGLRenderDevice::GetGL3Procs);

	bool loadOk = true;

	#define GL3_PROC(ret, func, params) loadOk &= GetGL3Proc(*(void**)&func, #func);
	#include "OpenGL3Funcs.h"
	#undef GL3_PROC

	return loadOk;

	unguard;
}

bool UOpenGLRenderDevice::FindGLExt(const char *pName) {
	guard(UOpenGLRenderDevice::FindGLExt);

	bool bRet = IsIndexedGLExtensionSupported(pName);
	if (bRet) {
		debugf(NAME_Init, TEXT("Device supports: %s"), appFromAnsi(pName));
	}
	if (DebugBit(DEBUG_BIT_BASIC)) {
		dbgPrintf("utglr: GL_EXT: %s = %d\n", pName, bRet);
	}

	return bRet;
	unguard;
}

void UOpenGLRenderDevice::GetGLExtProc(void*& ProcAddress, const char *pName, const char *pSupportName, bool& Supports) {
	guard(UOpenGLRenderDevice::GetGLExtProc);

	if (!Supports) {
		return;
	}

	ProcAddress = wglGetProcAddress(pName);
	if (!ProcAddress) {
		Supports = false;

		debugf(TEXT("   Missing function '%s' for '%s' support"), appFromAnsi(pName), appFromAnsi(pSupportName));
		dbgPrintf("Missing function '%s' for '%s' support\n", pName, pSupportName);
	}

	unguard;
}

void UOpenGLRenderDevice::GetGLExtProcs(void) {
	guard(UOpenGLRenderDevice::GetGLExtProcs);
	#define GL_EXT_NAME(name) SUPPORTS##name = FindGLExt(#name + 1);
	#define GL_EXT_PROC(ext, ret, func, params) GetGLExtProc(*(void**)&func, #func, #ext + 1, SUPPORTS##ext);
	#include "OpenGLExtFuncs.h"
	#undef GL_EXT_NAME
	#undef GL_EXT_PROC
	unguard;
}


UBOOL UOpenGLRenderDevice::FailedInitf(const TCHAR* Fmt, ...) {
	TCHAR TempStr[4096];
	GET_VARARGS(TempStr, ARRAY_COUNT(TempStr), Fmt);
	debugf(NAME_Init, TempStr);
	Exit();
	return 0;
}

void UOpenGLRenderDevice::Exit() {
	guard(UOpenGLRenderDevice::Exit);
	check(NumDevices > 0);

	// Shut down RC.
	if (m_hRC) {
		UnsetRes();
	}

	//Reset gamma ramp
	ResetGamma();

	//Timer shutdown
	ShutdownFrameRateLimitTimer();

	// Shut down this GL context. May fail if window was already destroyed.
	if (m_hDC) {
		ReleaseDC(m_hWnd, m_hDC);
	}

	// Shut down global GL.
	if (--NumDevices == 0) {
#if 0 // Broken on some drivers
		// Free modules.
		if (hModuleGlMain)
			verify(FreeLibrary(hModuleGlMain));
#endif
		{
			unsigned int u;

			for (u = 0; u < NUM_CTTree_TREES; u++) {
				m_sharedZeroPrefixBindTrees[u].clear(&m_DWORD_CTTree_Allocator);
			}
			for (u = 0; u < NUM_CTTree_TREES; u++) {
				m_sharedNonZeroPrefixBindTrees[u].clear(&m_QWORD_CTTree_Allocator);
			}
		}
		m_sharedNonZeroPrefixBindChain.mark_as_clear();
		//Texture ids and memory for non zero prefix tex id pool are not freed on exit
		m_sharedRGBA8TexPool.clear(&m_TexPoolMap_Allocator);

		AllContexts.~TArray<HGLRC>();
	}

	unguard;
}

void UOpenGLRenderDevice::ShutdownAfterError() {
	guard(UOpenGLRenderDevice::ShutdownAfterError);

	debugf(NAME_Exit, TEXT("UOpenGLRenderDevice::ShutdownAfterError"));

	if (DebugBit(DEBUG_BIT_BASIC)) {
		dbgPrintf("utglr: ShutdownAfterError\n");
	}

	//ChangeDisplaySettings(NULL, 0);

	//Reset gamma ramp
	ResetGamma();

	unguard;
}


UBOOL UOpenGLRenderDevice::SetRes(INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen) {
	guard(UOpenGLRenderDevice::SetRes);

	//Get debug bits
	{
		INT i = 0;
		if (!GConfig->GetInt(g_pSection, TEXT("DebugBits"), i)) i = 0;
		m_debugBits = i;
	}
	//Display debug bits
	if (DebugBit(DEBUG_BIT_ANY)) dbgPrintf("utglr: DebugBits = %u\n", m_debugBits);

	debugf(TEXT("Enter SetRes()"));

	// If not fullscreen, and color bytes hasn't changed, do nothing.
	if (m_hRC && !Fullscreen && !WasFullscreen && NewColorBytes == Viewport->ColorBytes) {
		if (!Viewport->ResizeViewport(BLIT_HardwarePaint | BLIT_OpenGL, NewX, NewY, NewColorBytes)) {
			return 0;
		}
		glViewport(0, 0, NewX, NewY);

		return 1;
	}

	// Exit res.
	if (m_hRC) {
		debugf(TEXT("UnSetRes() -> hRC != NULL"));
		UnsetRes();
	}

	// Change display settings.
	if (Fullscreen) {
		INT FindX = NewX, FindY = NewY, BestError = MAXINT;
		for (INT i = 0; i < Modes.Num(); i++) {
			if (Modes(i).Z==NewColorBytes*8) {
				INT Error
				=	(Modes(i).X-FindX)*(Modes(i).X-FindX)
				+	(Modes(i).Y-FindY)*(Modes(i).Y-FindY);
				if (Error < BestError) {
					NewX      = Modes(i).X;
					NewY      = Modes(i).Y;
					BestError = Error;
				}
			}
		}
		{
			DEVMODEA dma;
			DEVMODEW dmw;
			bool tryNoRefreshRate = true;

			ZeroMemory(&dma, sizeof(dma));
			ZeroMemory(&dmw, sizeof(dmw));
			dma.dmSize = sizeof(dma);
			dmw.dmSize = sizeof(dmw);
			dma.dmPelsWidth = dmw.dmPelsWidth = NewX;
			dma.dmPelsHeight = dmw.dmPelsHeight = NewY;
			dma.dmBitsPerPel = dmw.dmBitsPerPel = NewColorBytes * 8;
			dma.dmFields = dmw.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT;// | DM_BITSPERPEL;
			if (RefreshRate) {
				dma.dmDisplayFrequency = dmw.dmDisplayFrequency = RefreshRate;
				dma.dmFields |= DM_DISPLAYFREQUENCY;
				dmw.dmFields |= DM_DISPLAYFREQUENCY;

				if (TCHAR_CALL_OS(ChangeDisplaySettingsW(&dmw, CDS_FULLSCREEN), ChangeDisplaySettingsA(&dma, CDS_FULLSCREEN)) != DISP_CHANGE_SUCCESSFUL) {
					debugf(TEXT("ChangeDisplaySettings failed: %ix%i, %i Hz"), NewX, NewY, RefreshRate);
					dma.dmFields &= ~DM_DISPLAYFREQUENCY;
					dmw.dmFields &= ~DM_DISPLAYFREQUENCY;
				}
				else {
					tryNoRefreshRate = false;
				}
			}
			if (tryNoRefreshRate) {
				if (TCHAR_CALL_OS(ChangeDisplaySettingsW(&dmw, CDS_FULLSCREEN), ChangeDisplaySettingsA(&dma, CDS_FULLSCREEN)) != DISP_CHANGE_SUCCESSFUL) {
					debugf(TEXT("ChangeDisplaySettings failed: %ix%i"), NewX, NewY);

					return 0;
				}
			}
		}
	}

	// Change window size.
	UBOOL Result = Viewport->ResizeViewport(Fullscreen ? (BLIT_Fullscreen | BLIT_OpenGL) : (BLIT_HardwarePaint | BLIT_OpenGL), NewX, NewY, NewColorBytes);
	if (!Result) {
		if (Fullscreen) {
			TCHAR_CALL_OS(ChangeDisplaySettingsW(NULL, 0), ChangeDisplaySettingsA(NULL, 0));
		}

		return 0;
	}


	//Get function pointers needed for new OpenGL initialization
	new_gl_init_procs_t initProcs;
	if (!GetNewGLInitProcs(NewColorBytes, &initProcs)) {
		appErrorf(TEXT("Failed to get function pointers needed for new OpenGL initialization."));
	}

	//Set default numDepthBits in case any failures might prevent it from being set later
	m_numDepthBits = 16;

	//Restrict NumAASamples range
	if (NumAASamples < 1) NumAASamples = 1;

	m_usingAA = false;
	m_curAAEnable = true;
	m_defAAEnable = true;
	m_initNumAASamples = NumAASamples;

	HGLRC hShareGLRC = NULL;
	if (ShareLists && AllContexts.Num()) {
		hShareGLRC = AllContexts(0);
	}
	SetNewPixelFormat(NewColorBytes, hShareGLRC, &initProcs);
	AllContexts.AddItem(m_hRC);


	//Reset previous SwapBuffers status to okay
	m_prevSwapBuffersStatus = true;

	//Get OpenGL info
	debugf(NAME_Init, TEXT("GL_VENDOR     : %s"), appFromAnsi((const ANSICHAR *)glGetString(GL_VENDOR)));
	debugf(NAME_Init, TEXT("GL_RENDERER   : %s"), appFromAnsi((const ANSICHAR *)glGetString(GL_RENDERER)));
	debugf(NAME_Init, TEXT("GL_VERSION    : %s"), appFromAnsi((const ANSICHAR *)glGetString(GL_VERSION)));
	debugf(NAME_Init, TEXT("GLSL_VERSION  : %s"), appFromAnsi((const ANSICHAR *)glGetString(GL_SHADING_LANGUAGE_VERSION)));

	//Get OpenGL 3.3 function pointers
	{
		bool loadRet;

		loadRet = GetGL3Procs();
		if (!loadRet) {
			appErrorf(TEXT("Failed to get OpenGL 3.3 function pointer."));
		}
	}

	//Get OpenGL extension function pointers
	GetGLExtProcs();

	debugf(NAME_Init, TEXT("Depth bits: %u"), m_numDepthBits);
	if (DebugBit(DEBUG_BIT_BASIC)) dbgPrintf("utglr: Depth bits: %u\n", m_numDepthBits);
	if (m_usingAA) {
		debugf(NAME_Init, TEXT("AA samples: %u"), m_initNumAASamples);
		if (DebugBit(DEBUG_BIT_BASIC)) dbgPrintf("utglr: AA samples: %u\n", m_initNumAASamples);
	}

	//Get other defaults
	if (!GConfig->GetInt(g_pSection, TEXT("Brightness"), Brightness)) Brightness = 0;

	//Debug parameter listing
	if (DebugBit(DEBUG_BIT_BASIC)) {
		#define UTGLR_DEBUG_SHOW_PARAM_REG(name) DbgPrintInitParam(#name, name)
		#define UTGLR_DEBUG_SHOW_PARAM_DCV(name) DbgPrintInitParam(#name, DCV.name)

		UTGLR_DEBUG_SHOW_PARAM_REG(LODBias);
		UTGLR_DEBUG_SHOW_PARAM_REG(GammaOffset);
		UTGLR_DEBUG_SHOW_PARAM_REG(GammaOffsetRed);
		UTGLR_DEBUG_SHOW_PARAM_REG(GammaOffsetGreen);
		UTGLR_DEBUG_SHOW_PARAM_REG(GammaOffsetBlue);
		UTGLR_DEBUG_SHOW_PARAM_REG(Brightness);
		UTGLR_DEBUG_SHOW_PARAM_REG(GammaCorrectScreenshots);
		UTGLR_DEBUG_SHOW_PARAM_REG(OneXBlending);
		UTGLR_DEBUG_SHOW_PARAM_REG(MinLogTextureSize);
		UTGLR_DEBUG_SHOW_PARAM_REG(MaxLogTextureSize);
		UTGLR_DEBUG_SHOW_PARAM_REG(MaxAnisotropy);
		UTGLR_DEBUG_SHOW_PARAM_REG(TMUnits);
		UTGLR_DEBUG_SHOW_PARAM_REG(MaxTMUnits);
		UTGLR_DEBUG_SHOW_PARAM_REG(RefreshRate);
		UTGLR_DEBUG_SHOW_PARAM_REG(ShareLists);
		UTGLR_DEBUG_SHOW_PARAM_REG(UsePrecache);
		UTGLR_DEBUG_SHOW_PARAM_REG(UseTrilinear);
		UTGLR_DEBUG_SHOW_PARAM_REG(UseS3TC);
		UTGLR_DEBUG_SHOW_PARAM_REG(Use16BitTextures);
		UTGLR_DEBUG_SHOW_PARAM_REG(NoFiltering);
		UTGLR_DEBUG_SHOW_PARAM_REG(DetailMax);
		UTGLR_DEBUG_SHOW_PARAM_REG(ColorizeDetailTextures);
		UTGLR_DEBUG_SHOW_PARAM_REG(SinglePassFog);
		UTGLR_DEBUG_SHOW_PARAM_DCV(SinglePassDetail);
		UTGLR_DEBUG_SHOW_PARAM_REG(UseSSE);
		UTGLR_DEBUG_SHOW_PARAM_REG(UseSSE2);
		UTGLR_DEBUG_SHOW_PARAM_REG(UseTexIdPool);
		UTGLR_DEBUG_SHOW_PARAM_REG(UseTexPool);
		UTGLR_DEBUG_SHOW_PARAM_REG(CacheStaticMaps);
		UTGLR_DEBUG_SHOW_PARAM_REG(DynamicTexIdRecycleLevel);
		UTGLR_DEBUG_SHOW_PARAM_REG(TexDXT1ToDXT3);
		UTGLR_DEBUG_SHOW_PARAM_REG(SwapInterval);
		UTGLR_DEBUG_SHOW_PARAM_REG(FrameRateLimit);
		UTGLR_DEBUG_SHOW_PARAM_REG(SceneNodeHack);
		UTGLR_DEBUG_SHOW_PARAM_REG(SmoothMaskedTextures);
		UTGLR_DEBUG_SHOW_PARAM_REG(MaskedTextureHack);
		UTGLR_DEBUG_SHOW_PARAM_REG(UseAA);
		UTGLR_DEBUG_SHOW_PARAM_REG(NumAASamples);
		UTGLR_DEBUG_SHOW_PARAM_REG(NoAATiles);
		UTGLR_DEBUG_SHOW_PARAM_REG(ZRangeHack);

		#undef UTGLR_DEBUG_SHOW_PARAM_REG
		#undef UTGLR_DEBUG_SHOW_PARAM_DCV
	}

	//Special handling for WGL_EXT_swap_control
	//Restricted to a maximum of 10
	if ((SwapInterval >= 0) && (SwapInterval <= 10)) {
		const char *pWGLExt;

		//Get WGL extensions string
		pWGLExt = initProcs.p_wglGetExtensionsStringARB(m_hDC);

		//See if WGL_EXT_swap_control is supported
		if (IsGLExtensionSupported(pWGLExt, "WGL_EXT_swap_control")) {
			//Get the extension function pointer and use it if successful
			PFNWGLSWAPINTERVALEXTPROC p_wglSwapIntervalEXT = reinterpret_cast<PFNWGLSWAPINTERVALEXTPROC>(wglGetProcAddress("wglSwapIntervalEXT"));
			if (p_wglSwapIntervalEXT != NULL) {
				//Set the swap interval
				p_wglSwapIntervalEXT(SwapInterval);
			}
		}
	}


#ifdef UTGLR_INCLUDE_SSE_CODE
	if (UseSSE) {
		if (!CPU_DetectSSE()) {
			UseSSE = 0;
		}
	}
	if (UseSSE2) {
		if (!CPU_DetectSSE2()) {
			UseSSE2 = 0;
		}
	}
#else
	UseSSE = 0;
	UseSSE2 = 0;
#endif
	if (DebugBit(DEBUG_BIT_BASIC)) dbgPrintf("utglr: UseSSE = %u\n", UseSSE);
	if (DebugBit(DEBUG_BIT_BASIC)) dbgPrintf("utglr: UseSSE2 = %u\n", UseSSE2);

	SetGamma(Viewport->GetOuterUClient()->Brightness);

	//Restrict dynamic tex id recycle level range
	if (DynamicTexIdRecycleLevel < 10) DynamicTexIdRecycleLevel = 10;

	SupportsTC = UseS3TC;


	// Validate flags.

	//Special extensions validation for init only config pass
	if (!SUPPORTS_GL_EXT_texture_compression_s3tc) SupportsTC = 0;

	//DCV refresh
	ConfigValidate_RefreshDCV();

	//Required extensions config validation pass
	ConfigValidate_RequiredExtensions();


	if (!MaxTMUnits || (MaxTMUnits > MAX_TMUNITS)) {
		MaxTMUnits = MAX_TMUNITS;
	}

	glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS, &TMUnits);
	debugf(TEXT("Texture image units: %i"), TMUnits);
	if (TMUnits > MaxTMUnits) {
		TMUnits = MaxTMUnits;
	}


	//Main config validation pass (after set TMUnits)
	ConfigValidate_Main();


	//Allocate vertex buffer names
	m_allocatedVertexBufferNames = false;
	AllocateVertexBufferNamesSafe();

	//Allocate vertex buffers
	if (!InitializeVertexBuffers()) {
		appErrorf(TEXT("Failed to allocate vertex buffers."));
	}


	//Configure vertex buffers

	//Reset vertex buffer state tracking variables
	InitVertexBufferState();


	//Configure vertex arrays

	//Vertex array with vertices and color
	SetVertexArray(m_vaOneColor);
	{
		//Vertex and primary color
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(FGLVertexColor), 0);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 4, GL_UNSIGNED_BYTE, GL_TRUE, sizeof(FGLVertexColor), (const GLvoid *)offsetof(FGLVertexColor, color));
	}

	//Vertex array with vertices, color, and one texcoord
	SetVertexArray(m_vaOneColorOneTexcoord);
	{
		//Vertex and primary color
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(FGLVertexColor), 0);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 4, GL_UNSIGNED_BYTE, GL_TRUE, sizeof(FGLVertexColor), (const GLvoid *)offsetof(FGLVertexColor, color));

		//TexCoord arrays
		glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
		glEnableVertexAttribArray(4);
		glVertexAttribPointer(4, 2, GL_FLOAT, GL_FALSE, sizeof(FGLTexCoord), 0);
	}

	//Vertex array with vertices, two colors, and one tex coord
	SetVertexArray(m_vaTwoColorOneTexcoord);
	{
		//Vertex and primary color
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(FGLVertexColor), 0);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 4, GL_UNSIGNED_BYTE, GL_TRUE, sizeof(FGLVertexColor), (const GLvoid *)offsetof(FGLVertexColor, color));

		//Secondary color
		glBindBuffer(GL_ARRAY_BUFFER, m_vbSecondaryColor);
		glEnableVertexAttribArray(2);
		glVertexAttribPointer(2, 4, GL_UNSIGNED_BYTE, GL_TRUE, sizeof(FGLSecondaryColor), 0);

		//TexCoord arrays
		glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
		glEnableVertexAttribArray(4);
		glVertexAttribPointer(4, 2, GL_FLOAT, GL_FALSE, sizeof(FGLTexCoord), 0);
	}

	//Set to vertex array 0
	SetVertexArray(0);

	//Bind array buffer 0
	glBindBuffer(GL_ARRAY_BUFFER, 0);


	//Reset fragment program state tracking variables
	m_prCurrent = 0;

	//Mark shader names as not allocated
	m_allocatedShaderNames = false;
	//Setup shaders
	TryInitializeShaders();


	if (MaxAnisotropy < 0) {
		MaxAnisotropy = 0;
	}
	if (MaxAnisotropy) {
		GLint iMaxAnisotropyLimit = 1;
		glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &iMaxAnisotropyLimit);
		debugf(TEXT("MaxAnisotropy: %i"), iMaxAnisotropyLimit);
		if (MaxAnisotropy > iMaxAnisotropyLimit) {
			MaxAnisotropy = iMaxAnisotropyLimit;
		}
	}

	if (SupportsTC) {
		debugf(TEXT("Trying to use S3TC extension."));
	}

	//Use default if MaxLogTextureSize <= 0
	if (MaxLogTextureSize <= 0) MaxLogTextureSize = 12;

	INT MaxTextureSize;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &MaxTextureSize);
	INT Dummy = -1;
	while (MaxTextureSize > 0) {
		MaxTextureSize >>= 1;
		Dummy++;
	}

	if ((MaxLogTextureSize > Dummy) || (SupportsTC)) MaxLogTextureSize = Dummy;
	if ((MinLogTextureSize < 2) || (SupportsTC)) MinLogTextureSize = 2;

	MaxLogUOverV = MaxLogTextureSize;
	MaxLogVOverU = MaxLogTextureSize;

	debugf(TEXT("MinLogTextureSize: %i"), MinLogTextureSize);
	debugf(TEXT("MaxLogTextureSize: %i"), MaxLogTextureSize);


	// Verify hardware defaults.
	check(MinLogTextureSize >= 0);
	check(MaxLogTextureSize >= 0);
	check(MinLogTextureSize <= MaxLogTextureSize);

	// Flush textures.
	Flush(1);

	//Invalidate fixed texture ids
	m_noTextureId = 0;

	// Bind little white RGBA texture to ID 0.
	InitNoTextureSafe();

	//Clear fsBlendInfo to set any unused values to known state
	m_fsBlendInfo[0] = 0.0f;
	m_fsBlendInfo[1] = 0.0f;
	m_fsBlendInfo[2] = 0.0f;
	m_fsBlendInfo[3] = 0.0f;

	//Set alpha test disabled
	m_fsBlendInfo[0] = -1.0f;

	//Set base texture LOD bias
	m_fsBlendInfo[1] = LODBias;

	// Set permanent state.
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_LEQUAL);
	glPolygonOffset(-1.0f, -1.0f);
	glBlendFunc(GL_ONE, GL_ZERO);
	glDisable(GL_BLEND);
	glEnable(GL_DITHER);

	m_gpFogEnabled = false;


	// Init variables.
	m_bufferedVertsType = BV_TYPE_NONE;
	m_bufferedVerts = 0;

	m_curBlendFlags = PF_Occlude;
	m_smoothMaskedTexturesBit = 0;
	m_useAlphaToCoverageForMasked = false;
	m_alphaToCoverageEnabled = false;
	m_curPolyFlags = 0;
	m_curPolyFlags2 = 0;

	//Initialize color flags
	m_requestedColorFlags = 0;

	//Initialize Z range hack state
	m_useZRangeHack = false;
	m_nearZRangeHackProjectionActive = false;
	m_requestNearZRangeHackProjection = false;


	//Initialize previous lock variables
	PL_DetailTextures = DetailTextures;
	PL_OneXBlending = OneXBlending;
	PL_MaxLogUOverV = MaxLogUOverV;
	PL_MaxLogVOverU = MaxLogVOverU;
	PL_MinLogTextureSize = MinLogTextureSize;
	PL_MaxLogTextureSize = MaxLogTextureSize;
	PL_NoFiltering = NoFiltering;
	PL_UseTrilinear = UseTrilinear;
	PL_Use16BitTextures = Use16BitTextures;
	PL_TexDXT1ToDXT3 = TexDXT1ToDXT3;
	PL_MaxAnisotropy = MaxAnisotropy;
	PL_SmoothMaskedTextures = SmoothMaskedTextures;
	PL_MaskedTextureHack = MaskedTextureHack;
	PL_LODBias = LODBias;
	PL_SinglePassDetail = SinglePassDetail;
	PL_UseSSE = UseSSE;
	PL_UseSSE2 = UseSSE2;


	//Reset current frame count
	m_currentFrameCount = 0;

	// Remember fullscreenness.
	WasFullscreen = Fullscreen;

	return 1;

	unguard;
}

void UOpenGLRenderDevice::UnsetRes() {
	guard(UOpenGLRenderDevice::UnsetRes);

#ifdef _WIN32
	check(m_hRC)
#endif

	//Flush textures
	Flush(1);

	//Free fixed textures if they were allocated
	if (m_noTextureId != 0) {
		glDeleteTextures(1, &m_noTextureId);
		m_noTextureId = 0;
	}

	//Free vertex buffers if they were allocated
	FreeVertexBufferNamesSafe();

	//Free shaders if they were allocated and leave shader mode if necessary
	ShutdownShaderMode();

#ifdef _WIN32
	hCurrentRC = NULL;
	wglMakeCurrent(NULL, NULL);
//	verify(wglDeleteContext(m_hRC));
	wglDeleteContext(m_hRC);
	verify(AllContexts.RemoveItem(m_hRC) == 1);
	m_hRC = NULL;
	if (WasFullscreen) {
		TCHAR_CALL_OS(ChangeDisplaySettingsW(NULL, 0), ChangeDisplaySettingsA(NULL, 0));
	}
#endif

	unguard;
}


void UOpenGLRenderDevice::MakeCurrent(void) {
	guard(UOpenGLRenderDevice::MakeCurrent);
#ifdef _WIN32
	check(m_hRC);
	check(m_hDC);
	if (hCurrentRC != m_hRC) {
		verify(wglMakeCurrent(m_hDC, m_hRC));
		hCurrentRC = m_hRC;
	}
#endif
	unguard;
}

void UOpenGLRenderDevice::CheckGLErrorFlag(const TCHAR *pTag) {
	GLenum Error = glGetError();
	if ((Error != GL_NO_ERROR) && DebugBit(DEBUG_BIT_GL_ERROR)) {
		const TCHAR *pMsg;
		switch (Error) {
		case GL_INVALID_ENUM:
			pMsg = TEXT("GL_INVALID_ENUM");
			break;
		case GL_INVALID_VALUE:
			pMsg = TEXT("GL_INVALID_VALUE");
			break;
		case GL_INVALID_OPERATION:
			pMsg = TEXT("GL_INVALID_OPERATION");
			break;
		case GL_STACK_OVERFLOW:
			pMsg = TEXT("GL_STACK_OVERFLOW");
			break;
		case GL_STACK_UNDERFLOW:
			pMsg = TEXT("GL_STACK_UNDERFLOW");
			break;
		case GL_OUT_OF_MEMORY:
			pMsg = TEXT("GL_OUT_OF_MEMORY");
			break;
		default:
			pMsg = TEXT("UNKNOWN");
		}
		//appErrorf(TEXT("OpenGL Error: %s (%s)"), pMsg, pTag);
		debugf(TEXT("OpenGL Error: %s (%s)"), pMsg, pTag);
	}
}


void UOpenGLRenderDevice::ConfigValidate_RefreshDCV(void) {
	#define UTGLR_REFRESH_DCV(_name) _name = DCV._name

	UTGLR_REFRESH_DCV(SinglePassDetail);

	#undef UTGLR_REFRESH_DCV

	return;
}

void UOpenGLRenderDevice::ConfigValidate_RequiredExtensions(void) {
	if (!SUPPORTS_GL_EXT_texture_filter_anisotropic) MaxAnisotropy  = 0;

	return;
}

void UOpenGLRenderDevice::ConfigValidate_Main(void) {
	//Single pass detail texturing requires at least 4 texture units
	if (TMUnits < 4) SinglePassDetail = 0;

	//Limit maximum DetailMax
	if (DetailMax > 3) DetailMax = 3;

	return;
}


#ifdef _WIN32
void UOpenGLRenderDevice::SetNewPixelFormat(INT NewColorBytes, HGLRC hShareGLRC, new_gl_init_procs_t *pInitProcs) {
	BOOL bRet;
	bool havePixelFormat;
	int iFormats[1];
	int iAttributes[30];
	UINT nNumFormats;
	PIXELFORMATDESCRIPTOR tempPfd;

	//Set using AA to default of not using initially
	//Remember number of AA samples, if AA init successful
	m_usingAA = false;
	m_initNumAASamples = NumAASamples;

	if (DebugBit(DEBUG_BIT_BASIC)) dbgPrintf("utglr: SetNewPixelFormat\n");

	iAttributes[0]  = WGL_SUPPORT_OPENGL_ARB;
	iAttributes[1]  = GL_TRUE;
	iAttributes[2]  = WGL_DRAW_TO_WINDOW_ARB;
	iAttributes[3]  = GL_TRUE;
	iAttributes[4]  = WGL_ACCELERATION_ARB;
	iAttributes[5]  = WGL_FULL_ACCELERATION_ARB;
	iAttributes[6]  = WGL_COLOR_BITS_ARB;
	iAttributes[7]  = (NewColorBytes <= 2) ? 16 : 24;
	iAttributes[8]  = WGL_DEPTH_BITS_ARB;
	iAttributes[9]  = 32;
	iAttributes[10] = WGL_DOUBLE_BUFFER_ARB;
	iAttributes[11] = GL_TRUE;
	iAttributes[12] = WGL_SAMPLE_BUFFERS_ARB;
	iAttributes[13] = GL_TRUE;
	iAttributes[14] = WGL_SAMPLES_ARB;
	iAttributes[15] = NumAASamples;
	iAttributes[16] = 0;
	iAttributes[17] = 0;

	//Try to get an AA pixel format if AA enabled
	havePixelFormat = false;
	if (UseAA) {
		DWORD tryAACount;
		DWORD tryNumAASamples;

		//Try to choose pixel format with AA
		tryAACount = 0;
		tryNumAASamples = (NumAASamples < 1) ? 1 : NumAASamples;
		do {
			//Select how many AA samples to try to get
			if (tryAACount == 0) {
				//Use initial selection
			}
			else if (tryAACount < 3) {
				tryNumAASamples /= 2;
			}
			else {
				tryNumAASamples = 4;
			}
			//Set WGL_SAMPLES_ARB value
			iAttributes[15] = tryNumAASamples;

			//Try 32b, 24b, 16b depth buffer
			iAttributes[9] = 32;
			bRet = pInitProcs->p_wglChoosePixelFormatARB(m_hDC, iAttributes, NULL, 1, iFormats, &nNumFormats);
			if ((bRet == FALSE) || (nNumFormats == 0)) {
				iAttributes[9] = 24;
				bRet = pInitProcs->p_wglChoosePixelFormatARB(m_hDC, iAttributes, NULL, 1, iFormats, &nNumFormats);
			}
			if ((bRet == FALSE) || (nNumFormats == 0)) {
				iAttributes[9] = 16;
				bRet = pInitProcs->p_wglChoosePixelFormatARB(m_hDC, iAttributes, NULL, 1, iFormats, &nNumFormats);
			}
			if ((bRet == FALSE) || (nNumFormats == 0)) {
				if (DebugBit(DEBUG_BIT_BASIC)) dbgPrintf("utglr: Choose AA pixel format failed\n");
			}
			else {
				havePixelFormat = true;
				m_usingAA = true;
				break;
			}
		} while (++tryAACount < 4);
	}
	//Try to get a pixel format without AA if AA init failed or AA not enabled
	if (!havePixelFormat) {
		//Set WGL_SAMPLE_BUFFERS_ARB to GL_FALSE
		iAttributes[13] = GL_FALSE;
		//Set WGL_SAMPLES_ARB to 0
		iAttributes[15] = 0;

		iAttributes[9] = 32;
		bRet = pInitProcs->p_wglChoosePixelFormatARB(m_hDC, iAttributes, NULL, 1, iFormats, &nNumFormats);
		if ((bRet == FALSE) || (nNumFormats == 0)) {
			iAttributes[9] = 24;
			bRet = pInitProcs->p_wglChoosePixelFormatARB(m_hDC, iAttributes, NULL, 1, iFormats, &nNumFormats);
		}
		if ((bRet == FALSE) || (nNumFormats == 0)) {
			iAttributes[9] = 16;
			bRet = pInitProcs->p_wglChoosePixelFormatARB(m_hDC, iAttributes, NULL, 1, iFormats, &nNumFormats);
		}
		if ((bRet == FALSE) || (nNumFormats == 0)) {
			if (DebugBit(DEBUG_BIT_BASIC)) dbgPrintf("utglr: Choose pixel format failed\n");
		}
		else {
			havePixelFormat = true;
		}
	}
	if (!havePixelFormat) {
		appErrorf(TEXT("Failed to choose a usable pixel format."));
	}

	//Set pixel format
	appMemzero(&tempPfd, sizeof(tempPfd));
	tempPfd.nSize = sizeof(tempPfd);
	debugf(NAME_Init, TEXT("Using pixel format %i"), iFormats[0]);
	verify(SetPixelFormat(m_hDC, iFormats[0], &tempPfd));

	//Create GL3 context
	const int iAttribs[] = {
		WGL_CONTEXT_MAJOR_VERSION_ARB, 3,
		WGL_CONTEXT_MINOR_VERSION_ARB, 3,
#if 1
		WGL_CONTEXT_FLAGS_ARB, WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB,
		WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_CORE_PROFILE_BIT_ARB,
#else
		WGL_CONTEXT_FLAGS_ARB, 0x0,
		WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB,
#endif
		0
	};
	m_hRC = pInitProcs->p_wglCreateContextAttribsARB(m_hDC, hShareGLRC, iAttribs);
	if (!m_hRC) {
		appErrorf(TEXT("Failed to create OpenGL 3.3 rendering context."));
	}

	MakeCurrent();

	{
		int iAttribute = WGL_DEPTH_BITS_ARB;
		int iValue = m_numDepthBits;

		//Get actual number of depth bits
		iAttribute = WGL_DEPTH_BITS_ARB;
		iValue = m_numDepthBits;
		if (pInitProcs->p_wglGetPixelFormatAttribivARB(m_hDC, iFormats[0], 0, 1, &iAttribute, &iValue)) {
			m_numDepthBits = iValue;
		}

		//Get actual number of AA samples if using AA
		if (m_usingAA) {
			iAttribute = WGL_SAMPLES_ARB;
			iValue = m_initNumAASamples;
			if (pInitProcs->p_wglGetPixelFormatAttribivARB(m_hDC, iFormats[0], 0, 1, &iAttribute, &iValue)) {
				m_initNumAASamples = iValue;
			}
		}
	}

	return;
}


LRESULT CALLBACK UOpenGLRenderDevice::GetNewGLInitProcsWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch(uMsg) {
	case WM_CREATE:
		break;

	default:
		return TCHAR_CALL_OS(DefWindowProcW(hWnd, uMsg, wParam, lParam), DefWindowProcA(hWnd, uMsg, wParam, lParam));
	}

	return 0;
}

bool UOpenGLRenderDevice::GetNewGLInitProcs(INT NewColorBytes, new_gl_init_procs_t *pInitProcs) {
	guard(UOpenGLRenderDevice::GetNewGLInitProcs);

	bool bSuccess = false;
	HINSTANCE hInstance = TCHAR_CALL_OS(GetModuleHandleW(NULL), GetModuleHandleA(NULL));
	struct tagWNDCLASSA wcA;
	struct tagWNDCLASSW wcW;
	const CHAR *pClassNameA = "UOpenGLRenderDevice::GetNewGLInitProcs";
	const WCHAR *pClassNameW = L"UOpenGLRenderDevice::GetNewGLInitProcs";
	HWND hWnd;

	wcA.style = wcW.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	wcA.lpfnWndProc = wcW.lpfnWndProc = UOpenGLRenderDevice::GetNewGLInitProcsWndProc;
	wcA.cbClsExtra = wcW.cbClsExtra = 0;
	wcA.cbWndExtra = wcW.cbWndExtra = 0;
	wcA.hInstance = wcW.hInstance = hInstance;
	wcA.hIcon = wcW.hIcon = NULL;
	wcA.hCursor = wcW.hCursor = TCHAR_CALL_OS(LoadCursorW(NULL, MAKEINTRESOURCEW(32512)/*IDC_ARROW*/), LoadCursorA(NULL, MAKEINTRESOURCEA(32512)/*IDC_ARROW*/));
	wcA.hbrBackground = wcW.hbrBackground = NULL;
	wcA.lpszMenuName = NULL;
	wcW.lpszMenuName = NULL;
	wcA.lpszClassName = pClassNameA;
	wcW.lpszClassName = pClassNameW;

	if (TCHAR_CALL_OS(RegisterClassW(&wcW), RegisterClassA(&wcA)) == 0) {
		return false;
	}

	hWnd = TCHAR_CALL_OS(
		CreateWindowW(
		pClassNameW,
		pClassNameW,
		WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
		0, 0,
		20, 20,
		NULL,
		NULL,
		hInstance,
		0),
		CreateWindowA(
		pClassNameA,
		pClassNameA,
		WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
		0, 0,
		20, 20,
		NULL,
		NULL,
		hInstance,
		0)
		);
	if (hWnd == NULL) {
		TCHAR_CALL_OS(UnregisterClassW(pClassNameW, hInstance), UnregisterClassA(pClassNameA, hInstance));
		return false;
	}

	HGLRC hTempGLRC = NULL;
	do {
		HDC hDC = GetDC(hWnd);

		INT nPixelFormat;
		BYTE DesiredColorBits   = (NewColorBytes <= 2) ? 16 : 24;
		BYTE DesiredDepthBits   = 0;
		BYTE DesiredStencilBits = 0;
		PIXELFORMATDESCRIPTOR pfd = {
			sizeof(PIXELFORMATDESCRIPTOR),
			1,
			PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
			PFD_TYPE_RGBA,
			DesiredColorBits,
			0,0,0,0,0,0,
			0,0,
			0,0,0,0,0,
			DesiredDepthBits,
			DesiredStencilBits,
			0,
			PFD_MAIN_PLANE,
			0,
			0,0,0
		};

		nPixelFormat = ChoosePixelFormat(hDC, &pfd);
		if (nPixelFormat == 0) {
			break;
		}

		if (SetPixelFormat(hDC, nPixelFormat, &pfd) == FALSE) {
			break;
		}
		hTempGLRC = wglCreateContext(hDC);
		if (hTempGLRC == NULL) {
			break;
		}
		if (wglMakeCurrent(hDC, hTempGLRC) == FALSE) {
			break;
		}


		//Check for required extension support

		//Get wglGetExtensionsStringARB
		pInitProcs->p_wglGetExtensionsStringARB =
			reinterpret_cast<PFNWGLGETEXTENSIONSSTRINGARBPROC>(wglGetProcAddress("wglGetExtensionsStringARB"));
		if (!pInitProcs->p_wglGetExtensionsStringARB) {
			debugf(NAME_Init, TEXT("Failed to get wglGetExtensionsStringARB function address."));
			break;
		}

		//Get WGL extensions string
		const char *pWGLExt = pInitProcs->p_wglGetExtensionsStringARB(hDC);

		//Check for WGL_ARB_pixel_format, WGL_ARB_create_context, and WGL_ARB_create_context_profile
		if (!IsGLExtensionSupported(pWGLExt, "WGL_ARB_pixel_format")) {
			debugf(NAME_Init, TEXT("WGL_ARB_pixel_format extension not supported."));
			break;
		}
		if (!IsGLExtensionSupported(pWGLExt, "WGL_ARB_create_context")) {
			debugf(NAME_Init, TEXT("WGL_ARB_create_context extension not supported."));
			break;
		}
		if (!IsGLExtensionSupported(pWGLExt, "WGL_ARB_create_context_profile")) {
			debugf(NAME_Init, TEXT("WGL_ARB_create_context_profile extension not supported."));
			break;
		}

		//Get wglChoosePixelFormatARB
		pInitProcs->p_wglChoosePixelFormatARB =
			reinterpret_cast<PFNWGLCHOOSEPIXELFORMATARBPROC>(wglGetProcAddress("wglChoosePixelFormatARB"));
		if (!pInitProcs->p_wglChoosePixelFormatARB) {
			debugf(NAME_Init, TEXT("Failed to get wglChoosePixelFormatARB function address."));
			break;
		}

		//Get wglGetPixelFormatAttribivARB
		pInitProcs->p_wglGetPixelFormatAttribivARB =
			reinterpret_cast<PFNWGLGETPIXELFORMATATTRIBIVARBPROC>(wglGetProcAddress("wglGetPixelFormatAttribivARB"));
		if (!pInitProcs->p_wglGetPixelFormatAttribivARB) {
			debugf(NAME_Init, TEXT("Failed to get wglGetPixelFormatAttribivARB function address."));
			break;
		}

		//Get wglCreateContextAttribsARB
		pInitProcs->p_wglCreateContextAttribsARB =
			reinterpret_cast<PFNWGLCREATECONTEXTATTRIBSARBPROC>(wglGetProcAddress("wglCreateContextAttribsARB"));
		if (!pInitProcs->p_wglCreateContextAttribsARB) {
			debugf(NAME_Init, TEXT("Failed to get wglCreateContextAttribsARB function address."));
			break;
		}

		//Successful if get to here
		bSuccess = true;
	} while (0);

	//Delete temporary context if it was created
	if (hTempGLRC) {
		//Delete temporary context
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(hTempGLRC);
		hTempGLRC = NULL;
	}

	DestroyWindow(hWnd);

	TCHAR_CALL_OS(UnregisterClassW(pClassNameW, hInstance), UnregisterClassA(pClassNameA, hInstance));

	return bSuccess;

	unguard;
}
#endif


UBOOL UOpenGLRenderDevice::Init(UViewport* InViewport, INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen) {
	guard(UOpenGLRenderDevice::Init);

	debugf(TEXT("Initializing OpenGL3Drv..."));

	if (NumDevices == 0) {
		g_gammaFirstTime = true;
		g_haveOriginalGammaRamp = false;
	}

#ifdef WIN32
	// Get list of device modes.
	for (INT i = 0; ; i++) {
		UBOOL UnicodeOS;

#if defined(NO_UNICODE_OS_SUPPORT) || !defined(UNICODE)
		UnicodeOS = 0;
#elif defined(NO_ANSI_OS_SUPPORT)
		UnicodeOS = 1;
#else
		UnicodeOS = GUnicodeOS;
#endif

		if (!UnicodeOS) {
#if defined(NO_UNICODE_OS_SUPPORT) || !defined(UNICODE) || !defined(NO_ANSI_OS_SUPPORT)
			DEVMODEA Tmp;
			appMemzero(&Tmp, sizeof(Tmp));
			Tmp.dmSize = sizeof(Tmp);
			if (!EnumDisplaySettingsA(NULL, i, &Tmp)) {
				break;
			}
			Modes.AddUniqueItem(FPlane(Tmp.dmPelsWidth, Tmp.dmPelsHeight, Tmp.dmBitsPerPel, Tmp.dmDisplayFrequency));
#endif
		}
		else {
#if !defined(NO_UNICODE_OS_SUPPORT) && defined(UNICODE)
			DEVMODEW Tmp;
			appMemzero(&Tmp, sizeof(Tmp));
			Tmp.dmSize = sizeof(Tmp);
			if (!EnumDisplaySettingsW(NULL, i, &Tmp)) {
				break;
			}
			Modes.AddUniqueItem(FPlane(Tmp.dmPelsWidth, Tmp.dmPelsHeight, Tmp.dmBitsPerPel, Tmp.dmDisplayFrequency));
#endif
		}
	}

	// Init global GL.
	if (NumDevices == 0) {
		bool loadRet;

		// Find DLL's.
		hModuleGlMain = LoadLibraryA(GL_DLL);
		if (!hModuleGlMain) {
			debugf(NAME_Init, LocalizeError("NoFindGL"), appFromAnsi(GL_DLL));
			return 0;
		}

		//Get OpenGL 1.x function pointers
		loadRet = GetGL1Procs();
		if (!loadRet) {
			return 0;
		}
	}
#endif

	NumDevices++;

	// Init this GL rendering context.
	m_zeroPrefixBindTrees = ShareLists ? m_sharedZeroPrefixBindTrees : m_localZeroPrefixBindTrees;
	m_nonZeroPrefixBindTrees = ShareLists ? m_sharedNonZeroPrefixBindTrees : m_localNonZeroPrefixBindTrees;
	m_nonZeroPrefixBindChain = ShareLists ? &m_sharedNonZeroPrefixBindChain : &m_localNonZeroPrefixBindChain;
	m_nonZeroPrefixTexIdPool = ShareLists ? &m_sharedNonZeroPrefixTexIdPool : &m_localNonZeroPrefixTexIdPool;
	m_RGBA8TexPool = ShareLists ? &m_sharedRGBA8TexPool : &m_localRGBA8TexPool;

	Viewport = InViewport;


	//Remember main window handle and get its DC
	m_hWnd = (HWND)InViewport->GetWindow();
	check(m_hWnd);
	m_hDC = GetDC(m_hWnd);
	check(m_hDC);


#if 0
	{
		//Print all PFD's exposed
		INT pf;
		INT pfCount = DescribePixelFormat(m_hDC, 0, 0, NULL);
		for (pf = 1; pf <= pfCount; pf++) {
			PrintFormat(m_hDC, pf);
		}
	}
#endif

	if (!SetRes(NewX, NewY, NewColorBytes, Fullscreen)) {
		return FailedInitf(LocalizeError("ResFailed"));
	}

	return 1;
	unguard;
}

UBOOL UOpenGLRenderDevice::Exec(const TCHAR* Cmd, FOutputDevice& Ar) {
	guard(UOpenGLRenderDevice::Exec);

	if (URenderDevice::Exec(Cmd, Ar)) {
		return 1;
	}
	if (ParseCommand(&Cmd, TEXT("DGL"))) {
		if (ParseCommand(&Cmd, TEXT("BUILD"))) {
			debugf(TEXT("OpenGL3 renderer built: ?????"));
			return 1;
		}
		else if (ParseCommand(&Cmd, TEXT("AA"))) {
			if (m_usingAA) {
				m_defAAEnable = !m_defAAEnable;
				debugf(TEXT("AA Enable [%u]"), (m_defAAEnable) ? 1 : 0);
			}
			return 1;
		}

		return 0;
	}
	else if (ParseCommand(&Cmd, TEXT("GetRes"))) {
		TArray<FPlane> Relevant;
		INT i;
		for (i = 0; i < Modes.Num(); i++) {
			if (Modes(i).Z == (Viewport->ColorBytes * 8))
				if
				(	(Modes(i).X!=320 || Modes(i).Y!=200)
				&&	(Modes(i).X!=640 || Modes(i).Y!=400) )
				Relevant.AddUniqueItem(FPlane(Modes(i).X, Modes(i).Y, 0, 0));
		}
		appQsort(&Relevant(0), Relevant.Num(), sizeof(FPlane), (QSORT_COMPARE)CompareRes);
		FString Str;
		for (i = 0; i < Relevant.Num(); i++) {
			Str += FString::Printf(TEXT("%ix%i "), (INT)Relevant(i).X, (INT)Relevant(i).Y);
		}
		Ar.Log(*Str.LeftChop(1));
		return 1;
	}

	return 0;

	unguard;
}

void UOpenGLRenderDevice::Lock(FPlane InFlashScale, FPlane InFlashFog, FPlane ScreenClear, DWORD RenderLockFlags, BYTE* InHitData, INT* InHitSize) {
	UTGLR_DEBUG_CALL_COUNT(Lock);
	guard(UOpenGLRenderDevice::Lock);
	check(LockCount == 0);
	++LockCount;


	//Reset stats
	BindCycles = ImageCycles = ComplexCycles = GouraudCycles = TileCycles = 0;

	m_prSwitchCount = 0;
	m_AASwitchCount = 0;
	m_sceneNodeCount = 0;
	m_sceneNodeHackCount = 0;
	m_vbFlushCount = 0;
	m_stat0Count = 0;
	m_stat1Count = 0;


	// Make this context current.
	MakeCurrent();

	//Clear the Z-buffer
	if (1) {
		glClearColor(ScreenClear.X, ScreenClear.Y, ScreenClear.Z, ScreenClear.W);
		glClearDepth(1.0);
		glDepthRange(0.0, 1.0);
		glPolygonOffset(-1.0f, -1.0f);
		SetBlend(PF_Occlude);
		glClear(GL_DEPTH_BUFFER_BIT | ((RenderLockFlags & LOCKR_ClearScreen) ? GL_COLOR_BUFFER_BIT : 0));
	}
	glDepthFunc(GL_LEQUAL);


	//Scan for textures that may have been deleted by another context
	ScanForDeletedTextures();


	bool flushTextures = false;
	bool needShaderReload = false;


	//DCV refresh
	ConfigValidate_RefreshDCV();

	//Required extensions config validation pass
	ConfigValidate_RequiredExtensions();


	//Main config validation pass
	ConfigValidate_Main();


	//Prevent changes to these parameters
	MaxLogUOverV = PL_MaxLogUOverV;
	MaxLogVOverU = PL_MaxLogVOverU;
	MinLogTextureSize = PL_MinLogTextureSize;
	MaxLogTextureSize = PL_MaxLogTextureSize;

	//Detect changes in various texture related options and force texture flush if necessary
	if (NoFiltering != PL_NoFiltering) {
		PL_NoFiltering = NoFiltering;
		flushTextures = true;
	}
	if (UseTrilinear != PL_UseTrilinear) {
		PL_UseTrilinear = UseTrilinear;
		flushTextures = true;
	}
	if (Use16BitTextures != PL_Use16BitTextures) {
		PL_Use16BitTextures = Use16BitTextures;
		flushTextures = true;
	}
	if (TexDXT1ToDXT3 != PL_TexDXT1ToDXT3) {
		PL_TexDXT1ToDXT3 = TexDXT1ToDXT3;
		flushTextures = true;
	}
	//MaxAnisotropy cannot be negative
	if (MaxAnisotropy < 0) {
		MaxAnisotropy = 0;
	}
	if (MaxAnisotropy != PL_MaxAnisotropy) {
		if (MaxAnisotropy) {
			GLint iMaxAnisotropyLimit = 1;
			glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &iMaxAnisotropyLimit);
			if (MaxAnisotropy > iMaxAnisotropyLimit) {
				MaxAnisotropy = iMaxAnisotropyLimit;
			}
		}

		PL_MaxAnisotropy = MaxAnisotropy;
		flushTextures = true;
	}

	if (SmoothMaskedTextures != PL_SmoothMaskedTextures) {
		PL_SmoothMaskedTextures = SmoothMaskedTextures;

		//Clear masked blending state if set before adjusting smooth masked textures bit
		SetBlend(PF_Occlude);

		//Disable alpha to coverage if currently enabled
		if (m_alphaToCoverageEnabled) {
			m_alphaToCoverageEnabled = false;
			glDisable(GL_SAMPLE_ALPHA_TO_COVERAGE);
		}
	}

	//Smooth masked textures bit controls alpha blend for masked textures
	m_smoothMaskedTexturesBit = 0;
	m_useAlphaToCoverageForMasked = false;
	if (SmoothMaskedTextures) {
		//Use alpha to coverage if using AA and enough samples
		if (m_usingAA && (m_initNumAASamples >= 4)) {
			m_useAlphaToCoverageForMasked = true;
		}
		else {
			m_smoothMaskedTexturesBit = PF_Masked;
		}
	}

	if (MaskedTextureHack != PL_MaskedTextureHack) {
		PL_MaskedTextureHack = MaskedTextureHack;
		flushTextures = true;
	}

	if (LODBias != PL_LODBias) {
		PL_LODBias = LODBias;

		//Set base texture LOD bias
		m_fsBlendInfo[1] = LODBias;
	}

	if (DetailTextures != PL_DetailTextures) {
		PL_DetailTextures = DetailTextures;
		flushTextures = true;
		if (DetailTextures) {
			needShaderReload = true;
		}
	}

	if (SinglePassDetail != PL_SinglePassDetail) {
		PL_SinglePassDetail = SinglePassDetail;
		if (SinglePassDetail) {
			needShaderReload = true;
		}
	}


	//Check if fragment program reload is necessary
	if (needShaderReload) {
		//Attempt to initialize shaders
		TryInitializeShaders();
	}


	if (UseSSE != PL_UseSSE) {
#ifdef UTGLR_INCLUDE_SSE_CODE
		if (UseSSE) {
			if (!CPU_DetectSSE()) {
				UseSSE = 0;
			}
		}
#else
		UseSSE = 0;
#endif
		PL_UseSSE = UseSSE;
	}
	if (UseSSE2 != PL_UseSSE2) {
#ifdef UTGLR_INCLUDE_SSE_CODE
		if (UseSSE2) {
			if (!CPU_DetectSSE2()) {
				UseSSE2 = 0;
			}
		}
#else
		UseSSE2 = 0;
#endif
		PL_UseSSE2 = UseSSE2;
	}


	//Shared fragment program parameters
	if (1) {
		//Lightmap blend scale factor
		m_fsBlendInfo[3] = (OneXBlending) ? 1.0f : 2.0f;

		//Update shared uniform with fragment info
		glBindBuffer(GL_UNIFORM_BUFFER, m_ubFragInfo);
		glBufferSubData(GL_UNIFORM_BUFFER, 0, 16, m_fsBlendInfo);
	}


	//Initialize buffer verts proc pointers
	m_pBuffer3BasicVertsProc = Buffer3BasicVerts;
	m_pBuffer3ColoredVertsProc = Buffer3ColoredVerts;
	m_pBuffer3FoggedVertsProc = Buffer3FoggedVerts;

#ifdef UTGLR_INCLUDE_SSE_CODE
	//Initialize SSE buffer verts proc pointers
	if (UseSSE) {
		m_pBuffer3ColoredVertsProc = Buffer3ColoredVerts_SSE;
		m_pBuffer3FoggedVertsProc = Buffer3FoggedVerts_SSE;
	}
	if (UseSSE2) {
		m_pBuffer3ColoredVertsProc = Buffer3ColoredVerts_SSE2;
		m_pBuffer3FoggedVertsProc = Buffer3FoggedVerts_SSE2;
	}
#endif //UTGLR_INCLUDE_SSE_CODE

	m_pBuffer3VertsProc = NULL;


	//Initialize buffer detail texture data proc pointer
	m_pBufferDetailTextureDataProc = &UOpenGLRenderDevice::BufferDetailTextureData;
#ifdef UTGLR_INCLUDE_SSE_CODE
	if (UseSSE2) {
		m_pBufferDetailTextureDataProc = &UOpenGLRenderDevice::BufferDetailTextureData_SSE2;
	}
#endif //UTGLR_INCLUDE_SSE_CODE


	//Precalculate detail texture colors
	if (ColorizeDetailTextures) {
		m_detailTextureColor4ub = 0x00408040;
	}
	else {
		m_detailTextureColor4ub = 0x00808080;
	}

	//Precalculate mask for MaskedTextureHack based on if it's enabled
	m_maskedTextureHackMask = (MaskedTextureHack) ? TEX_CACHE_ID_FLAG_MASKED : 0;

	// Remember stuff.
	FlashScale = InFlashScale;
	FlashFog   = InFlashFog;

	//Selection setup
	m_HitData = InHitData;
	m_HitSize = InHitSize;
	m_HitCount = 0;
	if (m_HitData) {
		m_HitBufSize = *m_HitSize;
		*m_HitSize = 0;

		//Start selection
		m_gclip.SelectModeStart();
	}

	//Flush textures if necessary due to config change
	if (flushTextures) {
		Flush(1);
	}

	unguard;
}

void UOpenGLRenderDevice::SetSceneNode(FSceneNode* Frame) {
	UTGLR_DEBUG_CALL_COUNT(SetSceneNode);
	guard(UOpenGLRenderDevice::SetSceneNode);

	EndBuffering();		// Flush vertex array before changing the projection matrix!

	m_sceneNodeCount++;

	//No need to set default AA state here
	//No need to set default projection state as this function always sets/initializes it
	SetDefaultVertexArray();
	SetDefaultShaderProgram();

	// Precompute stuff.
	FLOAT rcpFrameFX = 1.0f / Frame->FX;
	m_Aspect = Frame->FY * rcpFrameFX;
	m_RProjZ = appTan(Viewport->Actor->FovAngle * PI / 360.0);
	m_RFX2 = 2.0f * m_RProjZ * rcpFrameFX;
	m_RFY2 = 2.0f * m_RProjZ * rcpFrameFX;

	//Remember Frame->X and Frame->Y for scene node hack
	m_sceneNodeX = Frame->X;
	m_sceneNodeY = Frame->Y;

	// Set viewport.
	glViewport(Frame->XB, Viewport->SizeY - Frame->Y - Frame->YB, Frame->X, Frame->Y);

	//Decide whether or not to use Z range hack
	m_useZRangeHack = false;
	if (ZRangeHack && !GIsEditor) {
		m_useZRangeHack = true;
	}

	// Set projection.
	if (Frame->Viewport->IsOrtho()) {
		//Don't use Z range hack if ortho projection
		m_useZRangeHack = false;

		SetOrthoProjection();
	}
	else {
		SetProjectionStateNoCheck(false);
	}

	//Set clip planes if doing selection
	if (m_HitData) {
		if (Frame->Viewport->IsOrtho()) {
			float cp[4];
			FLOAT nX = Viewport->HitX - Frame->FX2;
			FLOAT pX = nX + Viewport->HitXL;
			FLOAT nY = Viewport->HitY - Frame->FY2;
			FLOAT pY = nY + Viewport->HitYL;

			nX *= m_RFX2 * 0.5f;
			pX *= m_RFX2 * 0.5f;
			nY *= m_RFY2 * 0.5f;
			pY *= m_RFY2 * 0.5f;

			cp[0] = +1.0; cp[1] = 0.0; cp[2] = 0.0; cp[3] = -nX;
			m_gclip.SetCp(0, cp);
			m_gclip.SetCpEnable(0, true);

			cp[0] = 0.0; cp[1] = +1.0; cp[2] = 0.0; cp[3] = -nY;
			m_gclip.SetCp(1, cp);
			m_gclip.SetCpEnable(1, true);

			cp[0] = -1.0; cp[1] = 0.0; cp[2] = 0.0; cp[3] = +pX;
			m_gclip.SetCp(2, cp);
			m_gclip.SetCpEnable(2, true);

			cp[0] = 0.0; cp[1] = -1.0; cp[2] = 0.0; cp[3] = +pY;
			m_gclip.SetCp(3, cp);
			m_gclip.SetCpEnable(3, true);

			//Near clip plane
			cp[0] = 0.0f; cp[1] = 0.0f; cp[2] = 1.0f; cp[3] = -0.5f;
			m_gclip.SetCp(4, cp);
			m_gclip.SetCpEnable(4, true);
		}
		else {
			FVector N[4];
			float cp[4];
			INT i;

			FLOAT nX = Viewport->HitX - Frame->FX2;
			FLOAT pX = nX + Viewport->HitXL;
			FLOAT nY = Viewport->HitY - Frame->FY2;
			FLOAT pY = nY + Viewport->HitYL;

			N[0] = (FVector(nX * Frame->RProj.Z, 0, 1) ^ FVector(0, -1, 0)).SafeNormal();
			N[1] = (FVector(pX * Frame->RProj.Z, 0, 1) ^ FVector(0, +1, 0)).SafeNormal();
			N[2] = (FVector(0, nY * Frame->RProj.Z, 1) ^ FVector(+1, 0, 0)).SafeNormal();
			N[3] = (FVector(0, pY * Frame->RProj.Z, 1) ^ FVector(-1, 0, 0)).SafeNormal();

			for (i = 0; i < 4; i++) {
				cp[0] = N[i].X;
				cp[1] = N[i].Y;
				cp[2] = N[i].Z;
				cp[3] = 0.0f;
				m_gclip.SetCp(i, cp);
				m_gclip.SetCpEnable(i, true);
			}

			//Near clip plane
			cp[0] = 0.0f; cp[1] = 0.0f; cp[2] = 1.0f; cp[3] = -0.5f;
			m_gclip.SetCp(4, cp);
			m_gclip.SetCpEnable(4, true);
		}
	}

	unguard;
}

void UOpenGLRenderDevice::Unlock(UBOOL Blit) {
	UTGLR_DEBUG_CALL_COUNT(Unlock);
	guard(UOpenGLRenderDevice::Unlock);

	EndBuffering();

	SetDefaultAAState();
	SetDefaultProjectionState();
	SetDefaultVertexArray();
	SetDefaultShaderProgram();

	// Unlock and render.
	check(LockCount == 1);

	//glFlush();

	if (Blit) {
		CheckGLErrorFlag(TEXT("please report this bug"));

		//Swap buffers
		{
			bool SwapBuffersStatus;

			SwapBuffersStatus = (SwapBuffers(m_hDC)) ? true : false;

			if (!m_prevSwapBuffersStatus) {
				check(SwapBuffersStatus);
			}
			m_prevSwapBuffersStatus = SwapBuffersStatus;
		}
	}

	--LockCount;

	//If doing selection, end and return hits
	if (m_HitData) {
		INT i;

		//End selection
		m_gclip.SelectModeEnd();

		*m_HitSize = m_HitCount;

		//Disable clip planes
		for (i = 0; i < 5; i++) {
			m_gclip.SetCpEnable(i, false);
		}
	}

	//Scan for old textures
	if (UseTexIdPool) {
		//Scan for old textures
		ScanForOldTextures();
	}

	//Increment current frame count
	m_currentFrameCount++;

	//Check for optional frame rate limit
	if (FrameRateLimit >= 20) {
#if defined UTGLR_DX_BUILD || defined UTGLR_RUNE_BUILD
		FLOAT curFrameTimestamp;
#else
		FTime curFrameTimestamp;
#endif
		float timeDiff;
		float rcpFrameRateLimit;

		//First time timer init if necessary
		InitFrameRateLimitTimerSafe();

		curFrameTimestamp = appSeconds();
		timeDiff = curFrameTimestamp - m_prevFrameTimestamp;
		m_prevFrameTimestamp = curFrameTimestamp;

		rcpFrameRateLimit = 1.0f / FrameRateLimit;
		if (timeDiff < rcpFrameRateLimit) {
			float sleepTime;

			sleepTime = rcpFrameRateLimit - timeDiff;
			appSleep(sleepTime);

			m_prevFrameTimestamp = appSeconds();
		}
	}


#if 0
	dbgPrintf("PR switch count = %u\n", m_prSwitchCount);
	dbgPrintf("AA switch count = %u\n", m_AASwitchCount);
	dbgPrintf("Scene node count = %u\n", m_sceneNodeCount);
	dbgPrintf("Scene node hack count = %u\n", m_sceneNodeHackCount);
	dbgPrintf("VB flush count = %u\n", m_vbFlushCount);
	dbgPrintf("Stat 0 count = %u\n", m_stat0Count);
	dbgPrintf("Stat 1 count = %u\n", m_stat1Count);
#endif


	unguard;
}

void UOpenGLRenderDevice::Flush(UBOOL AllowPrecache) {
	guard(UOpenGLRenderDevice::Flush);
	unsigned int u;
	TArray<GLuint> Binds;

	for (u = 0; u < NUM_CTTree_TREES; u++) {
		DWORD_CTTree_t *zeroPrefixBindTree = &m_zeroPrefixBindTrees[u];
		for (DWORD_CTTree_t::node_t *zpbmPtr = zeroPrefixBindTree->begin(); zpbmPtr != zeroPrefixBindTree->end(); zpbmPtr = zeroPrefixBindTree->next_node(zpbmPtr)) {
			Binds.AddItem(zpbmPtr->data.Id);
		}
		zeroPrefixBindTree->clear(&m_DWORD_CTTree_Allocator);
	}

	for (u = 0; u < NUM_CTTree_TREES; u++) {
		QWORD_CTTree_t *nonZeroPrefixBindTree = &m_nonZeroPrefixBindTrees[u];
		for (QWORD_CTTree_t::node_t *nzpbmPtr = nonZeroPrefixBindTree->begin(); nzpbmPtr != nonZeroPrefixBindTree->end(); nzpbmPtr = nonZeroPrefixBindTree->next_node(nzpbmPtr)) {
			Binds.AddItem(nzpbmPtr->data.Id);
		}
		nonZeroPrefixBindTree->clear(&m_QWORD_CTTree_Allocator);
	}

	m_nonZeroPrefixBindChain->mark_as_clear();

	while (QWORD_CTTree_NodePool_t::node_t *nzptipPtr = m_nonZeroPrefixTexIdPool->try_remove()) {
		Binds.AddItem(nzptipPtr->data.Id);
		m_QWORD_CTTree_Allocator.free_node(nzptipPtr);
	}

	for (TexPoolMap_t::node_t *RGBA8TpPtr = m_RGBA8TexPool->begin(); RGBA8TpPtr != m_RGBA8TexPool->end(); RGBA8TpPtr = m_RGBA8TexPool->next_node(RGBA8TpPtr)) {
		while (QWORD_CTTree_NodePool_t::node_t *texPoolNodePtr = RGBA8TpPtr->data.try_remove()) {
			Binds.AddItem(texPoolNodePtr->data.Id);
			m_QWORD_CTTree_Allocator.free_node(texPoolNodePtr);
		}
	}
	m_RGBA8TexPool->clear(&m_TexPoolMap_Allocator);

	while (QWORD_CTTree_NodePool_t::node_t *nzpnpPtr = m_nonZeroPrefixNodePool.try_remove()) {
		m_QWORD_CTTree_Allocator.free_node(nzpnpPtr);
	}

	if (Binds.Num()) {
		glDeleteTextures(Binds.Num(), (GLuint*)&Binds(0));
	}
	AllocatedTextures = 0;

	//Reset current texture ids to hopefully unused values
	for (u = 0; u < MAX_TMUNITS; u++) {
		TexInfo[u].CurrentCacheID = TEX_CACHE_ID_UNUSED;
		TexInfo[u].pBind = NULL;
	}

	if (AllowPrecache && UsePrecache && !GIsEditor) {
		PrecacheOnFlip = 1;
	}

	SetGamma(Viewport->GetOuterUClient()->Brightness);

	unguard;
}


void UOpenGLRenderDevice::DrawComplexSurface(FSceneNode* Frame, FSurfaceInfo& Surface, FSurfaceFacet& Facet) {
	UTGLR_DEBUG_CALL_COUNT(DrawComplexSurface);
	guard(UOpenGLRenderDevice::DrawComplexSurface);

	EndBuffering();

	if (SceneNodeHack) {
		if ((Frame->X != m_sceneNodeX) || (Frame->Y != m_sceneNodeY)) {
			m_sceneNodeHackCount++;
			SetSceneNode(Frame);
		}
	}

	SetDefaultAAState();
	SetDefaultProjectionState();
	//This function uses cached vertex array information
	//This function uses cached shader program state information

	check(Surface.Texture);

	//Hit select path
	if (m_HitData) {
		for (FSavedPoly* Poly = Facet.Polys; Poly; Poly = Poly->Next) {
			INT NumPts = Poly->NumPts;
			CGClip::vec3_t triPts[3];
			INT i;
			const FTransform* Pt;

			Pt = Poly->Pts[0];
			triPts[0].x = Pt->Point.X;
			triPts[0].y = Pt->Point.Y;
			triPts[0].z = Pt->Point.Z;

			for (i = 2; i < NumPts; i++) {
				Pt = Poly->Pts[i - 1];
				triPts[1].x = Pt->Point.X;
				triPts[1].y = Pt->Point.Y;
				triPts[1].z = Pt->Point.Z;

				Pt = Poly->Pts[i];
				triPts[2].x = Pt->Point.X;
				triPts[2].y = Pt->Point.Y;
				triPts[2].z = Pt->Point.Z;

				m_gclip.SelectDrawTri(triPts);
			}
		}

		return;
	}

	clock(ComplexCycles);

	//Calculate UDot and VDot intermediates for complex surface
	m_csUDot = Facet.MapCoords.XAxis | Facet.MapCoords.Origin;
	m_csVDot = Facet.MapCoords.YAxis | Facet.MapCoords.Origin;

	//Buffer static geometry
	//Sets m_csPolyCount
	//m_csPtCount set later from return value
	//Sets MultiDrawFirstArray and MultiDrawCountArray
	INT numVerts;
	numVerts = BufferStaticComplexSurfaceGeometry_VP(Facet);

	//Reject invalid surfaces early
	if (numVerts == 0) {
		return;
	}

	//Save number of points
	m_csPtCount = numVerts;

	//See if detail texture should be drawn
	//FogMap and DetailTexture are mutually exclusive effects
	bool drawDetailTexture = false;
	if ((DetailTextures != 0) && Surface.DetailTexture && !Surface.FogMap) {
		drawDetailTexture = true;
	}

	//Check for detail texture
	if (drawDetailTexture == true) {
		DWORD anyIsNearBits;

		//Buffer detail texture data
		anyIsNearBits = (this->*m_pBufferDetailTextureDataProc)(380.0f);

		//Do not draw detail texture if no vertices are near
		if (anyIsNearBits == 0) {
			drawDetailTexture = false;
		}
	}


	//Do static render passes state setup
	{
		const FVector &XAxis = Facet.MapCoords.XAxis;
		const FVector &YAxis = Facet.MapCoords.YAxis;
		FLOAT vsParams[8] = { XAxis.X, XAxis.Y, XAxis.Z, m_csUDot,
							  YAxis.X, YAxis.Y, YAxis.Z, m_csVDot };

		//Update shared uniform with complex surface vertex info
		glBindBuffer(GL_UNIFORM_BUFFER, m_csVertInfo);
		glBufferSubData(GL_UNIFORM_BUFFER, 0, 32, vsParams);
	}

	//Always render first up to three layers [base, macrotexture, lightmap] as single pass
	//First base texture layer must always be present
	//Single pass fog layer added as optional and final second to fourth layer
	DWORD PolyFlags = Surface.PolyFlags;
	DWORD csTex = 0;
	bool renderedSinglePassFog = false;
	bool renderedBaseTexture = false;
	bool rpSetDepthEqual = false;

	//Base texture
	if (csTex < 4) {
		m_csTexInfo[csTex].Info = Surface.Texture;
		m_csTexInfo[csTex].PolyFlags = PolyFlags & ~PF_FlatShaded;
		m_csTexInfo[csTex].PanBias = 0.0f;
		csTex++;
	}

	//Macrotexture
	if ((csTex < 4) && Surface.MacroTexture) {
		m_csTexInfo[csTex].Info = Surface.MacroTexture;
		m_csTexInfo[csTex].PolyFlags = PF_Modulated;
		m_csTexInfo[csTex].PanBias = -0.5f;
		csTex++;
	}

	//Lightmap
	if ((csTex < 4) && Surface.LightMap) {
		m_csTexInfo[csTex].Info = Surface.LightMap;
		m_csTexInfo[csTex].PolyFlags = PF_Modulated;
		m_csTexInfo[csTex].PanBias = -0.5f;
		csTex++;
	}

	//Draw detail texture
	if (drawDetailTexture == true) {
		GLuint singlePassDetail_prId = 0;

		//Check if single pass detail mode is enabled and if can use it
		if (SinglePassDetail) {
			//Only attempt single pass detail if haven't rendered the base texture yet
			if (!renderedBaseTexture) {
				//Single pass detail must be done with one or two normal passes
				if (csTex == 1) {
					if (DetailMax >= 2) {
						singlePassDetail_prId = m_prSingleTextureAndDetailTextureTwoLayer;
					}
					else {
						singlePassDetail_prId = m_prSingleTextureAndDetailTexture;
					}
				}
				else if (csTex == 2) {
					if (DetailMax >= 2) {
						singlePassDetail_prId = m_prDualTextureAndDetailTextureTwoLayer;
					}
					else {
						singlePassDetail_prId = m_prDualTextureAndDetailTexture;
					}
				}
			}
		}

		//Render single pass detail if found a supported combination
		if (singlePassDetail_prId != 0) {
			m_csTexInfo[csTex].Info = Surface.DetailTexture;
			m_csTexInfo[csTex].PolyFlags = PF_Modulated;
			m_csTexInfo[csTex].PanBias = 0.0f;
			csTex++;

			RenderSingleOrDualTextureAndDetailTexture_FP(csTex, PolyFlags & ~PF_FlatShaded, singlePassDetail_prId);

			csTex = 0;
		}
		//Otherwise render any previously accumulated passes and then detail texture
		else {
			//Render first pass if not already done so
			if (csTex != 0) {
				GLuint prId = m_prComplexSurfaceSingleTexture;

				switch (csTex) {
				case 1: prId = m_prComplexSurfaceSingleTexture; break;
				case 2: prId = m_prComplexSurfaceDualTextureModulated; break;
				case 3: prId = m_prComplexSurfaceTripleTextureModulated; break;
				default:
					;
				}
				RenderComplexSurfacePassNoCheck_FP(csTex, PolyFlags & ~PF_FlatShaded, prId);
				
				csTex = 0;
				renderedBaseTexture = true;
			}

			//Set depth compare to equal before subsequent pass for masked textures
			if (PolyFlags & PF_Masked) {
				if (rpSetDepthEqual == false) {
					rpSetDepthEqual = true;
					glDepthFunc(GL_EQUAL);
				}
			}

			//This function should only be called if at least one polygon will be detail textured
			DrawDetailTexture_FP(*Surface.DetailTexture);
		}
	}

	//Single pass fog if allowed
	if ((csTex < 4) && Surface.FogMap && SinglePassFog) {
		m_csTexInfo[csTex].Info = Surface.FogMap;
		m_csTexInfo[csTex].PolyFlags = PF_Highlighted;
		m_csTexInfo[csTex].PanBias = -0.5f;
		csTex++;

		renderedSinglePassFog = true;

		//Must render after add single pass fog layer
		GLuint prId = m_prComplexSurfaceSingleTexture;

		switch (csTex) {
		case 2: prId = m_prComplexSurfaceSingleTextureWithFog; break;
		case 3: prId = m_prComplexSurfaceDualTextureModulatedWithFog; break;
		case 4: prId = m_prComplexSurfaceTripleTextureModulatedWithFog; break;
		default:
			;
		}
		RenderComplexSurfacePassNoCheck_FP(csTex, PolyFlags & ~PF_FlatShaded, prId);

		csTex = 0;
		renderedBaseTexture = true;
	}

	//Render first pass if not already done so
	if (csTex != 0) {
		GLuint prId = m_prComplexSurfaceSingleTexture;

		switch (csTex) {
		case 1: prId = m_prComplexSurfaceSingleTexture; break;
		case 2: prId = m_prComplexSurfaceDualTextureModulated; break;
		case 3: prId = m_prComplexSurfaceTripleTextureModulated; break;
		default:
			;
		}
		RenderComplexSurfacePassNoCheck_FP(csTex, PolyFlags & ~PF_FlatShaded, prId);
		
		csTex = 0;
		renderedBaseTexture = true;
	}

	//Subsequent own pass fog if required
	if ((csTex == 0) && Surface.FogMap && !renderedSinglePassFog) {
		GLuint prId = m_prComplexSurfaceSingleTextureNoLod;

		//Set depth compare to equal before subsequent pass for masked textures
		if (PolyFlags & PF_Masked) {
			if (rpSetDepthEqual == false) {
				rpSetDepthEqual = true;
				glDepthFunc(GL_EQUAL);
			}
		}

		m_csTexInfo[csTex].Info = Surface.FogMap;
		m_csTexInfo[csTex].PolyFlags = PF_Highlighted;
		m_csTexInfo[csTex].PanBias = -0.5f;
		csTex++;

		RenderComplexSurfacePassNoCheck_FP(csTex, PF_Highlighted, prId);

		csTex = 0;
	}

	// UnrealEd selection.
	if (GIsEditor && (PolyFlags & (PF_Selected | PF_FlatShaded))) {
		DWORD polyColor;

		//No need to set default AA state here as it is always set on entry to DrawComplexSurface
		//No need to set default projection state here as it is always set on entry to DrawComplexSurface
		SetDefaultVertexArray();
		SetDefaultShaderProgram();

		SetNoTexture(0);
		SetBlend(PF_Highlighted);

		if (PolyFlags & PF_FlatShaded) {
			FPlane Color;

			Color.X = Surface.FlatColor.R / 255.0f;
			Color.Y = Surface.FlatColor.G / 255.0f;
			Color.Z = Surface.FlatColor.B / 255.0f;
			Color.W = 0.85f;
			if (PolyFlags & PF_Selected) {
				Color.X *= 1.5f;
				Color.Y *= 1.5f;
				Color.Z *= 1.5f;
				Color.W = 1.0f;
			}

			polyColor = FPlaneTo_RGBAClamped(&Color);
		}
		else {
			polyColor = 0x7F7F0000;
		}

		for (FSavedPoly* Poly = Facet.Polys; Poly; Poly = Poly->Next) {
			INT NumPts = Poly->NumPts;

			//Make sure at least NumPts entries are left in the vertex buffers
			if ((m_curVertexBufferPos + NumPts) >= VERTEX_ARRAY_SIZE) {
				FlushVertexBuffers();
			}

			//Lock vertexColor and texCoord0 buffers
			glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
			LockVertexColorBuffer();
			glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
			LockTexCoordBuffer(0);

			FGLTexCoord *pTexCoordArray = m_pTexCoordArray[0];
			FGLVertexColor *pVertexColorArray = m_pVertexColorArray;

			for (INT i = 0; i < Poly->NumPts; i++) {
				pTexCoordArray[i].u = 0.5f;
				pTexCoordArray[i].v = 0.5f;

				pVertexColorArray[i].x = Poly->Pts[i]->Point.X;
				pVertexColorArray[i].y = Poly->Pts[i]->Point.Y;
				pVertexColorArray[i].z = Poly->Pts[i]->Point.Z;
				pVertexColorArray[i].color = polyColor;
			}

			//Unlock vertexColor and texCoord0 buffers
			glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
			UnlockVertexColorBuffer();
			glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
			UnlockTexCoordBuffer(0);

			//Draw the triangle fan
			glDrawArrays(GL_TRIANGLE_FAN, m_curVertexBufferPos, NumPts);

			//Advance vertex buffer position
			m_curVertexBufferPos += NumPts;
		}
	}

	if (rpSetDepthEqual == true) {
		glDepthFunc(GL_LEQUAL);
	}

	unclock(ComplexCycles);
	unguard;
}

#ifdef UTGLR_RUNE_BUILD
void UOpenGLRenderDevice::PreDrawFogSurface() {
	UTGLR_DEBUG_CALL_COUNT(PreDrawFogSurface);
	guard(UOpenGLRenderDevice::PreDrawFogSurface);

	EndBuffering();

	SetDefaultAAState();
	SetDefaultProjectionState();
	SetDefaultVertexArray();
	SetDefaultShaderProgram();

	SetBlend(PF_AlphaBlend);
	SetNoTexture(0);

	unguard;
}

void UOpenGLRenderDevice::PostDrawFogSurface() {
	UTGLR_DEBUG_CALL_COUNT(PostDrawFogSurface);
	guard(UOpenGLRenderDevice::PostDrawFogSurface);

	SetBlend(0);

	unguard;
}

void UOpenGLRenderDevice::DrawFogSurface(FSceneNode* Frame, FFogSurf &FogSurf) {
	UTGLR_DEBUG_CALL_COUNT(DrawFogSurface);
	guard(UOpenGLRenderDevice::DrawFogSurface);

	FPlane Modulate(FogSurf.FogColor.X, FogSurf.FogColor.Y, FogSurf.FogColor.Z, 0.0f);

	FLOAT RFogDistance = 1.0f / FogSurf.FogDistance;

	if (FogSurf.PolyFlags & PF_Masked) {
		glDepthFunc(GL_EQUAL);
	}

	for (FSavedPoly* Poly = FogSurf.Polys; Poly; Poly = Poly->Next) {
		INT NumPts = Poly->NumPts;

		//Make sure at least NumPts entries are left in the vertex buffers
		if ((m_curVertexBufferPos + NumPts) >= VERTEX_ARRAY_SIZE) {
			FlushVertexBuffers();
		}

		//Lock vertexColor and texCoord0 buffers
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		LockVertexColorBuffer();
		glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
		LockTexCoordBuffer(0);

		FGLTexCoord *pTexCoordArray = m_pTexCoordArray[0];
		FGLVertexColor *pVertexColorArray = m_pVertexColorArray;

		INT Index = 0;
		for (INT i = 0; i < NumPts; i++) {
			FTransform* P = Poly->Pts[i];

			Modulate.W = P->Point.Z * RFogDistance;
			if (Modulate.W > 1.0f) {
				Modulate.W = 1.0f;
			}
			else if (Modulate.W < 0.0f) {
				Modulate.W = 0.0f;
			}

			FGLVertexColor &destVertexColor = pVertexColorArray[Index];
			destVertexColor.x = P->Point.X;
			destVertexColor.y = P->Point.Y;
			destVertexColor.z = P->Point.Z;
			destVertexColor.color = FPlaneTo_RGBA(&Modulate);

			FGLTexCoord &destTexCoord = pTexCoordArray[Index];
			destTexCoord.u = 0.0f;
			destTexCoord.v = 0.0f;

			Index++;
		}

		//Unlock vertexColor and texCoord0 buffers
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		UnlockVertexColorBuffer();
		glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
		UnlockTexCoordBuffer(0);

		//Draw the triangle fan
		glDrawArrays(GL_TRIANGLE_FAN, m_curVertexBufferPos, NumPts);

		//Advance vertex buffer position
		m_curVertexBufferPos += NumPts;
	}

	if (FogSurf.PolyFlags & PF_Masked) {
		glDepthFunc(GL_LEQUAL);
	}

	unguard;
}

void UOpenGLRenderDevice::PreDrawGouraud(FSceneNode* Frame, FLOAT FogDistance, FPlane FogColor) {
	UTGLR_DEBUG_CALL_COUNT(PreDrawGouraud);
	guard(UOpenGLRenderDevice::PreDrawGouraud);

	if (FogDistance > 0.0f) {
		EndBuffering();

		//Enable fog
		m_gpFogEnabled = true;
		{
			FLOAT fsParams[8];
			FLOAT rcpFogLen;

			//Fog color
			fsParams[0] = FogColor.X;
			fsParams[1] = FogColor.Y;
			fsParams[2] = FogColor.Z;
			fsParams[3] = FogColor.W;

			//Fog distance info
			rcpFogLen = 1.0f / (FogDistance - 0.0f);
			fsParams[4] = rcpFogLen;
			fsParams[5] = FogDistance * rcpFogLen;
			fsParams[6] = 0.0f;
			fsParams[7] = 0.0f;

			//Update shared uniform with fog info
			glBindBuffer(GL_UNIFORM_BUFFER, m_ubFragInfo);
			glBufferSubData(GL_UNIFORM_BUFFER, 16, 32, fsParams);
		}
	}

	unguard;
}

void UOpenGLRenderDevice::PostDrawGouraud(FLOAT FogDistance) {
	UTGLR_DEBUG_CALL_COUNT(PostDrawGouraud);
	guard(UOpenGLRenderDevice::PostDrawGouraud);

	if (FogDistance > 0.0f) {
		EndBuffering();

		//Disable fog
		m_gpFogEnabled = false;
	}

	unguard;
}
#endif

void UOpenGLRenderDevice::DrawGouraudPolygon(FSceneNode* Frame, FTextureInfo& Info, FTransTexture** Pts, INT NumPts, DWORD PolyFlags, FSpanBuffer* Span) {
	UTGLR_DEBUG_CALL_COUNT(DrawGouraudPolygon);
	guard(UOpenGLRenderDevice::DrawGouraudPolygon);

	//DrawGouraudPolygon only uses PolyFlags2 locally
	DWORD PolyFlags2 = 0;

	EndBufferingExcept(BV_TYPE_GOURAUD_POLYS);

	if (SceneNodeHack) {
		if ((Frame->X != m_sceneNodeX) || (Frame->Y != m_sceneNodeY)) {
			m_sceneNodeHackCount++;
			SetSceneNode(Frame);
		}
	}

	//Reject invalid polygons early so that other parts of the code do not have to deal with them
	if (NumPts < 3) {
		return;
	}
	if (NumPts > 30) {
		return;
	}

	//Hit select path
	if (m_HitData) {
		CGClip::vec3_t triPts[3];
		const FTransTexture* Pt;
		INT i;

		Pt = Pts[0];
		triPts[0].x = Pt->Point.X;
		triPts[0].y = Pt->Point.Y;
		triPts[0].z = Pt->Point.Z;

		for (i = 2; i < NumPts; i++) {
			Pt = Pts[i - 1];
			triPts[1].x = Pt->Point.X;
			triPts[1].y = Pt->Point.Y;
			triPts[1].z = Pt->Point.Z;

			Pt = Pts[i];
			triPts[2].x = Pt->Point.X;
			triPts[2].y = Pt->Point.Y;
			triPts[2].z = Pt->Point.Z;

			m_gclip.SelectDrawTri(triPts);
		}

		return;
	}

	//Load texture cache id
	QWORD CacheID = Info.CacheID;

	//Only attempt to alter texture cache id on certain textures
	if ((CacheID & 0xFF) == 0xE0) {
		//Alter texture cache id if masked texture hack is enabled and texture is masked
		CacheID |= ((PolyFlags & PF_Masked) ? TEX_CACHE_ID_FLAG_MASKED : 0) & m_maskedTextureHackMask;

		//Check for 16 bit texture option
		if (Use16BitTextures) {
			if (Info.Palette && (Info.Palette[128].A == 255)) {
				CacheID |= TEX_CACHE_ID_FLAG_16BIT;
			}
		}
	}

	//Decide if should request near Z range hack projection
	if (m_useZRangeHack && (GUglyHackFlags & 0x1)) {
		PolyFlags2 |= PF2_NEAR_Z_RANGE_HACK;
	}

	if ((m_curPolyFlags != PolyFlags) ||
		(m_curPolyFlags2 != PolyFlags2) ||
		(TexInfo[0].CurrentCacheID != CacheID) ||
		((m_curVertexBufferPos + m_bufferedVerts + NumPts) >= (VERTEX_ARRAY_SIZE - 30)) ||
		(m_bufferedVerts == 0))
	// flush drawing and set the state!
	{
		//Flush any previously buffered gouraud polys
		EndBuffering();

		//Check if vertex buffer flush is required
		if ((m_curVertexBufferPos + m_bufferedVerts + NumPts) >= (VERTEX_ARRAY_SIZE - 30)) {
			FlushVertexBuffers();
		}

		//Start gouraud polygon buffering
		StartBuffering(BV_TYPE_GOURAUD_POLYS);

		//Update current poly flags
		m_curPolyFlags = PolyFlags;
		m_curPolyFlags2 = PolyFlags2;

		//Set request near Z range hack projection flag
		m_requestNearZRangeHackProjection = (PolyFlags2 & PF2_NEAR_Z_RANGE_HACK) ? true : false;

		SetBlend(PolyFlags);
		SetTextureNoPanBias(0, Info, PolyFlags);

		//Check if should render fog
		//Set color flags
		if (PolyFlags & PF_Modulated) {
			m_requestedColorFlags = 0;
		}
		else {
			m_requestedColorFlags = CF_COLOR_ARRAY;

#ifdef UTGLR_RUNE_BUILD
			if ((PolyFlags & (PF_RenderFog | PF_Translucent | PF_Modulated | PF_AlphaBlend)) == PF_RenderFog) {
#elif UTGLR_UNREAL_227_BUILD
			if ((PolyFlags & (PF_RenderFog | PF_Translucent | PF_Modulated | PF_AlphaBlend)) == PF_RenderFog) {
#else
			if ((PolyFlags & (PF_RenderFog | PF_Translucent | PF_Modulated)) == PF_RenderFog) {
#endif
				m_requestedColorFlags = CF_COLOR_ARRAY | CF_FOG_MODE;
			}
		}

		//Set vertex array and shader program
		{
			GLuint vaState = (m_requestedColorFlags & CF_FOG_MODE) ? m_vaTwoColorOneTexcoord : m_vaOneColorOneTexcoord;
			GLuint prId = m_prDefaultRenderingState;

			SetVertexArray(vaState);

			if (m_requestedColorFlags & CF_FOG_MODE) {
				prId = m_prDefaultRenderingStateWithFog;
			}
#ifdef UTGLR_RUNE_BUILD
			if (m_gpFogEnabled) {
				prId = m_prDefaultRenderingStateWithLinearFog;
			}
#endif

			SetShaderProgram(prId);
		}

		//Lock vertexColor and texCoord0 buffers
		//Lock secondary color buffer if fog
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		LockVertexColorBuffer();
		if (m_requestedColorFlags & CF_FOG_MODE) {
			glBindBuffer(GL_ARRAY_BUFFER, m_vbSecondaryColor);
			LockSecondaryColorBuffer();
		}
		glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
		LockTexCoordBuffer(0);

		//Select a buffer verts proc
		if (m_requestedColorFlags & CF_FOG_MODE) {
			m_pBuffer3VertsProc = m_pBuffer3FoggedVertsProc;
		}
		else if (m_requestedColorFlags & CF_COLOR_ARRAY) {
			m_pBuffer3VertsProc = m_pBuffer3ColoredVertsProc;
		}
		else {
			m_pBuffer3VertsProc = m_pBuffer3BasicVertsProc;
		}
#ifdef UTGLR_RUNE_BUILD
		m_gpAlpha = 255;
		if (PolyFlags & PF_AlphaBlend) {
			m_gpAlpha = appRound(Info.Texture->Alpha * 255.0f);
			m_pBuffer3VertsProc = Buffer3Verts;
		}
#endif
	}

	//Buffer 3 vertices from the first (and perhaps only) triangle
	(m_pBuffer3VertsProc)(this, Pts);

	if (NumPts > 3) {
		//Buffer additional vertices from a clipped triangle
		BufferAdditionalClippedVerts(Pts, NumPts);
	}

	unguard;
}

void UOpenGLRenderDevice::DrawTile(FSceneNode* Frame, FTextureInfo& Info, FLOAT X, FLOAT Y, FLOAT XL, FLOAT YL, FLOAT U, FLOAT V, FLOAT UL, FLOAT VL, class FSpanBuffer* Span, FLOAT Z, FPlane Color, FPlane Fog, DWORD PolyFlags) {
	UTGLR_DEBUG_CALL_COUNT(DrawTile);
	guard(UOpenGLRenderDevice::DrawTile);

	//DrawTile does not use PolyFlags2
	const DWORD PolyFlags2 = 0;

	EndBufferingExcept(BV_TYPE_TILES);

	if (SceneNodeHack) {
		if ((Frame->X != m_sceneNodeX) || (Frame->Y != m_sceneNodeY)) {
			m_sceneNodeHackCount++;
			SetSceneNode(Frame);
		}
	}

	//Adjust Z coordinate if Z range hack is active
	if (m_useZRangeHack) {
		if ((Z >= 0.5f) && (Z < 8.0f)) {
			Z = (((Z - 0.5f) / 7.5f) * 4.0f) + 4.0f;
		}
	}

	FLOAT PX1 = X - Frame->FX2;
	FLOAT PX2 = PX1 + XL;
	FLOAT PY1 = Y - Frame->FY2;
	FLOAT PY2 = PY1 + YL;

	FLOAT RPX1 = m_RFX2 * PX1;
	FLOAT RPX2 = m_RFX2 * PX2;
	FLOAT RPY1 = m_RFY2 * PY1;
	FLOAT RPY2 = m_RFY2 * PY2;
	if (!Frame->Viewport->IsOrtho()) {
		RPX1 *= Z;
		RPX2 *= Z;
		RPY1 *= Z;
		RPY2 *= Z;
	}

	//Hit select path
	if (m_HitData) {
		CGClip::vec3_t triPts[3];

		triPts[0].x = RPX1;
		triPts[0].y = RPY1;
		triPts[0].z = Z;

		triPts[1].x = RPX2;
		triPts[1].y = RPY1;
		triPts[1].z = Z;

		triPts[2].x = RPX2;
		triPts[2].y = RPY2;
		triPts[2].z = Z;

		m_gclip.SelectDrawTri(triPts);

		triPts[0].x = RPX1;
		triPts[0].y = RPY1;
		triPts[0].z = Z;

		triPts[1].x = RPX2;
		triPts[1].y = RPY2;
		triPts[1].z = Z;

		triPts[2].x = RPX1;
		triPts[2].y = RPY2;
		triPts[2].z = Z;

		m_gclip.SelectDrawTri(triPts);

		return;
	}

	if (1) {
		//Load texture cache id
		QWORD CacheID = Info.CacheID;

		//Only attempt to alter texture cache id on certain textures
		if ((CacheID & 0xFF) == 0xE0) {
			//Alter texture cache id if masked texture hack is enabled and texture is masked
			CacheID |= ((PolyFlags & PF_Masked) ? TEX_CACHE_ID_FLAG_MASKED : 0) & m_maskedTextureHackMask;

			//Check for 16 bit texture option
			if (Use16BitTextures) {
				if (Info.Palette && (Info.Palette[128].A == 255)) {
					CacheID |= TEX_CACHE_ID_FLAG_16BIT;
				}
			}
		}

		//Check if need to start new tile buffering
		if ((m_curPolyFlags != PolyFlags) ||
			(m_curPolyFlags2 != PolyFlags2) ||
			(TexInfo[0].CurrentCacheID != CacheID) ||
			(m_curVertexBufferPos + m_bufferedVerts >= (VERTEX_ARRAY_SIZE - 6)) ||
			(m_bufferedVerts == 0))
		{
			//Flush any previously buffered tiles
			EndBuffering();

			//Check if vertex buffer flush is required
			if ((m_curVertexBufferPos + m_bufferedVerts) >= (VERTEX_ARRAY_SIZE - 6)) {
				FlushVertexBuffers();
			}

			//Start tile buffering
			StartBuffering(BV_TYPE_TILES);

			SetDefaultVertexArray();
			SetDefaultShaderProgram();

			//Update current poly flags (before possible local modification)
			m_curPolyFlags = PolyFlags;
			m_curPolyFlags2 = PolyFlags2;

#ifdef UTGLR_RUNE_BUILD
			if (Info.Palette && Info.Palette[128].A != 255 && !(PolyFlags & (PF_Translucent | PF_AlphaBlend))) {
#elif UTGLR_UNREAL_227_BUILD
			if (Info.Palette && Info.Palette[128].A != 255 && !(PolyFlags & (PF_Translucent | PF_AlphaBlend))) {
#else
			if (Info.Palette && Info.Palette[128].A != 255 && !(PolyFlags & PF_Translucent)) {
#endif
				PolyFlags |= PF_Highlighted | PF_Occlude;
			}

			SetBlend(PolyFlags);
			SetTextureNoPanBias(0, Info, PolyFlags);

			if (PolyFlags & PF_Modulated) {
				m_requestedColorFlags = 0;
			}
			else {
				m_requestedColorFlags = CF_COLOR_ARRAY;
			}

			//Lock vertexColor and texCoord0 buffers
			glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
			LockVertexColorBuffer();
			glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
			LockTexCoordBuffer(0);
		}

		//Get tile color
		DWORD tileColor;
		tileColor = 0xFFFFFFFF;
		if (!(PolyFlags & PF_Modulated)) {
			if (UseSSE2) {
#ifdef UTGLR_INCLUDE_SSE_CODE
				static __m128 fColorMul = { 255.0f, 255.0f, 255.0f, 0.0f };
				__m128 fColorMulReg;
				__m128 fColor;
				__m128 fAlpha;
				__m128i iColor;

				fColorMulReg = fColorMul;
				fColor = _mm_loadu_ps(&Color.X);
				fColor = _mm_mul_ps(fColor, fColorMulReg);

				fAlpha = _mm_setzero_ps();
				fAlpha = _mm_move_ss(fAlpha, fColorMulReg);
#ifdef UTGLR_RUNE_BUILD
				if (PolyFlags & PF_AlphaBlend) {
					fAlpha = _mm_mul_ss(fAlpha, _mm_load_ss(&Info.Texture->Alpha));
				}
#elif UTGLR_UNREAL_227_BUILD
				if (PolyFlags & PF_AlphaBlend) {
					fAlpha = _mm_mul_ss(fAlpha, _mm_load_ss(&Color.W));
				}
#endif
				fAlpha = _mm_shuffle_ps(fAlpha, fAlpha,  _MM_SHUFFLE(0, 1, 1, 1));

				fColor = _mm_or_ps(fColor, fAlpha);

				iColor = _mm_cvtps_epi32(fColor);
				iColor = _mm_packs_epi32(iColor, iColor);
				iColor = _mm_packus_epi16(iColor, iColor);

				tileColor = _mm_cvtsi128_si32(iColor);
#endif
			}
			else {
#ifdef UTGLR_RUNE_BUILD
				if (PolyFlags & PF_AlphaBlend) {
					Color.W = Info.Texture->Alpha;
					tileColor = FPlaneTo_RGBAClamped(&Color);
				}
				else {
					tileColor = FPlaneTo_RGBClamped_A255(&Color);
				}
#elif UTGLR_UNREAL_227_BUILD
				if (PolyFlags & PF_AlphaBlend) {
					tileColor = FPlaneTo_RGBAClamped(&Color);
				}
				else {
					tileColor = FPlaneTo_RGBClamped_A255(&Color);
				}
#else
				tileColor = FPlaneTo_RGBClamped_A255(&Color);
#endif
			}
		}

		//Buffer the tile
		FGLTexCoord *pTexCoordArray = &m_pTexCoordArray[0][m_bufferedVerts];
		FGLVertexColor *pVertexColorArray = &m_pVertexColorArray[m_bufferedVerts];

		pVertexColorArray[0].x = RPX1;
		pVertexColorArray[0].y = RPY1;
		pVertexColorArray[0].z = Z;
		pVertexColorArray[0].color = tileColor;

		pVertexColorArray[1].x = RPX2;
		pVertexColorArray[1].y = RPY1;
		pVertexColorArray[1].z = Z;
		pVertexColorArray[1].color = tileColor;

		pVertexColorArray[2].x = RPX2;
		pVertexColorArray[2].y = RPY2;
		pVertexColorArray[2].z = Z;
		pVertexColorArray[2].color = tileColor;

		pVertexColorArray[3].x = RPX1;
		pVertexColorArray[3].y = RPY1;
		pVertexColorArray[3].z = Z;
		pVertexColorArray[3].color = tileColor;

		pVertexColorArray[4].x = RPX2;
		pVertexColorArray[4].y = RPY2;
		pVertexColorArray[4].z = Z;
		pVertexColorArray[4].color = tileColor;

		pVertexColorArray[5].x = RPX1;
		pVertexColorArray[5].y = RPY2;
		pVertexColorArray[5].z = Z;
		pVertexColorArray[5].color = tileColor;

		FLOAT TexInfoUMult = TexInfo[0].UMult;
		FLOAT TexInfoVMult = TexInfo[0].VMult;

		FLOAT SU1 = (U) * TexInfoUMult;
		FLOAT SU2 = (U + UL) * TexInfoUMult;
		FLOAT SV1 = (V) * TexInfoVMult;
		FLOAT SV2 = (V + VL) * TexInfoVMult;

		pTexCoordArray[0].u = SU1;
		pTexCoordArray[0].v = SV1;

		pTexCoordArray[1].u = SU2;
		pTexCoordArray[1].v = SV1;

		pTexCoordArray[2].u = SU2;
		pTexCoordArray[2].v = SV2;

		pTexCoordArray[3].u = SU1;
		pTexCoordArray[3].v = SV1;

		pTexCoordArray[4].u = SU2;
		pTexCoordArray[4].v = SV2;

		pTexCoordArray[5].u = SU1;
		pTexCoordArray[5].v = SV2;

		m_bufferedVerts += 6;
	}

	unguard;
}

void UOpenGLRenderDevice::Draw3DLine(FSceneNode* Frame, FPlane Color, DWORD LineFlags, FVector P1, FVector P2) {
	UTGLR_DEBUG_CALL_COUNT(Draw3DLine);
	guard(UOpenGLRenderDevice::Draw3DLine);

	P1 = P1.TransformPointBy(Frame->Coords);
	P2 = P2.TransformPointBy(Frame->Coords);
	if (Frame->Viewport->IsOrtho()) {
		// Zoom.
		FLOAT rcpZoom = 1.0f / Frame->Zoom;
		P1.X = (P1.X * rcpZoom) + Frame->FX2;
		P1.Y = (P1.Y * rcpZoom) + Frame->FY2;
		P2.X = (P2.X * rcpZoom) + Frame->FX2;
		P2.Y = (P2.Y * rcpZoom) + Frame->FY2;
		P1.Z = P2.Z = 1;

		// See if points form a line parallel to our line of sight (i.e. line appears as a dot).
		if (Abs(P2.X - P1.X) + Abs(P2.Y - P1.Y) >= 0.2) {
			Draw2DLine(Frame, Color, LineFlags, P1, P2);
		}
		else if (Frame->Viewport->Actor->OrthoZoom < ORTHO_LOW_DETAIL) {
			Draw2DPoint(Frame, Color, LINE_None, P1.X - 1, P1.Y - 1, P1.X + 1, P1.Y + 1, P1.Z);
		}
	}
	else {
		EndBufferingExcept(BV_TYPE_LINES);

		//Hit select path
		if (m_HitData) {
			CGClip::vec3_t lnPts[2];

			lnPts[0].x = P1.X;
			lnPts[0].y = P1.Y;
			lnPts[0].z = P1.Z;

			lnPts[1].x = P2.X;
			lnPts[1].y = P2.Y;
			lnPts[1].z = P2.Z;

			m_gclip.SelectDrawLine(lnPts);

			return;
		}

		//Draw3DLine does not use PolyFlags2
		const DWORD PolyFlags = PF_Highlighted | PF_Occlude;
		const DWORD PolyFlags2 = 0;

		//Check if need to start new line buffering
		if ((m_curPolyFlags != PolyFlags) ||
			(m_curPolyFlags2 != PolyFlags2) ||
			(TexInfo[0].CurrentCacheID != TEX_CACHE_ID_NO_TEX) ||
			(m_curVertexBufferPos + m_bufferedVerts >= (VERTEX_ARRAY_SIZE - 2)) ||
			(m_bufferedVerts == 0))
		{
			//Flush any previously buffered lines
			EndBuffering();

			//Check if vertex buffer flush is required
			if ((m_curVertexBufferPos + m_bufferedVerts) >= (VERTEX_ARRAY_SIZE - 2)) {
				FlushVertexBuffers();
			}

			//Start line buffering
			StartBuffering(BV_TYPE_LINES);

			SetDefaultAAState();
			SetDefaultProjectionState();
			SetDefaultVertexArray();
			SetDefaultShaderProgram();

			//Update current poly flags
			m_curPolyFlags = PolyFlags;
			m_curPolyFlags2 = PolyFlags2;

			//Set blending and no texture for lines
			SetBlend(PolyFlags);
			SetNoTexture(0);

			//Lock vertexColor and texCoord0 buffers
			glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
			LockVertexColorBuffer();
			glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
			LockTexCoordBuffer(0);
		}

		//Get line color
		DWORD lineColor = FPlaneTo_RGBClamped_A255(&Color);

		//Buffer the line
		FGLTexCoord *pTexCoordArray = &m_pTexCoordArray[0][m_bufferedVerts];
		FGLVertexColor *pVertexColorArray = &m_pVertexColorArray[m_bufferedVerts];

		pVertexColorArray[0].x = P1.X;
		pVertexColorArray[0].y = P1.Y;
		pVertexColorArray[0].z = P1.Z;
		pVertexColorArray[0].color = lineColor;

		pVertexColorArray[1].x = P2.X;
		pVertexColorArray[1].y = P2.Y;
		pVertexColorArray[1].z = P2.Z;
		pVertexColorArray[1].color = lineColor;

		pTexCoordArray[0].u = 0.0f;
		pTexCoordArray[0].v = 0.0f;

		pTexCoordArray[1].u = 1.0f;
		pTexCoordArray[1].v = 0.0f;

		m_bufferedVerts += 2;
	}
	unguard;
}

void UOpenGLRenderDevice::Draw2DLine(FSceneNode* Frame, FPlane Color, DWORD LineFlags, FVector P1, FVector P2) {
	UTGLR_DEBUG_CALL_COUNT(Draw2DLine);
	guard(UOpenGLRenderDevice::Draw2DLine);

	EndBufferingExcept(BV_TYPE_LINES);


	//Get line coordinates back in 3D
	FLOAT X1Pos = m_RFX2 * (P1.X - Frame->FX2);
	FLOAT Y1Pos = m_RFY2 * (P1.Y - Frame->FY2);
	FLOAT X2Pos = m_RFX2 * (P2.X - Frame->FX2);
	FLOAT Y2Pos = m_RFY2 * (P2.Y - Frame->FY2);
	if (!Frame->Viewport->IsOrtho()) {
		X1Pos *= P1.Z;
		Y1Pos *= P1.Z;
		X2Pos *= P2.Z;
		Y2Pos *= P2.Z;
	}

	//Hit select path
	if (m_HitData) {
		CGClip::vec3_t lnPts[2];

		lnPts[0].x = X1Pos;
		lnPts[0].y = Y1Pos;
		lnPts[0].z = P1.Z;

		lnPts[1].x = X2Pos;
		lnPts[1].y = Y2Pos;
		lnPts[1].z = P2.Z;

		m_gclip.SelectDrawLine(lnPts);

		return;
	}

	//Draw2DLine does not use PolyFlags2
	const DWORD PolyFlags = PF_Highlighted | PF_Occlude;
	const DWORD PolyFlags2 = 0;

	//Check if need to start new line buffering
	if ((m_curPolyFlags != PolyFlags) ||
		(m_curPolyFlags2 != PolyFlags2) ||
		(TexInfo[0].CurrentCacheID != TEX_CACHE_ID_NO_TEX) ||
		(m_curVertexBufferPos + m_bufferedVerts >= (VERTEX_ARRAY_SIZE - 2)) ||
		(m_bufferedVerts == 0))
	{
		//Flush any previously buffered lines
		EndBuffering();

		//Check if vertex buffer flush is required
		if ((m_curVertexBufferPos + m_bufferedVerts) >= (VERTEX_ARRAY_SIZE - 2)) {
			FlushVertexBuffers();
		}

		//Start line buffering
		StartBuffering(BV_TYPE_LINES);

		SetDefaultAAState();
		SetDefaultProjectionState();
		SetDefaultVertexArray();
		SetDefaultShaderProgram();

		//Update current poly flags
		m_curPolyFlags = PolyFlags;
		m_curPolyFlags2 = PolyFlags2;

		//Set blending and no texture for lines
		SetBlend(PolyFlags);
		SetNoTexture(0);

		//Lock vertexColor and texCoord0 buffers
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		LockVertexColorBuffer();
		glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
		LockTexCoordBuffer(0);
	}

	//Get line color
	DWORD lineColor = FPlaneTo_RGBClamped_A255(&Color);

	//Buffer the line
	FGLTexCoord *pTexCoordArray = &m_pTexCoordArray[0][m_bufferedVerts];
	FGLVertexColor *pVertexColorArray = &m_pVertexColorArray[m_bufferedVerts];

	pVertexColorArray[0].x = X1Pos;
	pVertexColorArray[0].y = Y1Pos;
	pVertexColorArray[0].z = P1.Z;
	pVertexColorArray[0].color = lineColor;

	pVertexColorArray[1].x = X2Pos;
	pVertexColorArray[1].y = Y2Pos;
	pVertexColorArray[1].z = P2.Z;
	pVertexColorArray[1].color = lineColor;

	pTexCoordArray[0].u = 0.0f;
	pTexCoordArray[0].v = 0.0f;

	pTexCoordArray[1].u = 1.0f;
	pTexCoordArray[1].v = 0.0f;

	m_bufferedVerts += 2;

	unguard;
}

void UOpenGLRenderDevice::Draw2DPoint(FSceneNode* Frame, FPlane Color, DWORD LineFlags, FLOAT X1, FLOAT Y1, FLOAT X2, FLOAT Y2, FLOAT Z) {
	UTGLR_DEBUG_CALL_COUNT(Draw2DPoint);
	guard(UOpenGLRenderDevice::Draw2DPoint);

	EndBufferingExcept(BV_TYPE_POINTS);


	// Hack to fix UED selection problem with selection brush
	if (GIsEditor) {
		Z = 1.0f;
	}

	//Get point coordinates back in 3D
	FLOAT X1Pos = m_RFX2 * (X1 - Frame->FX2);
	FLOAT Y1Pos = m_RFY2 * (Y1 - Frame->FY2);
	FLOAT X2Pos = m_RFX2 * (X2 - Frame->FX2);
	FLOAT Y2Pos = m_RFY2 * (Y2 - Frame->FY2);
	if (!Frame->Viewport->IsOrtho()) {
		X1Pos *= Z;
		Y1Pos *= Z;
		X2Pos *= Z;
		Y2Pos *= Z;
	}

	//Hit select path
	if (m_HitData) {
		CGClip::vec3_t triPts[3];

		triPts[0].x = X1Pos;
		triPts[0].y = Y1Pos;
		triPts[0].z = Z;

		triPts[1].x = X2Pos;
		triPts[1].y = Y1Pos;
		triPts[1].z = Z;

		triPts[2].x = X2Pos;
		triPts[2].y = Y2Pos;
		triPts[2].z = Z;

		m_gclip.SelectDrawTri(triPts);

		triPts[0].x = X1Pos;
		triPts[0].y = Y1Pos;
		triPts[0].z = Z;

		triPts[1].x = X2Pos;
		triPts[1].y = Y2Pos;
		triPts[1].z = Z;

		triPts[2].x = X1Pos;
		triPts[2].y = Y2Pos;
		triPts[2].z = Z;

		m_gclip.SelectDrawTri(triPts);

		return;
	}

	//Draw2DPoint does not use PolyFlags2
	const DWORD PolyFlags = PF_Highlighted | PF_Occlude;
	const DWORD PolyFlags2 = 0;

	//Check if need to start new point buffering
	if ((m_curPolyFlags != PolyFlags) ||
		(m_curPolyFlags2 != PolyFlags2) ||
		(TexInfo[0].CurrentCacheID != TEX_CACHE_ID_NO_TEX) ||
		(m_curVertexBufferPos + m_bufferedVerts >= (VERTEX_ARRAY_SIZE - 6)) ||
		(m_bufferedVerts == 0))
	{
		//Flush any previously buffered points
		EndBuffering();

		//Check if vertex buffer flush is required
		if ((m_curVertexBufferPos + m_bufferedVerts) >= (VERTEX_ARRAY_SIZE - 6)) {
			FlushVertexBuffers();
		}

		//Start point buffering
		StartBuffering(BV_TYPE_POINTS);

		SetDefaultAAState();
		SetDefaultProjectionState();
		SetDefaultVertexArray();
		SetDefaultShaderProgram();

		//Update current poly flags
		m_curPolyFlags = PolyFlags;
		m_curPolyFlags2 = PolyFlags2;

		//Set blending and no texture for points
		SetBlend(PolyFlags);
		SetNoTexture(0);

		//Lock vertexColor and texCoord0 buffers
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		LockVertexColorBuffer();
		glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
		LockTexCoordBuffer(0);
	}

	//Get point color
	DWORD pointColor = FPlaneTo_RGBClamped_A255(&Color);

	//Buffer the point
	FGLTexCoord *pTexCoordArray = &m_pTexCoordArray[0][m_bufferedVerts];
	FGLVertexColor *pVertexColorArray = &m_pVertexColorArray[m_bufferedVerts];

	pVertexColorArray[0].x = X1Pos;
	pVertexColorArray[0].y = Y1Pos;
	pVertexColorArray[0].z = Z;
	pVertexColorArray[0].color = pointColor;

	pVertexColorArray[1].x = X2Pos;
	pVertexColorArray[1].y = Y1Pos;
	pVertexColorArray[1].z = Z;
	pVertexColorArray[1].color = pointColor;

	pVertexColorArray[2].x = X2Pos;
	pVertexColorArray[2].y = Y2Pos;
	pVertexColorArray[2].z = Z;
	pVertexColorArray[2].color = pointColor;

	pVertexColorArray[3].x = X1Pos;
	pVertexColorArray[3].y = Y1Pos;
	pVertexColorArray[3].z = Z;
	pVertexColorArray[3].color = pointColor;

	pVertexColorArray[4].x = X2Pos;
	pVertexColorArray[4].y = Y2Pos;
	pVertexColorArray[4].z = Z;
	pVertexColorArray[4].color = pointColor;

	pVertexColorArray[5].x = X1Pos;
	pVertexColorArray[5].y = Y2Pos;
	pVertexColorArray[5].z = Z;
	pVertexColorArray[5].color = pointColor;

	FLOAT SU1 = 0.0f;
	FLOAT SU2 = 1.0f;
	FLOAT SV1 = 0.0f;
	FLOAT SV2 = 1.0f;

	pTexCoordArray[0].u = SU1;
	pTexCoordArray[0].v = SV1;

	pTexCoordArray[1].u = SU2;
	pTexCoordArray[1].v = SV1;

	pTexCoordArray[2].u = SU2;
	pTexCoordArray[2].v = SV2;

	pTexCoordArray[3].u = SU1;
	pTexCoordArray[3].v = SV1;

	pTexCoordArray[4].u = SU2;
	pTexCoordArray[4].v = SV2;

	pTexCoordArray[5].u = SU1;
	pTexCoordArray[5].v = SV2;

	m_bufferedVerts += 6;

	unguard;
}


void UOpenGLRenderDevice::ClearZ(FSceneNode* Frame) {
	UTGLR_DEBUG_CALL_COUNT(ClearZ);
	guard(UOpenGLRenderDevice::ClearZ);

	EndBuffering();

	//Default AA state not required for glClear
	//Default projection state not required for glClear
	//Default color state not required for glClear
	//Default shader program state not required for glClear

	SetBlend(PF_Occlude);
	glClear(GL_DEPTH_BUFFER_BIT);

	unguard;
}

void UOpenGLRenderDevice::PushHit(const BYTE* Data, INT Count) {
	guard(UOpenGLRenderDevice::PushHit);

	INT i;

	EndBuffering();

	//Add to stack
	for (i = 0; i < Count; i += 4) {
		DWORD hitName = *(DWORD *)(Data + i);
		m_gclip.PushHitName(hitName);
	}

	unguard;
}

void UOpenGLRenderDevice::PopHit(INT Count, UBOOL bForce) {
	guard(UOpenGLRenderDevice::PopHit);

	EndBuffering();

	INT i;
	bool selHit;

	//Handle hit
	selHit = m_gclip.CheckNewSelectHit();
	if (selHit || bForce) {
		DWORD nHitNameBytes;

		nHitNameBytes = m_gclip.GetHitNameStackSize() * 4;
		if (nHitNameBytes <= m_HitBufSize) {
			m_gclip.GetHitNameStackValues((unsigned int *)m_HitData, nHitNameBytes / 4);
			m_HitCount = nHitNameBytes;
		}
		else {
			m_HitCount = 0;
		}
	}

	//Remove from stack
	for (i = 0; i < Count; i += 4) {
		m_gclip.PopHitName();
	}

	unguard;
}

void UOpenGLRenderDevice::GetStats(TCHAR* Result) {
	guard(UOpenGLRenderDevice::GetStats);

	double msPerCycle = GSecondsPerCycle * 1000.0f;
	appSprintf(
		Result,
		TEXT("OpenGL stats: Bind=%04.1f Image=%04.1f Complex=%04.1f Gouraud=%04.1f Tile=%04.1f"),
		msPerCycle * BindCycles,
		msPerCycle * ImageCycles,
		msPerCycle * ComplexCycles,
		msPerCycle * GouraudCycles,
		msPerCycle * TileCycles
	);

	unguard;
}

void UOpenGLRenderDevice::ReadPixels(FColor* Pixels) {
	guard(UOpenGLRenderDevice::ReadPixels);

	INT x, y;
	INT SizeX, SizeY;

	SizeX = Viewport->SizeX;
	SizeY = Viewport->SizeY;

	glReadPixels(0, 0, SizeX, SizeY, GL_RGBA, GL_UNSIGNED_BYTE, Pixels);
	for (y = 0; y < SizeY / 2; y++) {
		for (x = 0; x < SizeX; x++) {
			Exchange(Pixels[x + y * SizeX].R, Pixels[x + (SizeY - 1 - y) * SizeX].B);
			Exchange(Pixels[x + y * SizeX].G, Pixels[x + (SizeY - 1 - y) * SizeX].G);
			Exchange(Pixels[x + y * SizeX].B, Pixels[x + (SizeY - 1 - y) * SizeX].R);
		}
	}

	//Gamma correct screenshots if the option is true and the gamma ramp was set successfully
	if (GammaCorrectScreenshots && m_setGammaRampSucceeded) {
		FByteGammaRamp gammaByteRamp;
		BuildGammaRamp(SavedGammaCorrection, SavedGammaCorrection, SavedGammaCorrection, Brightness, gammaByteRamp);
		for (y = 0; y < SizeY; y++) {
			for (x = 0; x < SizeX; x++) {
				Pixels[x + y * SizeX].R = gammaByteRamp.red[Pixels[x + y * SizeX].R];
				Pixels[x + y * SizeX].G = gammaByteRamp.green[Pixels[x + y * SizeX].G];
				Pixels[x + y * SizeX].B = gammaByteRamp.blue[Pixels[x + y * SizeX].B];
			}
		}
	}

	unguard;
}

void UOpenGLRenderDevice::EndFlash() {
	UTGLR_DEBUG_CALL_COUNT(EndFlash);
	guard(UOpenGLRenderDevice::EndFlash);
	if ((FlashScale != FPlane(0.5f, 0.5f, 0.5f, 0.0f)) || (FlashFog != FPlane(0.0f, 0.0f, 0.0f, 0.0f))) {
		EndBuffering();

		SetDefaultAAState();
		SetDefaultProjectionState();
		SetDefaultVertexArray();
		SetDefaultShaderProgram();

		SetBlend(PF_Highlighted);
		SetNoTexture(0);

		FPlane tempPlane = FPlane(FlashFog.X, FlashFog.Y, FlashFog.Z, 1.0f - Min(FlashScale.X * 2.0f, 1.0f));
		DWORD flashColor = FPlaneTo_RGBA(&tempPlane);

		FLOAT RFX2 = m_RProjZ;
		FLOAT RFY2 = m_RProjZ * m_Aspect;

		//Adjust Z coordinate if Z range hack is active
		FLOAT ZCoord = 1.0f;
		if (m_useZRangeHack) {
			ZCoord = (((ZCoord - 0.5f) / 7.5f) * 4.0f) + 4.0f;
		}

		//Make sure at least 4 entries are left in the vertex buffers
		if ((m_curVertexBufferPos + 4) >= VERTEX_ARRAY_SIZE) {
			FlushVertexBuffers();
		}

		//Lock vertexColor and texCoord0 buffers
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		LockVertexColorBuffer();
		glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
		LockTexCoordBuffer(0);

		FGLTexCoord *pTexCoordArray = m_pTexCoordArray[0];
		FGLVertexColor *pVertexColorArray = m_pVertexColorArray;

		pTexCoordArray[0].u = 0.0f;
		pTexCoordArray[0].v = 0.0f;

		pTexCoordArray[1].u = 1.0f;
		pTexCoordArray[1].v = 0.0f;

		pTexCoordArray[2].u = 1.0f;
		pTexCoordArray[2].v = 1.0f;

		pTexCoordArray[3].u = 0.0f;
		pTexCoordArray[3].v = 1.0f;

		pVertexColorArray[0].x = RFX2 * (-1.0f * ZCoord);
		pVertexColorArray[0].y = RFY2 * (-1.0f * ZCoord);
		pVertexColorArray[0].z = ZCoord;
		pVertexColorArray[0].color = flashColor;

		pVertexColorArray[1].x = RFX2 * (+1.0f * ZCoord);
		pVertexColorArray[1].y = RFY2 * (-1.0f * ZCoord);
		pVertexColorArray[1].z = ZCoord;
		pVertexColorArray[1].color = flashColor;

		pVertexColorArray[2].x = RFX2 * (+1.0f * ZCoord);
		pVertexColorArray[2].y = RFY2 * (+1.0f * ZCoord);
		pVertexColorArray[2].z = ZCoord;
		pVertexColorArray[2].color = flashColor;

		pVertexColorArray[3].x = RFX2 * (-1.0f * ZCoord);
		pVertexColorArray[3].y = RFY2 * (+1.0f * ZCoord);
		pVertexColorArray[3].z = ZCoord;
		pVertexColorArray[3].color = flashColor;

		//Unlock vertexColor and texCoord0 buffers
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		UnlockVertexColorBuffer();
		glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
		UnlockTexCoordBuffer(0);

		//Draw the square
		glDrawArrays(GL_TRIANGLE_FAN, m_curVertexBufferPos, 4);

		//Advance vertex buffer position
		m_curVertexBufferPos += 4;
	}
	unguard;
}

void UOpenGLRenderDevice::PrecacheTexture(FTextureInfo& Info, DWORD PolyFlags) {
	guard(UOpenGLRenderDevice::PrecacheTexture);
	SetTextureNoPanBias(0, Info, PolyFlags);
	unguard;
}


//This function is safe to call multiple times to initialize once
void UOpenGLRenderDevice::InitNoTextureSafe(void) {
	guard(UOpenGLRenderDevice::InitNoTexture);
	unsigned int u;
	DWORD Data[4*4];

	//Return early if already initialized
	if (m_noTextureId != 0) {
		return;
	}

	for (u = 0; u < (4*4); u++) {
		Data[u] = 0xFFFFFFFF;
	}

	glGenTextures(1, &m_noTextureId);

	SetNoTexture(0);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, 4, 4, 0, GL_RGBA, GL_UNSIGNED_BYTE, Data);

	return;
	unguard;
}

void UOpenGLRenderDevice::ScanForDeletedTextures(void) {
	guard(UOpenGLRenderDevice::ScanForDeletedTextures);

	unsigned int u;

	//Check if currently bound texture may have been removed by another context
	for (u = 0; u < MAX_TMUNITS; u++) {
		FCachedTexture *pBind = TexInfo[u].pBind;
		if (pBind != NULL) {
			//Check if the texture may have been removed by another context
			if (NULL == FindCachedTexture(TexInfo[u].CurrentCacheID)) {
				//Cause texture to be updated next time
				//Some parts of texture resource may not be released while still bound
				TexInfo[u].CurrentCacheID = TEX_CACHE_ID_UNUSED;
				TexInfo[u].pBind = NULL;
			}
		}
	}

	unguard;
}

void UOpenGLRenderDevice::ScanForOldTextures(void) {
	guard(UOpenGLRenderDevice::ScanForOldTextures);

	unsigned int u;
	FCachedTexture *pCT;

	//Prevent currently bound textures from being recycled
	for (u = 0; u < MAX_TMUNITS; u++) {
		FCachedTexture *pBind = TexInfo[u].pBind;
		if (pBind != NULL) {
			//Update last used frame count so that the texture will not be recycled
			pBind->LastUsedFrameCount = m_currentFrameCount;

			//Move node to tail of linked list if in LRU list
			if (pBind->bindType == BIND_TYPE_NON_ZERO_PREFIX_LRU_LIST) {
				m_nonZeroPrefixBindChain->unlink(pBind);
				m_nonZeroPrefixBindChain->link_to_tail(pBind);
			}
		}
	}

	pCT = m_nonZeroPrefixBindChain->begin();
	while (pCT != m_nonZeroPrefixBindChain->end()) {
		DWORD numFramesSinceUsed = m_currentFrameCount - pCT->LastUsedFrameCount;
		if (numFramesSinceUsed > DynamicTexIdRecycleLevel) {
			//See if the tex pool is not enabled, or the tex format is not RGBA8, or the texture has mipmaps
			if (!UseTexPool || (pCT->texInternalFormat != GL_RGBA8) || (pCT->texParams.hasMipmaps)) {
				//Remove node from linked list
				m_nonZeroPrefixBindChain->unlink(pCT);

				//Get pointer to node in bind map
				QWORD_CTTree_t::node_t *pNode = (QWORD_CTTree_t::node_t *)((BYTE *)pCT - (DWORD)&(((QWORD_CTTree_t::node_t *)0)->data));
				//Extract tree index
				BYTE treeIndex = pCT->treeIndex;
				//Advanced cached texture pointer to next entry in linked list
				pCT = pCT->pNext;

				//Remove node from bind map
				m_nonZeroPrefixBindTrees[treeIndex].remove(pNode);

				//Add node plus texture id to texture id pool
				m_nonZeroPrefixTexIdPool->add(pNode);

				continue;
			}
			else {
				TexPoolMap_t::node_t *texPoolPtr;

#if 0
{
	static unsigned int s_c;
	dbgPrintf("utglr: TexPool free = %u, Id = %u, u = %u, v = %u\n",
		s_c++, pCT->Id, pCT->UBits, pCT->VBits);
}
#endif

				//Remove node from linked list
				m_nonZeroPrefixBindChain->unlink(pCT);

				//Create a key from the lg2 width and height of the texture object
				TexPoolMapKey_t texPoolKey = MakeTexPoolMapKey(pCT->UBits, pCT->VBits);

				//Get pointer to node in bind map
				QWORD_CTTree_t::node_t *pNode = (QWORD_CTTree_t::node_t *)((BYTE *)pCT - (DWORD)&(((QWORD_CTTree_t::node_t *)0)->data));
				//Extract tree index
				BYTE treeIndex = pCT->treeIndex;
				//Advanced cached texture pointer to next entry in linked list
				pCT = pCT->pNext;

				//Remove node from bind map
				m_nonZeroPrefixBindTrees[treeIndex].remove(pNode);

				//See if the key does not yet exist
				texPoolPtr = m_RGBA8TexPool->find(texPoolKey);
				//If the key does not yet exist, add an empty vector in its place
				if (texPoolPtr == 0) {
					texPoolPtr = m_TexPoolMap_Allocator.alloc_node();
					texPoolPtr->key = texPoolKey;
					texPoolPtr->data = QWORD_CTTree_NodePool_t();
					m_RGBA8TexPool->insert(texPoolPtr);
				}

				//Add node plus texture id to a list in the tex pool based on its dimensions
				texPoolPtr->data.add(pNode);

				continue;
			}
		}

		//The list is sorted
		//Stop searching on first one not to be recycled
		break;

		pCT = pCT->pNext;
	}

	unguard;
}

void UOpenGLRenderDevice::SetNoTextureNoCheck(INT Multi) {
	guard(UOpenGLRenderDevice::SetNoTexture);

	// Set small white texture.
	clock(BindCycles);

	//Set texture
	glBindTexture(GL_TEXTURE_2D, m_noTextureId);

	TexInfo[Multi].CurrentCacheID = TEX_CACHE_ID_NO_TEX;
	TexInfo[Multi].pBind = NULL;

	unclock(BindCycles);

	unguard;
}

FCachedTexture *UOpenGLRenderDevice::FindCachedTexture(QWORD CacheID) {
	bool isZeroPrefixCacheID = ((CacheID & 0xFFFFFFFF00000000ULL) == 0) ? true : false;
	FCachedTexture *pBind = NULL;

	if (isZeroPrefixCacheID) {
		DWORD CacheIDSuffix = (CacheID & 0x00000000FFFFFFFFULL);

		DWORD_CTTree_t *zeroPrefixBindTree = &m_zeroPrefixBindTrees[CTZeroPrefixCacheIDSuffixToTreeIndex(CacheIDSuffix)];
		DWORD_CTTree_t::node_t *bindTreePtr = zeroPrefixBindTree->find(CacheIDSuffix);
		if (bindTreePtr != 0) {
			pBind = &bindTreePtr->data;
		}
	}
	else {
		DWORD CacheIDSuffix = (CacheID & 0x00000000FFFFFFFF);
		DWORD treeIndex = CTNonZeroPrefixCacheIDSuffixToTreeIndex(CacheIDSuffix);

		QWORD_CTTree_t *nonZeroPrefixBindTree = &m_nonZeroPrefixBindTrees[treeIndex];
		QWORD_CTTree_t::node_t *bindTreePtr = nonZeroPrefixBindTree->find(CacheID);
		if (bindTreePtr != 0) {
			pBind = &bindTreePtr->data;
		}
	}

	return pBind;
}

UOpenGLRenderDevice::QWORD_CTTree_NodePool_t::node_t *UOpenGLRenderDevice::TryAllocFromTexPool(TexPoolMapKey_t texPoolKey) {
	TexPoolMap_t::node_t *texPoolPtr;

	//Search for the key in the map
	texPoolPtr = m_RGBA8TexPool->find(texPoolKey);
	if (texPoolPtr != 0) {
		QWORD_CTTree_NodePool_t::node_t *texPoolNodePtr;

		//Get a reference to the pool of nodes with tex ids of the right dimension
		QWORD_CTTree_NodePool_t &texPool = texPoolPtr->data;

		//Attempt to get a texture id from the tex pool
		if ((texPoolNodePtr = texPool.try_remove()) != 0) {
#if 0
{
	static unsigned int s_c;
	dbgPrintf("utglr: TexPool retrieve = %u, Id = %u, u = %u, v = %u\n",
		s_c++, texPoolNodePtr->data.Id, texPoolNodePtr->data.UBits, texPoolNodePtr->data.VBits);
}
#endif

			return texPoolNodePtr;
		}
	}

	return NULL;
}

BYTE UOpenGLRenderDevice::GenerateTexFilterParams(DWORD PolyFlags, FCachedTexture *pBind) {
	BYTE texFilter;

	//Generate tex filter params
	texFilter = 0;
	if (NoFiltering) {
		texFilter |= CT_MIN_FILTER_NEAREST;
	}
	else if (PolyFlags & PF_NoSmooth) {
		texFilter |= (!pBind->texParams.hasMipmaps) ? CT_MIN_FILTER_NEAREST : CT_MIN_FILTER_NEAREST_MIPMAP_NEAREST;
	}
	else {
		texFilter |= (!pBind->texParams.hasMipmaps) ? CT_MIN_FILTER_LINEAR : (UseTrilinear ? CT_MIN_FILTER_LINEAR_MIPMAP_LINEAR : CT_MIN_FILTER_LINEAR_MIPMAP_NEAREST);
		texFilter |= CT_MAG_FILTER_NEAREST_OR_LINEAR_BIT;
		if (MaxAnisotropy) {
			texFilter |= CT_ANISOTROPIC_FILTER_BIT;
		}
	}
	texFilter |= (pBind->texParams.texObjFilter & (CT_ADDRESS_U_CLAMP | CT_ADDRESS_V_CLAMP));

	return texFilter;
}

void UOpenGLRenderDevice::SetTexFilterNoCheck(FCachedTexture *pBind, BYTE texFilter) {
	BYTE texFilterXor;

	//Get changed tex filter params
	texFilterXor = pBind->texParams.filter ^ texFilter;

	//Update main copy of tex filter params early
	pBind->texParams.filter = texFilter;

	//Update changed params
	if (texFilterXor & CT_MIN_FILTER_MASK) {
		GLint intParam = GL_NEAREST;

		switch (texFilter & CT_MIN_FILTER_MASK) {
		case CT_MIN_FILTER_NEAREST: intParam = GL_NEAREST; break;
		case CT_MIN_FILTER_LINEAR: intParam = GL_LINEAR; break;
		case CT_MIN_FILTER_NEAREST_MIPMAP_NEAREST: intParam = GL_NEAREST_MIPMAP_NEAREST; break;
		case CT_MIN_FILTER_LINEAR_MIPMAP_NEAREST: intParam = GL_LINEAR_MIPMAP_NEAREST; break;
		case CT_MIN_FILTER_NEAREST_MIPMAP_LINEAR: intParam = GL_NEAREST_MIPMAP_LINEAR; break;
		case CT_MIN_FILTER_LINEAR_MIPMAP_LINEAR: intParam = GL_LINEAR_MIPMAP_LINEAR; break;
		default:
			;
		}

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, intParam);
	}
	if (texFilterXor & CT_MAG_FILTER_NEAREST_OR_LINEAR_BIT) {
		GLint intParam;

		intParam = (texFilter & CT_MAG_FILTER_NEAREST_OR_LINEAR_BIT) ? GL_LINEAR : GL_NEAREST;

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, intParam);
	}
	if (texFilterXor & CT_ANISOTROPIC_FILTER_BIT) {
		GLfloat floatParam;

		floatParam = (texFilter & CT_ANISOTROPIC_FILTER_BIT) ? (GLfloat)MaxAnisotropy : 1.0f;

		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, floatParam);
	}
	if (texFilterXor & CT_ADDRESS_U_CLAMP) {
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, (texFilter & CT_ADDRESS_U_CLAMP) ? GL_CLAMP_TO_EDGE : GL_REPEAT);
	}
	if (texFilterXor & CT_ADDRESS_V_CLAMP) {
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, (texFilter & CT_ADDRESS_V_CLAMP) ? GL_CLAMP_TO_EDGE : GL_REPEAT);
	}

	return;
}

//This function must use Tex.CurrentCacheID and NEVER use Info.CacheID to reference the texture cache id
//This makes it work with the masked texture hack code
void UOpenGLRenderDevice::SetTextureNoCheck(FTexInfo& Tex, FTextureInfo& Info, DWORD PolyFlags) {
	guard(UOpenGLRenderDevice::SetTexture);

	// Make current.
	clock(BindCycles);

	//Search for texture in cache
	FCachedTexture *pBind;
	bool existingBind = false;
	bool needTexAllocate = true;
	pBind = FindCachedTexture(Tex.CurrentCacheID);

	if (pBind) {
		//Update when texture last referenced
		pBind->LastUsedFrameCount = m_currentFrameCount;

		//Check if texture is in LRU list
		if (pBind->bindType == BIND_TYPE_NON_ZERO_PREFIX_LRU_LIST) {
			//Move node to tail of linked list
			m_nonZeroPrefixBindChain->unlink(pBind);
			m_nonZeroPrefixBindChain->link_to_tail(pBind);
		}

		existingBind = true;
		needTexAllocate = false;
	}
	else {
		bool isZeroPrefixCacheID = ((Tex.CurrentCacheID & 0xFFFFFFFF00000000ULL) == 0) ? true : false;
		if (isZeroPrefixCacheID) {
			DWORD CacheIDSuffix = (Tex.CurrentCacheID & 0x00000000FFFFFFFFULL);
			DWORD_CTTree_t *zeroPrefixBindTree = &m_zeroPrefixBindTrees[CTZeroPrefixCacheIDSuffixToTreeIndex(CacheIDSuffix)];
			DWORD_CTTree_t::node_t *pNewNode;

			//Insert new texture info
			pNewNode = m_DWORD_CTTree_Allocator.alloc_node();
			pNewNode->key = CacheIDSuffix;
			zeroPrefixBindTree->insert(pNewNode);
			pBind = &pNewNode->data;
			pBind->LastUsedFrameCount = m_currentFrameCount;

			//Set bind type
			pBind->bindType = BIND_TYPE_ZERO_PREFIX;

			//Set default tex params
			pBind->texParams = CT_DEFAULT_TEX_PARAMS;
			pBind->dynamicTexBits = (PolyFlags & PF_NoSmooth) ? DT_NO_SMOOTH_BIT : 0;

			//Cache texture info for the new texture
			CacheTextureInfo(pBind, Info, PolyFlags);

			//Allocate a new texture Id
			glGenTextures(1, &pBind->Id);
			AllocatedTextures++;
		}
		else {
			DWORD CacheIDSuffix = (Tex.CurrentCacheID & 0x00000000FFFFFFFF);
			DWORD treeIndex = CTNonZeroPrefixCacheIDSuffixToTreeIndex(CacheIDSuffix);
			QWORD_CTTree_t *nonZeroPrefixBindTree = &m_nonZeroPrefixBindTrees[treeIndex];
			QWORD_CTTree_t::node_t *pNewNode;

			//Allocate a new node
			//Use the node pool if it is not empty
			pNewNode = m_nonZeroPrefixNodePool.try_remove();
			if (!pNewNode) {
				pNewNode = m_QWORD_CTTree_Allocator.alloc_node();
			}

			//Insert new texture info
			pNewNode->key = Tex.CurrentCacheID;
			nonZeroPrefixBindTree->insert(pNewNode);
			pBind = &pNewNode->data;
			pBind->LastUsedFrameCount = m_currentFrameCount;

			//Set bind type
			pBind->bindType = BIND_TYPE_NON_ZERO_PREFIX_LRU_LIST;
			if (CacheStaticMaps && ((Tex.CurrentCacheID & 0xFF) == 0x18)) {
				pBind->bindType = BIND_TYPE_NON_ZERO_PREFIX;
			}

			//Save tree index
			pBind->treeIndex = (BYTE)treeIndex;

			//Set default tex params
			pBind->texParams = CT_DEFAULT_TEX_PARAMS;
			pBind->dynamicTexBits = (PolyFlags & PF_NoSmooth) ? DT_NO_SMOOTH_BIT : 0;

			//Check if texture should be in LRU list
			if (pBind->bindType == BIND_TYPE_NON_ZERO_PREFIX_LRU_LIST) {
				//Add node to linked list
				m_nonZeroPrefixBindChain->link_to_tail(pBind);
			}

			//Cache texture info for the new texture
			CacheTextureInfo(pBind, Info, PolyFlags);

			//See if the tex pool is enabled
			bool needTexIdAllocate = true;
			if (UseTexPool) {
				//See if the format will be RGBA8
				//Only textures without mipmaps are stored in the tex pool
				if ((pBind->texType == TEX_TYPE_NORMAL) && (Info.NumMips == 1)) {
					QWORD_CTTree_NodePool_t::node_t *texPoolNodePtr;

					//Create a key from the lg2 width and height of the texture object
					TexPoolMapKey_t texPoolKey = MakeTexPoolMapKey(pBind->UBits, pBind->VBits);

					//Attempt to allocate from the tex pool
					texPoolNodePtr = TryAllocFromTexPool(texPoolKey);
					if (texPoolNodePtr) {
						//Use texture id from node in tex pool
						pBind->Id = texPoolNodePtr->data.Id;

						//Use tex params from node in tex pool
						pBind->texParams = texPoolNodePtr->data.texParams;
						pBind->dynamicTexBits = texPoolNodePtr->data.dynamicTexBits;

						//Then add node to free list
						m_nonZeroPrefixNodePool.add(texPoolNodePtr);

						//Clear the need tex id and tex allocate flags
						needTexIdAllocate = false;
						needTexAllocate = false;
					}
				}
			}
			if (needTexIdAllocate) {
				QWORD_CTTree_NodePool_t::node_t *nzptipPtr;

				//Allocate a new texture id
				if (UseTexIdPool && ((nzptipPtr = m_nonZeroPrefixTexIdPool->try_remove()) != 0)) {
					//Use texture id from node in tex id pool
					pBind->Id = nzptipPtr->data.Id;

					//Use tex params from node in tex id pool
					pBind->texParams = nzptipPtr->data.texParams;
					pBind->dynamicTexBits = nzptipPtr->data.dynamicTexBits;

					//Then add node to free list
					m_nonZeroPrefixNodePool.add(nzptipPtr);
				}
				else {
					glGenTextures(1, &pBind->Id);
					AllocatedTextures++;
				}
			}
		}
	}

	//Save pointer to current texture bind for current texture unit
	Tex.pBind = pBind;

	//Set texture
	glBindTexture(GL_TEXTURE_2D, pBind->Id);

	unclock(BindCycles);

	// Account for all the impact on scale normalization.
	Tex.UMult = pBind->UMult;
	Tex.VMult = pBind->VMult;

	//Check for any changes to dynamic texture object parameters
	{
		BYTE desiredDynamicTexBits;

		desiredDynamicTexBits = (PolyFlags & PF_NoSmooth) ? DT_NO_SMOOTH_BIT : 0;
		if (desiredDynamicTexBits != pBind->dynamicTexBits) {
			BYTE dynamicTexBitsXor;

			dynamicTexBitsXor = desiredDynamicTexBits ^ pBind->dynamicTexBits;

			//Update dynamic tex bits early as there are no subsequent dependencies
			pBind->dynamicTexBits = desiredDynamicTexBits;

			if (dynamicTexBitsXor & DT_NO_SMOOTH_BIT) {
				BYTE texFilter;

				//Generate tex filter params
				texFilter = GenerateTexFilterParams(PolyFlags, pBind);

				//Set tex filter state
				SetTexFilter(pBind, texFilter);
			}
		}
	}

	// Upload if needed.
	if (!existingBind || Info.bRealtimeChanged) {
		UploadTextureExec(Info, PolyFlags, pBind, existingBind, needTexAllocate);
	}

	unguard;
}

void UOpenGLRenderDevice::UploadTextureExec(FTextureInfo& Info, DWORD PolyFlags, FCachedTexture *pBind, bool existingBind, bool needTexAllocate) {
	FColor paletteIndex0;

	// Cleanup texture flags.
	if (SupportsLazyTextures) {
		Info.Load();
	}
	Info.bRealtimeChanged = 0;

	//Set palette index 0 to black for masked paletted textures
	if (Info.Palette && (PolyFlags & PF_Masked)) {
		paletteIndex0 = Info.Palette[0];
		Info.Palette[0] = FColor(0,0,0,0);
	}

	// Download the texture.
	clock(ImageCycles);

	// width * height * 4 bytes per pixel
	DWORD memAllocSize = 1 << (pBind->UBits + pBind->VBits + 2);
	if (memAllocSize > LOCAL_TEX_COMPOSE_BUFFER_SIZE) {
		m_texComposeMemMark = FMemMark(GMem);
		m_texConvertCtx.pCompose = New<BYTE>(GMem, memAllocSize);
	}
	else {
		m_texConvertCtx.pCompose = m_localTexComposeBuffer;
	}

	m_texConvertCtx.pBind = pBind;

	UBOOL SkipMipmaps = (Info.NumMips == 1);
	INT MaxLevel = pBind->MaxLevel;

	//Only update texture state for new textures
	if (!existingBind) {
		DWORD texMaxLevel;
		BYTE texFilter;

		//Set tex max level param
		//This is set once for new textures and never changed afterwards
		texMaxLevel = 1000;
		if (!SkipMipmaps) {
			texMaxLevel = MaxLevel;
		}
		else {
			texMaxLevel = 0;
		}
		//Update if different than default for new textures
		if (texMaxLevel != 1000) {
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, texMaxLevel);
		}

		//Flag indicating if texture has mipmaps
		//This is set once for new textures and never changed afterwards
		pBind->texParams.hasMipmaps = (!SkipMipmaps) ? true : false;

		//Texture filter params set once for each texture
		pBind->texParams.texObjFilter = 0;
#ifdef UTGLR_UNREAL_227_BUILD
		// Check if clamped or wrapped (f.e. for Skyboxes). Smirftsch
		if (Info.UClampMode) {
			pBind->texParams.texObjFilter |= CT_ADDRESS_U_CLAMP;
		}
		if (Info.VClampMode) {
			pBind->texParams.texObjFilter |= CT_ADDRESS_V_CLAMP;
		}
#endif


		//Generate tex filter params
		texFilter = GenerateTexFilterParams(PolyFlags, pBind);

		//Set tex filter state
		SetTexFilter(pBind, texFilter);
	}


	//Some textures only upload the base texture
	INT MaxUploadLevel = MaxLevel;
	if (SkipMipmaps) {
		MaxUploadLevel = 0;
	}


	//Set initial texture width and height in the context structure
	//Setup code must ensure that both UBits and VBits are greater than or equal to 0
	m_texConvertCtx.texWidthPow2 = 1 << pBind->UBits;
	m_texConvertCtx.texHeightPow2 = 1 << pBind->VBits;

	INT Level;
	for (Level = 0; Level <= MaxUploadLevel; Level++) {
		// Convert the mipmap.
		INT MipIndex = pBind->BaseMip + Level;
		INT stepBits = 0;
		if (MipIndex >= Info.NumMips) {
			stepBits = MipIndex - (Info.NumMips - 1);
			MipIndex = Info.NumMips - 1;
		}
		m_texConvertCtx.stepBits = stepBits;

		FMipmapBase* Mip = Info.Mips[MipIndex];
		if (!Mip || !Mip->DataPtr) {
			//Skip looking at any subsequent mipmap pointers
			break;
		}
		else {
			//Texture data conversion if necessary
			switch (pBind->texType) {
			case TEX_TYPE_COMPRESSED_DXT1:
				//No conversion required for compressed DXT1 textures
				break;

			case TEX_TYPE_COMPRESSED_DXT3:
				//No conversion required for compressed DXT3 textures
				break;

			case TEX_TYPE_COMPRESSED_DXT5:
				//No conversion required for compressed DXT5 textures
				break;

			case TEX_TYPE_COMPRESSED_DXT1_TO_DXT3:
				guard(ConvertDXT1_DXT3);
				ConvertDXT1_DXT3(Mip, Level);
				unguard;
				break;

			case TEX_TYPE_PALETTED:
				//Not supported
				break;

			case TEX_TYPE_HAS_PALETTE:
				guard(ConvertP8_RGBA8888);
				if (stepBits == 0) {
					ConvertP8_RGBA8888_NoStep(Mip, Info.Palette, Level);
				}
				else {
					ConvertP8_RGBA8888(Mip, Info.Palette, Level);
				}
				unguard;
				break;

			default:
				guard(ConvertBGRA7777);
				(this->*pBind->pConvertBGRA7777)(Mip, Level);
				unguard;
			}

			BYTE *Src = (BYTE*)m_texConvertCtx.pCompose;
			DWORD texWidth, texHeight;

			//Get current texture width and height
			texWidth = m_texConvertCtx.texWidthPow2;
			texHeight = m_texConvertCtx.texHeightPow2;

			//Calculate and save next texture width and height
			//Both are divided by two down to a floor of 1
			//Texture width and height must be even powers of 2 for the following code to work
			m_texConvertCtx.texWidthPow2 = (texWidth & 0x1) | (texWidth >> 1);
			m_texConvertCtx.texHeightPow2 = (texHeight & 0x1) | (texHeight >> 1);

			if (!needTexAllocate) {
				//Update existing texture
				switch (pBind->texType) {
				case TEX_TYPE_COMPRESSED_DXT1:
				case TEX_TYPE_COMPRESSED_DXT1_TO_DXT3:
				case TEX_TYPE_COMPRESSED_DXT3:
				case TEX_TYPE_COMPRESSED_DXT5:
					guard(glCompressedTexSubImage2D);
#if 0
{
	static unsigned int s_c;
	dbgPrintf("utglr: glCompressedTexSubImage2D = %u\n", s_c++);
}
#endif
					glCompressedTexSubImage2D(
						GL_TEXTURE_2D,
						Level,
						0,
						0,
						texWidth,
						texHeight,
						pBind->texInternalFormat,
						(pBind->texType == TEX_TYPE_COMPRESSED_DXT1) ? (texWidth * texHeight / 2) : (texWidth * texHeight),
						(pBind->texType == TEX_TYPE_COMPRESSED_DXT1_TO_DXT3) ? Src : Mip->DataPtr);
					unguard;

					break;

				default:
					guard(glTexSubImage2D);
#if 0
{
	static unsigned int s_c;
	dbgPrintf("utglr: glTexSubImage2D = %u\n", s_c++);
}
#endif
					glTexSubImage2D(
						GL_TEXTURE_2D,
						Level,
						0,
						0,
						texWidth,
						texHeight,
						pBind->texSourceFormat,
						GL_UNSIGNED_BYTE,
						Src);
					unguard;
				}
			}
			else {
				//Create new texture and update
				switch (pBind->texType) {
				case TEX_TYPE_COMPRESSED_DXT1:
				case TEX_TYPE_COMPRESSED_DXT1_TO_DXT3:
				case TEX_TYPE_COMPRESSED_DXT3:
				case TEX_TYPE_COMPRESSED_DXT5:
					guard(glCompressedTexImage2D);
#if 0
{
	static unsigned int s_c;
	dbgPrintf("utglr: glCompressedTexImage2D = %u\n", s_c++);
}
#endif
					glCompressedTexImage2D(
						GL_TEXTURE_2D,
						Level,
						pBind->texInternalFormat,
						texWidth,
						texHeight,
						0,
						(pBind->texType == TEX_TYPE_COMPRESSED_DXT1) ? (texWidth * texHeight / 2) : (texWidth * texHeight),
						(pBind->texType == TEX_TYPE_COMPRESSED_DXT1_TO_DXT3) ? Src : Mip->DataPtr);
					unguard;

					break;

				default:
					guard(glTexImage2D);
#if 0
{
	static unsigned int s_c;
	dbgPrintf("utglr: glTexImage2D = %u\n", s_c++);
}
#endif
					glTexImage2D(
						GL_TEXTURE_2D,
						Level,
						pBind->texInternalFormat,
						texWidth,
						texHeight,
						0,
						pBind->texSourceFormat,
						GL_UNSIGNED_BYTE,
						Src);
					unguard;
				}
			}
		}
	}

	if (memAllocSize > LOCAL_TEX_COMPOSE_BUFFER_SIZE) {
		m_texComposeMemMark.Pop();
	}

	unclock(ImageCycles);

	//Restore palette index 0 for masked paletted textures
	if (Info.Palette && (PolyFlags & PF_Masked)) {
		Info.Palette[0] = paletteIndex0;
	}

	// Cleanup.
	if (SupportsLazyTextures) {
		Info.Unload();
	}

	return;
}

void UOpenGLRenderDevice::CacheTextureInfo(FCachedTexture *pBind, const FTextureInfo &Info, DWORD PolyFlags) {
#if 0
{
	dbgPrintf("utglr: CacheId = %08X:%08X\n",
		(DWORD)((QWORD)Info.CacheID >> 32), (DWORD)((QWORD)Info.CacheID & 0xFFFFFFFF));
}
{
	const UTexture *pTexture = Info.Texture;
	const TCHAR *pName = pTexture->GetFullName();
	if (pName) dbgPrintf("utglr: TexName = %s\n", appToAnsi(pName));
}
{
	dbgPrintf("utglr: NumMips = %d\n", Info.NumMips);
}
{
	unsigned int u;
	TCHAR dbgStr[1024];
	TCHAR numStr[32];

	dbgStr[0] = _T('\0');
	appStrcat(dbgStr, TEXT("utglr: ZPBindTree Size = "));
	for (u = 0; u < NUM_CTTree_TREES; u++) {
		appSprintf(numStr, TEXT("%u"), m_zeroPrefixBindTrees[u].calc_size());
		appStrcat(dbgStr, numStr);
		if (u != (NUM_CTTree_TREES - 1)) appStrcat(dbgStr, TEXT(", "));
	}
	dbgPrintf("%s\n", appToAnsi(dbgStr));

	dbgStr[0] = _T('\0');
	appStrcat(dbgStr, TEXT("utglr: NZPBindTree Size = "));
	for (u = 0; u < NUM_CTTree_TREES; u++) {
		appSprintf(numStr, TEXT("%u"), m_nonZeroPrefixBindTrees[u].calc_size());
		appStrcat(dbgStr, numStr);
		if (u != (NUM_CTTree_TREES - 1)) appStrcat(dbgStr, TEXT(", "));
	}
	dbgPrintf("%s\n", appToAnsi(dbgStr));
}
#endif

	// Figure out OpenGL-related scaling for the texture.
	DWORD texFlags = 0;
	INT BaseMip = 0;
	INT MaxLevel;
	INT UBits = Info.Mips[0]->UBits;
	INT VBits = Info.Mips[0]->VBits;
	INT UCopyBits = 0;
	INT VCopyBits = 0;
	if ((UBits - VBits) > MaxLogUOverV) {
		VCopyBits += (UBits - VBits) - MaxLogUOverV;
		VBits = UBits - MaxLogUOverV;
	}
	if ((VBits - UBits) > MaxLogVOverU) {
		UCopyBits += (VBits - UBits) - MaxLogVOverU;
		UBits = VBits - MaxLogVOverU;
	}
	if (UBits < MinLogTextureSize) {
		UCopyBits += MinLogTextureSize - UBits;
		UBits += MinLogTextureSize - UBits;
	}
	if (VBits < MinLogTextureSize) {
		VCopyBits += MinLogTextureSize - VBits;
		VBits += MinLogTextureSize - VBits;
	}
	if (UBits > MaxLogTextureSize) {
		BaseMip += UBits - MaxLogTextureSize;
		VBits -= UBits - MaxLogTextureSize;
		UBits = MaxLogTextureSize;
		if (VBits < 0) {
			VCopyBits = -VBits;
			VBits = 0;
		}
	}
	if (VBits > MaxLogTextureSize) {
		BaseMip += VBits - MaxLogTextureSize;
		UBits -= VBits - MaxLogTextureSize;
		VBits = MaxLogTextureSize;
		if (UBits < 0) {
			UCopyBits = -UBits;
			UBits = 0;
		}
	}

	pBind->BaseMip = BaseMip;
	MaxLevel = Min(UBits, VBits) - MinLogTextureSize;
	if (MaxLevel < 0) {
		MaxLevel = 0;
	}
	pBind->MaxLevel = MaxLevel;
	pBind->UBits = UBits;
	pBind->VBits = VBits;

	pBind->UMult = 1.0f / (Info.UScale * (Info.USize << UCopyBits));
	pBind->VMult = 1.0f / (Info.VScale * (Info.VSize << VCopyBits));

	pBind->UClampVal = Info.UClamp - 1;
	pBind->VClampVal = Info.VClamp - 1;

	//Check for texture that does not require clamping
	//No clamp required if ((Info.UClamp == Info.USize) & (Info.VClamp == Info.VSize))
	if (((Info.UClamp ^ Info.USize) | (Info.VClamp ^ Info.VSize)) == 0) {
		texFlags |= TEX_FLAG_NO_CLAMP;
	}


	//Determine texture type
	//PolyFlags PF_Masked cannot change if existing texture is updated as it caches texture type information here
	pBind->texType = TEX_TYPE_NONE;
	if (SupportsTC) {
		switch (Info.Format) {
		case TEXF_DXT1:
			if (TexDXT1ToDXT3 && (!(PolyFlags & PF_Masked))) {
				pBind->texType = TEX_TYPE_COMPRESSED_DXT1_TO_DXT3;
				//texSourceFormat not used for compressed textures
				pBind->texInternalFormat = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
			}
			else {
				pBind->texType = TEX_TYPE_COMPRESSED_DXT1;
				//texSourceFormat not used for compressed textures
				pBind->texInternalFormat = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT;
			}
			break;

#ifdef UTGLR_UNREAL_227_BUILD
		case TEXF_DXT3:
			pBind->texType = TEX_TYPE_COMPRESSED_DXT3;
			//texSourceFormat not used for compressed textures
			pBind->texInternalFormat = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
			break;

		case TEXF_DXT5:
			pBind->texType = TEX_TYPE_COMPRESSED_DXT5;
			//texSourceFormat not used for compressed textures
			pBind->texInternalFormat = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
			break;
#endif

		default:
			;
		}
	}
	if (pBind->texType != TEX_TYPE_NONE) {
		//Using compressed texture
	}
	else if (Info.Palette) {
		pBind->texType = TEX_TYPE_HAS_PALETTE;
		pBind->texSourceFormat = GL_RGBA;
		pBind->texInternalFormat = GL_RGBA8;
		//Check if texture should be 16-bit
		if (PolyFlags & PF_Memorized) {
			pBind->texInternalFormat = (PolyFlags & PF_Masked) ? GL_RGB5_A1 : GL_RGB5;
		}
	}
	else {
		pBind->texType = TEX_TYPE_NORMAL;

		pBind->texSourceFormat = GL_BGRA;
		if (texFlags & TEX_FLAG_NO_CLAMP) {
			pBind->pConvertBGRA7777 = &UOpenGLRenderDevice::ConvertBGRA7777_BGRA8888_NoClamp;
		}
		else {
			pBind->pConvertBGRA7777 = &UOpenGLRenderDevice::ConvertBGRA7777_BGRA8888;
		}

		pBind->texInternalFormat = GL_RGBA8;
	}

	return;
}


void UOpenGLRenderDevice::ConvertDXT1_DXT3(const FMipmapBase *Mip, INT Level) {
	DWORD *pSrc = (DWORD *)Mip->DataPtr;
	DWORD *pDest = (DWORD *)m_texConvertCtx.pCompose;
	DWORD numBlocks = 1 << (Max(0, (INT)m_texConvertCtx.pBind->UBits - Level - 2) + Max(0, (INT)m_texConvertCtx.pBind->VBits - Level - 2));
	for (DWORD block = 0; block < numBlocks; block++) {
		*pDest = 0xFFFFFFFF;
		*(pDest + 1) = 0xFFFFFFFF;
		*(pDest + 2) = *pSrc;
		*(pDest + 3) = *(pSrc + 1);
		pSrc += 2;
		pDest += 4;
	}

	UTGLR_DEBUG_TEX_CONVERT_COUNT(ConvertDXT1_DXT3);
}

void UOpenGLRenderDevice::ConvertP8_RGBA8888(const FMipmapBase *Mip, const FColor *Palette, INT Level) {
	FColor* Ptr = (FColor*)m_texConvertCtx.pCompose;
	INT StepBits = m_texConvertCtx.stepBits;
	DWORD UMask = Mip->USize - 1;
	DWORD VMask = Mip->VSize - 1;
	INT ij_inc = 1 << StepBits;
	INT i_stop = 1 << Max(0, (INT)m_texConvertCtx.pBind->VBits - Level + StepBits);
	INT j_stop = 1 << Max(0, (INT)m_texConvertCtx.pBind->UBits - Level + StepBits);
	INT i = 0;
	do { //i_stop always >= 1
		BYTE* Base = (BYTE*)Mip->DataPtr + (i & VMask) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1
			*Ptr++ = Palette[Base[j & UMask]];
		} while ((j += ij_inc) < j_stop);
	} while ((i += ij_inc) < i_stop);

	UTGLR_DEBUG_TEX_CONVERT_COUNT(ConvertP8_RGBA8888);
}

void UOpenGLRenderDevice::ConvertP8_RGBA8888_NoStep(const FMipmapBase *Mip, const FColor *Palette, INT Level) {
	FColor* Ptr = (FColor*)m_texConvertCtx.pCompose;
	DWORD UMask = Mip->USize - 1;
	DWORD VMask = Mip->VSize - 1;
	INT i_stop = m_texConvertCtx.texHeightPow2;
	INT j_stop = m_texConvertCtx.texWidthPow2;
	INT i = 0;
	do { //i_stop always >= 1
		BYTE* Base = (BYTE*)Mip->DataPtr + (i & VMask) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1
			*Ptr++ = Palette[Base[j & UMask]];
		} while ((j += 1) < j_stop);
	} while ((i += 1) < i_stop);

	UTGLR_DEBUG_TEX_CONVERT_COUNT(ConvertP8_RGBA8888_NoStep);
}

void UOpenGLRenderDevice::ConvertBGRA7777_BGRA8888(const FMipmapBase *Mip, INT Level) {
	FColor* Ptr = (FColor*)m_texConvertCtx.pCompose;
	INT StepBits = m_texConvertCtx.stepBits;
	DWORD VMask = Mip->VSize - 1;
	DWORD VClampVal = m_texConvertCtx.pBind->VClampVal;
	DWORD UMask = Mip->USize - 1;
	DWORD UClampVal = m_texConvertCtx.pBind->UClampVal;
	INT ij_inc = 1 << StepBits;
	INT i_stop = 1 << Max(0, (INT)m_texConvertCtx.pBind->VBits - Level + StepBits);
	INT j_stop = 1 << Max(0, (INT)m_texConvertCtx.pBind->UBits - Level + StepBits);
	INT i = 0;
	do { //i_stop always >= 1
		FColor* Base = (FColor*)Mip->DataPtr + Min<DWORD>(i & VMask, VClampVal) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1;
			const FColor& Src = Base[Min<DWORD>(j & UMask, UClampVal)];
			*(DWORD *)Ptr = *((DWORD *)&Src) * 2; // because of 7777
			Ptr++;
		} while ((j += ij_inc) < j_stop);
	} while ((i += ij_inc) < i_stop);

	UTGLR_DEBUG_TEX_CONVERT_COUNT(ConvertBGRA7777_BGRA8888);
}

void UOpenGLRenderDevice::ConvertBGRA7777_BGRA8888_NoClamp(const FMipmapBase *Mip, INT Level) {
	FColor* Ptr = (FColor*)m_texConvertCtx.pCompose;
	INT StepBits = m_texConvertCtx.stepBits;
	DWORD VMask = Mip->VSize - 1;
	DWORD UMask = Mip->USize - 1;
	INT ij_inc = 1 << StepBits;
	INT i_stop = 1 << Max(0, (INT)m_texConvertCtx.pBind->VBits - Level + StepBits);
	INT j_stop = 1 << Max(0, (INT)m_texConvertCtx.pBind->UBits - Level + StepBits);
	INT i = 0;
	do { //i_stop always >= 1
		FColor* Base = (FColor*)Mip->DataPtr + (DWORD)(i & VMask) * Mip->USize;
		INT j = 0;
		do { //j_stop always >= 1;
			const FColor& Src = Base[(DWORD)(j & UMask)];
			*(DWORD *)Ptr = *((DWORD *)&Src) * 2; // because of 7777
			Ptr++;
		} while ((j += ij_inc) < j_stop);
	} while ((i += ij_inc) < i_stop);

	UTGLR_DEBUG_TEX_CONVERT_COUNT(ConvertBGRA7777_BGRA8888_NoClamp);
}


void UOpenGLRenderDevice::SetBlendNoCheck(DWORD blendFlags) {
	guardSlow(UOpenGLRenderDevice::SetBlend);

	// Detect changes in the blending modes.
	DWORD Xor = m_curBlendFlags ^ blendFlags;

	//Save copy of current blend flags
	DWORD curBlendFlags = m_curBlendFlags;

	//Update main copy of current blend flags early
	m_curBlendFlags = blendFlags;

#ifdef UTGLR_RUNE_BUILD
	const DWORD GL_BLEND_FLAG_BITS = PF_Translucent | PF_Modulated | PF_Highlighted | PF_AlphaBlend;
#elif UTGLR_UNREAL_227_BUILD
	const DWORD GL_BLEND_FLAG_BITS = PF_Translucent | PF_Modulated | PF_Highlighted | PF_AlphaBlend;
#else
	const DWORD GL_BLEND_FLAG_BITS = PF_Translucent | PF_Modulated | PF_Highlighted;
#endif
	DWORD relevantBlendFlagBits = GL_BLEND_FLAG_BITS | m_smoothMaskedTexturesBit;
	if (Xor & (relevantBlendFlagBits)) {
		if (!(blendFlags & (relevantBlendFlagBits))) {
			glDisable(GL_BLEND);
		}
		else {
			if (!(curBlendFlags & (relevantBlendFlagBits))) {
				glEnable(GL_BLEND);
			}
			if (blendFlags & PF_Translucent) {
				glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
			}
			else if (blendFlags & PF_Modulated) {
				glBlendFunc(GL_DST_COLOR, GL_SRC_COLOR);
			}
			else if (blendFlags & PF_Highlighted) {
				glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
			}
#ifdef UTGLR_RUNE_BUILD
			else if (blendFlags & PF_AlphaBlend) {
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			}
#elif UTGLR_UNREAL_227_BUILD
			else if (blendFlags & PF_AlphaBlend) {
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			}
#endif
			else if (blendFlags & PF_Masked) {
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			}
		}
	}
#ifdef UTGLR_UNREAL_227_BUILD
	if (Xor & (PF_Masked | PF_AlphaBlend)) {
#else
	if (Xor & PF_Masked) {
#endif
		if (blendFlags & PF_Masked) {
#ifdef UTGLR_UNREAL_227_BUILD
			//Enable alpha test with alpha ref of 0.5
			m_fsBlendInfo[0] = 0.5f;

			//Update shared uniform with fragment info
			glBindBuffer(GL_UNIFORM_BUFFER, m_ubFragInfo);
			glBufferSubData(GL_UNIFORM_BUFFER, 0, 16, m_fsBlendInfo);

			if (m_useAlphaToCoverageForMasked) {
				m_alphaToCoverageEnabled = true;
				glEnable(GL_SAMPLE_ALPHA_TO_COVERAGE);
			}
		}
		else if (blendFlags & PF_AlphaBlend) {
			//Enable alpha test with alpha ref of 0.01
			m_fsBlendInfo[0] = 0.01f;
#else
			//Enable alpha test with alpha ref of 0.5
			m_fsBlendInfo[0] = 0.5f;
#endif

			//Update shared uniform with fragment info
			glBindBuffer(GL_UNIFORM_BUFFER, m_ubFragInfo);
			glBufferSubData(GL_UNIFORM_BUFFER, 0, 16, m_fsBlendInfo);

			if (m_useAlphaToCoverageForMasked) {
				m_alphaToCoverageEnabled = true;
				glEnable(GL_SAMPLE_ALPHA_TO_COVERAGE);
			}
		}
		else {
			//Disable alpha test
			m_fsBlendInfo[0] = -1.0f;

			//Update shared uniform with fragment info
			glBindBuffer(GL_UNIFORM_BUFFER, m_ubFragInfo);
			glBufferSubData(GL_UNIFORM_BUFFER, 0, 16, m_fsBlendInfo);

			if (m_useAlphaToCoverageForMasked) {
				m_alphaToCoverageEnabled = false;
				glDisable(GL_SAMPLE_ALPHA_TO_COVERAGE);
			}
		}
	}
	if (Xor & PF_Invisible) {
		UBOOL flag = ((blendFlags & PF_Invisible) == 0) ? GL_TRUE : GL_FALSE;
		glColorMask(flag, flag, flag, flag);
	}
	if (Xor & PF_Occlude) {
		UBOOL flag = ((blendFlags & PF_Occlude) == 0) ? GL_FALSE : GL_TRUE;
		glDepthMask(flag);
	}

	unguardSlow;
}



void UOpenGLRenderDevice::SetAAStateNoCheck(bool AAEnable) {
	//Save new AA state
	m_curAAEnable = AAEnable;

	m_AASwitchCount++;

	//Set new AA state
	if (AAEnable) {
		glEnable(GL_MULTISAMPLE);
	}
	else {
		glDisable(GL_MULTISAMPLE);
	}

	return;
}


void UOpenGLRenderDevice::AllocateVertexBufferNamesSafe(void) {
	//Do not allocate names if already allocated
	if (m_allocatedVertexBufferNames) {
		return;
	}

	//Bind vertex array 0
	m_vaCurrent = 0;
	glBindVertexArray(0);

	//Bind array buffer 0
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	//Allocate names
	glGenVertexArrays(1, &m_vaOneColor);
	glGenVertexArrays(1, &m_vaOneColorOneTexcoord);
	glGenVertexArrays(1, &m_vaTwoColorOneTexcoord);

	glGenBuffers(1, &m_vbVertexColor);
	glGenBuffers(1, &m_vbSecondaryColor);
	glGenBuffers(1, m_vbTexCoord);

	//Mark names as allocated
	m_allocatedVertexBufferNames = true;

	return;
}

void UOpenGLRenderDevice::FreeVertexBufferNamesSafe(void) {
	//Do not free names if not allocated
	if (!m_allocatedVertexBufferNames) {
		return;
	}

	//Bind vertex array 0
	m_vaCurrent = 0;
	glBindVertexArray(0);

	//Bind array buffer 0
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	//Delete names
	glDeleteVertexArrays(1, &m_vaOneColor);
	glDeleteVertexArrays(1, &m_vaOneColorOneTexcoord);
	glDeleteVertexArrays(1, &m_vaTwoColorOneTexcoord);

	glDeleteBuffers(1, &m_vbVertexColor);
	glDeleteBuffers(1, &m_vbSecondaryColor);
	glDeleteBuffers(1, m_vbTexCoord);

	//Mark names as not allocated
	m_allocatedVertexBufferNames = false;

	return;
}

bool UOpenGLRenderDevice::InitializeVertexBuffers(void) {
	GLenum glError;

	//Clear GL error
	while ((glError = glGetError()) != GL_NO_ERROR) {
	}


	//Vertex and primary color
	glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
	glBufferData(GL_ARRAY_BUFFER, sizeof(FGLVertexColor) * VERTEX_ARRAY_SIZE, NULL, GL_DYNAMIC_DRAW);

	//Secondary color
	glBindBuffer(GL_ARRAY_BUFFER, m_vbSecondaryColor);
	glBufferData(GL_ARRAY_BUFFER, sizeof(FGLSecondaryColor) * VERTEX_ARRAY_SIZE, NULL, GL_DYNAMIC_DRAW);

	//Normal array


	//TexCoord arrays
	glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(FGLTexCoord) * VERTEX_ARRAY_SIZE, NULL, GL_DYNAMIC_DRAW);


	//Bind array buffer 0
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	//Check for any errors creating vertex buffers
	glError = glGetError();
	if (glError != GL_NO_ERROR) {
		return false;
	}

	return true;
}


bool UOpenGLRenderDevice::LoadShader(GLuint shaderId, const char *pShader, const char *pName) {
	const unsigned int SHADER_INFO_LOG_SIZE = 1024;
	char shaderInfoLog[SHADER_INFO_LOG_SIZE];
	GLint compileStatus;

	if (DebugBit(DEBUG_BIT_BASIC)) {
		dbgPrintf("utglr: Loading shader \"%s\"\n", pName);
	}

	//Load shader source
	glShaderSource(shaderId, 1, &pShader, NULL);

	//Complile shader
	glCompileShader(shaderId);
	glGetShaderiv(shaderId, GL_COMPILE_STATUS, &compileStatus);
	if (GL_FALSE == compileStatus) {
		if (DebugBit(DEBUG_BIT_BASIC)) {
			//Get shader info log
			shaderInfoLog[0] = '\0';
			glGetShaderInfoLog(shaderId, SHADER_INFO_LOG_SIZE, NULL, shaderInfoLog);
			dbgPrintf("utglr: Shader compile failed\n");
			dbgPrintf("utglr: Shader info log:\n%s\n", shaderInfoLog);
		}
		return false;
	}

	return true;
}

bool UOpenGLRenderDevice::LinkShaderProgram(GLuint prId, GLuint vsId, GLuint gsId, GLuint fsId, const char *pName) {
	const unsigned int PROGRAM_INFO_LOG_SIZE = 1024;
	char programInfoLog[PROGRAM_INFO_LOG_SIZE];
	GLint linkStatus;

	if (DebugBit(DEBUG_BIT_BASIC)) {
		dbgPrintf("utglr: Linking shader program \"%s\"\n", pName);
	}

	//Attach shaders
	glAttachShader(prId, vsId);
	if (gsId) {
		glAttachShader(prId, gsId);
	}
	glAttachShader(prId, fsId);

	//Link program
	glLinkProgram(prId);
	glGetProgramiv(prId, GL_LINK_STATUS, &linkStatus);
	if (GL_FALSE == linkStatus) {
		if (DebugBit(DEBUG_BIT_BASIC)) {
			//Get program info log
			programInfoLog[0] = '\0';
			glGetProgramInfoLog(prId, PROGRAM_INFO_LOG_SIZE, NULL, programInfoLog);
			dbgPrintf("utglr: Program link failed\n");
			dbgPrintf("utglr: Program info log:\n%s\n", programInfoLog);
		}
		return false;
	}

	return true;
}

void UOpenGLRenderDevice::AllocateShaderNamesSafe(void) {
	unsigned int u;

	//Do not allocate names if already allocated
	if (m_allocatedShaderNames) {
		return;
	}

	//Allocate names
	glGenBuffers(1, &m_ubTrInfo);
	glGenBuffers(1, &m_csVertInfo);
	glGenBuffers(1, &m_ubFragInfo);

	m_vsDefaultRenderingState = glCreateShader(GL_VERTEX_SHADER);
	m_vsDefaultRenderingStateWithFog = glCreateShader(GL_VERTEX_SHADER);
	m_vsDefaultRenderingStateWithLinearFog = glCreateShader(GL_VERTEX_SHADER);
	for (u = 0; u < 4; u++) {
		m_vsComplexSurface[u] = glCreateShader(GL_VERTEX_SHADER);
	}
	m_vsDetailTexture = glCreateShader(GL_VERTEX_SHADER);
	for (u = 0; u < 2; u++) {
		m_vsComplexSurfaceAndDetailTexture[u] = glCreateShader(GL_VERTEX_SHADER);
	}

	m_fsDefaultRenderingState = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsDefaultRenderingStateWithFog = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsDefaultRenderingStateWithLinearFog = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsComplexSurfaceSingleTexture = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsComplexSurfaceSingleTextureNoLod = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsComplexSurfaceDualTextureModulated = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsComplexSurfaceTripleTextureModulated = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsComplexSurfaceSingleTextureWithFog = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsComplexSurfaceDualTextureModulatedWithFog = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsComplexSurfaceTripleTextureModulatedWithFog = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsDetailTexture = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsDetailTextureTwoLayer = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsSingleTextureAndDetailTexture = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsSingleTextureAndDetailTextureTwoLayer = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsDualTextureAndDetailTexture = glCreateShader(GL_FRAGMENT_SHADER);
	m_fsDualTextureAndDetailTextureTwoLayer = glCreateShader(GL_FRAGMENT_SHADER);

	m_prDefaultRenderingState = glCreateProgram();
	m_prDefaultRenderingStateWithFog = glCreateProgram();
	m_prDefaultRenderingStateWithLinearFog = glCreateProgram();
	m_prComplexSurfaceSingleTexture = glCreateProgram();
	m_prComplexSurfaceSingleTextureNoLod = glCreateProgram();
	m_prComplexSurfaceDualTextureModulated = glCreateProgram();
	m_prComplexSurfaceTripleTextureModulated = glCreateProgram();
	m_prComplexSurfaceSingleTextureWithFog = glCreateProgram();
	m_prComplexSurfaceDualTextureModulatedWithFog = glCreateProgram();
	m_prComplexSurfaceTripleTextureModulatedWithFog = glCreateProgram();
	m_prDetailTexture = glCreateProgram();
	m_prDetailTextureTwoLayer = glCreateProgram();
	m_prSingleTextureAndDetailTexture = glCreateProgram();
	m_prSingleTextureAndDetailTextureTwoLayer = glCreateProgram();
	m_prDualTextureAndDetailTexture = glCreateProgram();
	m_prDualTextureAndDetailTextureTwoLayer = glCreateProgram();

	//Mark names as allocated
	m_allocatedShaderNames = true;

	return;
}

void UOpenGLRenderDevice::FreeShaderNamesSafe(void) {
	unsigned int u;

	//Do not free names if not allocated
	if (!m_allocatedShaderNames) {
		return;
	}

	//Delete names
	glDeleteBuffers(1, &m_ubTrInfo);
	glDeleteBuffers(1, &m_csVertInfo);
	glDeleteBuffers(1, &m_ubFragInfo);

	glDeleteShader(m_vsDefaultRenderingState);
	glDeleteShader(m_vsDefaultRenderingStateWithFog);
	glDeleteShader(m_vsDefaultRenderingStateWithLinearFog);
	for (u = 0; u < 4; u++) {
		glDeleteShader(m_vsComplexSurface[u]);
	}
	for (u = 0; u < 2; u++) {
		glDeleteShader(m_vsComplexSurfaceAndDetailTexture[u]);
	}
	glDeleteShader(m_vsDetailTexture);

	glDeleteShader(m_fsDefaultRenderingState);
	glDeleteShader(m_fsDefaultRenderingStateWithFog);
	glDeleteShader(m_fsDefaultRenderingStateWithLinearFog);
	glDeleteShader(m_fsComplexSurfaceSingleTexture);
	glDeleteShader(m_fsComplexSurfaceSingleTextureNoLod);
	glDeleteShader(m_fsComplexSurfaceDualTextureModulated);
	glDeleteShader(m_fsComplexSurfaceTripleTextureModulated);
	glDeleteShader(m_fsComplexSurfaceSingleTextureWithFog);
	glDeleteShader(m_fsComplexSurfaceDualTextureModulatedWithFog);
	glDeleteShader(m_fsComplexSurfaceTripleTextureModulatedWithFog);
	glDeleteShader(m_fsDetailTexture);
	glDeleteShader(m_fsDetailTextureTwoLayer);
	glDeleteShader(m_fsSingleTextureAndDetailTexture);
	glDeleteShader(m_fsSingleTextureAndDetailTextureTwoLayer);
	glDeleteShader(m_fsDualTextureAndDetailTexture);
	glDeleteShader(m_fsDualTextureAndDetailTextureTwoLayer);

	glDeleteProgram(m_prDefaultRenderingState);
	glDeleteProgram(m_prDefaultRenderingStateWithFog);
	glDeleteProgram(m_prDefaultRenderingStateWithLinearFog);
	glDeleteProgram(m_prComplexSurfaceSingleTexture);
	glDeleteProgram(m_prComplexSurfaceSingleTextureNoLod);
	glDeleteProgram(m_prComplexSurfaceDualTextureModulated);
	glDeleteProgram(m_prComplexSurfaceTripleTextureModulated);
	glDeleteProgram(m_prComplexSurfaceSingleTextureWithFog);
	glDeleteProgram(m_prComplexSurfaceDualTextureModulatedWithFog);
	glDeleteProgram(m_prComplexSurfaceTripleTextureModulatedWithFog);
	glDeleteProgram(m_prDetailTexture);
	glDeleteProgram(m_prDetailTextureTwoLayer);
	glDeleteProgram(m_prSingleTextureAndDetailTexture);
	glDeleteProgram(m_prSingleTextureAndDetailTextureTwoLayer);
	glDeleteProgram(m_prDualTextureAndDetailTexture);
	glDeleteProgram(m_prDualTextureAndDetailTextureTwoLayer);

	//Mark names as not allocated
	m_allocatedShaderNames = false;

	return;
}

bool UOpenGLRenderDevice::InitializeShaders(void) {
	bool initOk = true;
	GLuint prId;
	GLuint ubi;


	//Vertex shaders

	//Default rendering state
	initOk &= LoadShader(m_vsDefaultRenderingState, g_vsDefaultRenderingState,
		"VS: Default rendering state");

	//Default rendering state with fog
	initOk &= LoadShader(m_vsDefaultRenderingStateWithFog, g_vsDefaultRenderingStateWithFog,
		"VS: Default rendering state with fog");

	//Default rendering state with linear fog
	initOk &= LoadShader(m_vsDefaultRenderingStateWithLinearFog, g_vsDefaultRenderingStateWithLinearFog,
		"VS: Default rendering state with linear fog");

	//Complex surface single texture
	initOk &= LoadShader(m_vsComplexSurface[0], g_vsComplexSurfaceSingleTexture,
		"VS: Complex surface single texture");

	//Complex surface dual texture
	initOk &= LoadShader(m_vsComplexSurface[1], g_vsComplexSurfaceDualTexture,
		"VS: Complex surface dual texture");

	//Complex surface triple texture
	initOk &= LoadShader(m_vsComplexSurface[2], g_vsComplexSurfaceTripleTexture,
		"VS: Complex surface triple texture");

	//Complex surface quad texture
	initOk &= LoadShader(m_vsComplexSurface[3], g_vsComplexSurfaceQuadTexture,
		"VS: Complex surface quad texture");

	//Detail texture
	initOk &= LoadShader(m_vsDetailTexture, g_vsDetailTexture,
		"VS: Detail texture");

	//Complex surface single texture and detail texture
	initOk &= LoadShader(m_vsComplexSurfaceAndDetailTexture[0], g_vsComplexSurfaceSingleTextureAndDetailTexture,
		"VS: Complex surface single texture and detail texture");

	//Complex surface dual texture and detail texture
	initOk &= LoadShader(m_vsComplexSurfaceAndDetailTexture[1], g_vsComplexSurfaceDualTextureAndDetailTexture,
		"VS: Complex surface dual texture and detail texture");


	//Fragment shaders

	//Default rendering state
	initOk &= LoadShader(m_fsDefaultRenderingState, g_fsDefaultRenderingState,
		"FS: Default rendering state");

	//Default rendering state with fog
	initOk &= LoadShader(m_fsDefaultRenderingStateWithFog, g_fsDefaultRenderingStateWithFog,
		"FS: Default rendering state with fog");

	//Default rendering state with linear fog
	initOk &= LoadShader(m_fsDefaultRenderingStateWithLinearFog, g_fsDefaultRenderingStateWithLinearFog,
		"FS: Default rendering state with linear fog");


	//Complex surface single texture
	initOk &= LoadShader(m_fsComplexSurfaceSingleTexture, g_fsComplexSurfaceSingleTexture,
		"FS: Complex surface single texture");

	//Complex surface single texture with no LOD bias
	initOk &= LoadShader(m_fsComplexSurfaceSingleTextureNoLod, g_fsComplexSurfaceSingleTextureNoLod,
		"FS: Complex surface single texture no LOD");

	//Complex surface dual texture modulated
	initOk &= LoadShader(m_fsComplexSurfaceDualTextureModulated, g_fsComplexSurfaceDualTextureModulated,
		"FS: Complex surface dual texture modulated");

	//Complex surface triple texture modulated
	initOk &= LoadShader(m_fsComplexSurfaceTripleTextureModulated, g_fsComplexSurfaceTripleTextureModulated,
		"FS: Complex surface triple texture modulated");


	//Complex surface single texture with fog
	initOk &= LoadShader(m_fsComplexSurfaceSingleTextureWithFog, g_fsComplexSurfaceSingleTextureWithFog,
		"FS: Complex surface single texture with fog");

	//Complex surface dual texture modulated with fog
	initOk &= LoadShader(m_fsComplexSurfaceDualTextureModulatedWithFog, g_fsComplexSurfaceDualTextureModulatedWithFog,
		"FS: Complex surface dual texture modulated with fog");

	//Complex surface triple texture modulated with fog
	initOk &= LoadShader(m_fsComplexSurfaceTripleTextureModulatedWithFog, g_fsComplexSurfaceTripleTextureModulatedWithFog,
		"FS: Complex surface triple texture modulated with fog");


	//Detail texture
	initOk &= LoadShader(m_fsDetailTexture, g_fsDetailTexture,
		"FS: Detail texture");

	//Detail texture two layer
	initOk &= LoadShader(m_fsDetailTextureTwoLayer, g_fsDetailTextureTwoLayer,
		"FS: Detail texture two layer");

	//Single texture and detail texture
	initOk &= LoadShader(m_fsSingleTextureAndDetailTexture, g_fsSingleTextureAndDetailTexture,
		"FS: Complex surface single texture and detail texture");

	//Single texture and detail texture two layer
	initOk &= LoadShader(m_fsSingleTextureAndDetailTextureTwoLayer, g_fsSingleTextureAndDetailTextureTwoLayer,
		"FS: Complex surface single texture and detail texture two layer");

	//Dual texture and detail texture
	initOk &= LoadShader(m_fsDualTextureAndDetailTexture, g_fsDualTextureAndDetailTexture,
		"FS: Complex surface dual texture and detail texture");

	//Dual texture and detail texture two layer
	initOk &= LoadShader(m_fsDualTextureAndDetailTextureTwoLayer, g_fsDualTextureAndDetailTextureTwoLayer,
		"FS: Complex surface dual texture and detail texture two layer");


	//Create uniform buffer objects
	glBindBuffer(GL_UNIFORM_BUFFER, m_ubTrInfo);
	glBufferData(GL_UNIFORM_BUFFER, 64, NULL, GL_DYNAMIC_DRAW);
	glBindBufferBase(GL_UNIFORM_BUFFER, 0, m_ubTrInfo);

	glBindBuffer(GL_UNIFORM_BUFFER, m_csVertInfo);
	glBufferData(GL_UNIFORM_BUFFER, 96, NULL, GL_DYNAMIC_DRAW);
	glBindBufferBase(GL_UNIFORM_BUFFER, 1, m_csVertInfo);

	glBindBuffer(GL_UNIFORM_BUFFER, m_ubFragInfo);
	glBufferData(GL_UNIFORM_BUFFER, 48, NULL, GL_DYNAMIC_DRAW);
	glBindBufferBase(GL_UNIFORM_BUFFER, 2, m_ubFragInfo);


	//Default rendering state
	prId = m_prDefaultRenderingState;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	glBindAttribLocation(prId, 4, "viTC0");
	initOk &= LinkShaderProgram(prId,
		m_vsDefaultRenderingState,
		0,
		m_fsDefaultRenderingState,
		"Default rendering state");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);

	//Default rendering state with fog
	prId = m_prDefaultRenderingStateWithFog;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	glBindAttribLocation(prId, 2, "viSecCol");
	glBindAttribLocation(prId, 4, "viTC0");
	initOk &= LinkShaderProgram(prId,
		m_vsDefaultRenderingStateWithFog,
		0,
		m_fsDefaultRenderingStateWithFog,
		"Default rendering state with fog");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);

	//Default rendering state with linear fog
	prId = m_prDefaultRenderingStateWithLinearFog;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	glBindAttribLocation(prId, 4, "viTC0");
	initOk &= LinkShaderProgram(prId,
		m_vsDefaultRenderingStateWithLinearFog,
		0,
		m_fsDefaultRenderingStateWithLinearFog,
		"Default rendering state with linear fog");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);


	//Complex surface single texture
	prId = m_prComplexSurfaceSingleTexture;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurface[0],
		0,
		m_fsComplexSurfaceSingleTexture,
		"Complex surface single texture");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);

	//Complex surface single texture with no LOD bias
	prId = m_prComplexSurfaceSingleTextureNoLod;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurface[0],
		0,
		m_fsComplexSurfaceSingleTextureNoLod,
		"Complex surface single texture no LOD");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);

	//Complex surface dual texture
	prId = m_prComplexSurfaceDualTextureModulated;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurface[1],
		0,
		m_fsComplexSurfaceDualTextureModulated,
		"Complex surface dual texture");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	glUniform1i(glGetUniformLocation(prId, "tex1"), 1);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);

	//Complex surface triple texture
	prId = m_prComplexSurfaceTripleTextureModulated;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurface[2],
		0,
		m_fsComplexSurfaceTripleTextureModulated,
		"Complex surface triple texture");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	glUniform1i(glGetUniformLocation(prId, "tex1"), 1);
	glUniform1i(glGetUniformLocation(prId, "tex2"), 2);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);


	//Complex surface single texture with fog
	prId = m_prComplexSurfaceSingleTextureWithFog;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurface[1],
		0,
		m_fsComplexSurfaceSingleTextureWithFog,
		"Complex surface single texture with fog");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	glUniform1i(glGetUniformLocation(prId, "tex1"), 1);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);

	//Complex surface dual texture with fog
	prId = m_prComplexSurfaceDualTextureModulatedWithFog;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurface[2],
		0,
		m_fsComplexSurfaceDualTextureModulatedWithFog,
		"Complex surface dual texture with fog");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	glUniform1i(glGetUniformLocation(prId, "tex1"), 1);
	glUniform1i(glGetUniformLocation(prId, "tex2"), 2);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);

	//Complex surface triple texture with fog
	prId = m_prComplexSurfaceTripleTextureModulatedWithFog;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurface[3],
		0,
		m_fsComplexSurfaceTripleTextureModulatedWithFog,
		"Complex surface triple texture with fog");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	glUniform1i(glGetUniformLocation(prId, "tex1"), 1);
	glUniform1i(glGetUniformLocation(prId, "tex2"), 2);
	glUniform1i(glGetUniformLocation(prId, "tex3"), 3);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);


	//Detail texture
	prId = m_prDetailTexture;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsDetailTexture,
		0,
		m_fsDetailTexture,
		"Detail texture");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);

	//Detail texture two layer
	prId = m_prDetailTextureTwoLayer;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsDetailTexture,
		0,
		m_fsDetailTextureTwoLayer,
		"Detail texture two layer");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);

	//Complex surface single texture and detail texture
	prId = m_prSingleTextureAndDetailTexture;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurfaceAndDetailTexture[0],
		0,
		m_fsSingleTextureAndDetailTexture,
		"Complex surface single texture and detail texture");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	glUniform1i(glGetUniformLocation(prId, "tex1"), 1);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);

	//Complex surface single texture and detail texture two layer
	prId = m_prSingleTextureAndDetailTextureTwoLayer;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurfaceAndDetailTexture[0],
		0,
		m_fsSingleTextureAndDetailTextureTwoLayer,
		"Complex surface single texture and detail texture two layer");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	glUniform1i(glGetUniformLocation(prId, "tex1"), 1);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);

	//Complex surface dual texture and detail texture
	prId = m_prDualTextureAndDetailTexture;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurfaceAndDetailTexture[1],
		0,
		m_fsDualTextureAndDetailTexture,
		"Complex surface dual texture and detail texture");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	glUniform1i(glGetUniformLocation(prId, "tex1"), 1);
	glUniform1i(glGetUniformLocation(prId, "tex2"), 2);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);

	//Complex surface dual texture and detail texture two layer
	prId = m_prDualTextureAndDetailTextureTwoLayer;
	glBindAttribLocation(prId, 0, "viPos");
	glBindAttribLocation(prId, 1, "viPriCol");
	initOk &= LinkShaderProgram(prId,
		m_vsComplexSurfaceAndDetailTexture[1],
		0,
		m_fsDualTextureAndDetailTextureTwoLayer,
		"Complex surface dual texture and detail texture two layer");
	glUseProgram(prId);
	glUniform1i(glGetUniformLocation(prId, "tex0"), 0);
	glUniform1i(glGetUniformLocation(prId, "tex1"), 1);
	glUniform1i(glGetUniformLocation(prId, "tex2"), 2);
	ubi = glGetUniformBlockIndex(prId, "trInfo");
	glUniformBlockBinding(prId, ubi, 0);
	ubi = glGetUniformBlockIndex(prId, "csVertInfo");
	glUniformBlockBinding(prId, ubi, 1);
	ubi = glGetUniformBlockIndex(prId, "fragInfo");
	glUniformBlockBinding(prId, ubi, 2);


	//Reset to default program object
	glUseProgram(0);
	m_prCurrent = 0;

	return initOk;
}

//Attempts to initializes shaders
//Safe to call multiple times as all names are always allocated
//Will reload any existing programs if called multiple times
void UOpenGLRenderDevice::TryInitializeShaders(void) {
	//Allocate shader names
	AllocateShaderNamesSafe();

	//Initialize shaders
	if (InitializeShaders() == false) {
		//Free shader names
		FreeShaderNamesSafe();

		if (DebugBit(DEBUG_BIT_BASIC)) dbgPrintf("utglr: Shader initialization failed\n");
	}

	return;
}

//Shuts down shader mode if it is active
//Freeing the shader names takes care of releasing resources
//Safe to call even if shader mode is not supported or was never initialized
void UOpenGLRenderDevice::ShutdownShaderMode(void) {
	//Free shader names
	FreeShaderNamesSafe();

	//Select invalid program object
	glUseProgram(0);
	m_prCurrent = 0;

	return;
}


void UOpenGLRenderDevice::SetProjectionStateNoCheck(bool requestNearZRangeHackProjection) {
	FLOAT zScaleVal;
	FLOAT zNear;
	FLOAT zFar;

	//Save new Z range hack projection state
	m_nearZRangeHackProjectionActive = requestNearZRangeHackProjection;

	//Set default zNear
	zNear = 0.5f;

	zScaleVal = 1.0f;
	if (requestNearZRangeHackProjection) {
#ifdef UTGLR_DEBUG_Z_RANGE_HACK_WIREFRAME
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
#endif

		zScaleVal = 0.125f;
		zNear = 4.0f;
	}
	else {
#ifdef UTGLR_DEBUG_Z_RANGE_HACK_WIREFRAME
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
#endif

		if (m_useZRangeHack) {
			zNear = 4.0f;
		}
	}

	//Set zFar
#ifdef UTGLR_UNREAL_227_BUILD
	zFar = 49152.0f;
#else
	zFar = 32768.0f;
#endif

	//Build glFrustum like matrix
	{
		float left, right, bottom, top;
		float rcpRightMinusLeft, rcpTopMinusBottom, rcpNearMinusFar;
		FLOAT projMatrix[4][4];
		FLOAT vsTransMatrix[16];

		left  = -m_RProjZ * zNear;
		right = +m_RProjZ * zNear;
		bottom = -m_Aspect*m_RProjZ * zNear;
		top    = +m_Aspect*m_RProjZ * zNear;

		rcpRightMinusLeft = 1.0f / (right - left);
		rcpTopMinusBottom = 1.0f / (top - bottom);
		rcpNearMinusFar = 1.0f / (zNear - zFar);

		projMatrix[0][0] = 2.0f * zNear * rcpRightMinusLeft;
		projMatrix[0][1] = 0.0f;
		projMatrix[0][2] = 0.0f;
		projMatrix[0][3] = 0.0f;

		projMatrix[1][0] = 0.0f;
		projMatrix[1][1] = 2.0f * zNear * rcpTopMinusBottom;
		projMatrix[1][2] = 0.0f;
		projMatrix[1][3] = 0.0f;

		projMatrix[2][0] = (right + left) * rcpRightMinusLeft;
		projMatrix[2][1] = (top + bottom) * rcpTopMinusBottom;
		projMatrix[2][2] = ((zFar + zNear) * rcpNearMinusFar) * zScaleVal * zScaleVal;
		projMatrix[2][3] = -1.0f;

		projMatrix[3][0] = 0.0f;
		projMatrix[3][1] = 0.0f;
		projMatrix[3][2] = ((2.0f * zNear * zFar) * rcpNearMinusFar) * zScaleVal * zScaleVal;
		projMatrix[3][3] = 0.0f;

		//Scale by -1y and -1z
		vsTransMatrix[0]  =  projMatrix[0][0];
		vsTransMatrix[1]  =  projMatrix[0][1];
		vsTransMatrix[2]  =  projMatrix[0][2];
		vsTransMatrix[3]  =  projMatrix[0][3];
		vsTransMatrix[4]  = -projMatrix[1][0];
		vsTransMatrix[5]  = -projMatrix[1][1];
		vsTransMatrix[6]  = -projMatrix[1][2];
		vsTransMatrix[7]  = -projMatrix[1][3];
		vsTransMatrix[8]  = -projMatrix[2][0];
		vsTransMatrix[9]  = -projMatrix[2][1];
		vsTransMatrix[10] = -projMatrix[2][2];
		vsTransMatrix[11] = -projMatrix[2][3];
		vsTransMatrix[12] =  projMatrix[3][0];
		vsTransMatrix[13] =  projMatrix[3][1];
		vsTransMatrix[14] =  projMatrix[3][2];
		vsTransMatrix[15] =  projMatrix[3][3];

		//Update shared uniform with transform matrix
		glBindBuffer(GL_UNIFORM_BUFFER, m_ubTrInfo);
		glBufferSubData(GL_UNIFORM_BUFFER, 0, 64, vsTransMatrix);
	}

	return;
}

void UOpenGLRenderDevice::SetOrthoProjection(void) {
	//Save new Z range hack projection state
	m_nearZRangeHackProjectionActive = false;

	//Build glOrtho like matrix
	{
		float left, right, bottom, top, zNear, zFar;
		float rcpRightMinusLeft, rcpTopMinusBottom, rcpNearMinusFar;
		FLOAT projMatrix[4][4];
		FLOAT vsTransMatrix[16];

		left  = -m_RProjZ * 0.5f;
		right = +m_RProjZ * 0.5f;
		bottom = -m_Aspect*m_RProjZ * 0.5f;
		top    = +m_Aspect*m_RProjZ * 0.5f;
		zNear = 1.0f * 0.5f;
		zFar  = 32768.0f;

		rcpRightMinusLeft = 1.0f / (right - left);
		rcpTopMinusBottom = 1.0f / (top - bottom);
		rcpNearMinusFar = 1.0f / (zNear - zFar);

		projMatrix[0][0] = 2.0f * rcpRightMinusLeft;
		projMatrix[0][1] = 0.0f;
		projMatrix[0][2] = 0.0f;
		projMatrix[0][3] = 0.0f;

		projMatrix[1][0] = 0.0f;
		projMatrix[1][1] = 2.0f * rcpTopMinusBottom;
		projMatrix[1][2] = 0.0f;
		projMatrix[1][3] = 0.0f;

		projMatrix[2][0] = 0.0f;
		projMatrix[2][1] = 0.0f;
		projMatrix[2][2] = 2.0f * rcpNearMinusFar;
		projMatrix[2][3] = 0.0f;

		projMatrix[3][0] = -(right + left) * rcpRightMinusLeft;
		projMatrix[3][1] = -(top + bottom) * rcpTopMinusBottom;
		projMatrix[3][2] = (zFar + zNear) * rcpNearMinusFar;
		projMatrix[3][3] = 1.0f;

		//Scale by -1y and -1z
		vsTransMatrix[0]  =  projMatrix[0][0];
		vsTransMatrix[1]  =  projMatrix[0][1];
		vsTransMatrix[2]  =  projMatrix[0][2];
		vsTransMatrix[3]  =  projMatrix[0][3];
		vsTransMatrix[4]  = -projMatrix[1][0];
		vsTransMatrix[5]  = -projMatrix[1][1];
		vsTransMatrix[6]  = -projMatrix[1][2];
		vsTransMatrix[7]  = -projMatrix[1][3];
		vsTransMatrix[8]  = -projMatrix[2][0];
		vsTransMatrix[9]  = -projMatrix[2][1];
		vsTransMatrix[10] = -projMatrix[2][2];
		vsTransMatrix[11] = -projMatrix[2][3];
		vsTransMatrix[12] =  projMatrix[3][0];
		vsTransMatrix[13] =  projMatrix[3][1];
		vsTransMatrix[14] =  projMatrix[3][2];
		vsTransMatrix[15] =  projMatrix[3][3];

		//Update shared uniform with transform matrix
		glBindBuffer(GL_UNIFORM_BUFFER, m_ubTrInfo);
		glBufferSubData(GL_UNIFORM_BUFFER, 0, 64, vsTransMatrix);
	}

	return;
}


//Must be called with (numCsTex > 0)
void UOpenGLRenderDevice::RenderComplexSurfacePassNoCheck_FP(DWORD numCsTex, DWORD BlendPolyFlags, GLuint prId) {
	guard(UOpenGLRenderDevice::RenderComplexSurfacePassNoCheck_FP);

	DWORD u;
	FLOAT vsParams[4 * 4];

	SetShaderProgram(prId);
	SetVertexArray(m_vaOneColor);

	SetBlend(BlendPolyFlags);

	u = 0;
	do {
		if (u != 0) {
			glActiveTexture(GL_TEXTURE0 + u);
		}

		SetTexture(u, *m_csTexInfo[u].Info, m_csTexInfo[u].PolyFlags, m_csTexInfo[u].PanBias);
		
		vsParams[(u * 4) + 0] = TexInfo[u].UPan;
		vsParams[(u * 4) + 1] = TexInfo[u].VPan;
		vsParams[(u * 4) + 2] = TexInfo[u].UMult;
		vsParams[(u * 4) + 3] = TexInfo[u].VMult;
	} while (++u < numCsTex);

	//Update shared uniform with complex surface texture info
	glBindBuffer(GL_UNIFORM_BUFFER, m_csVertInfo);
	glBufferSubData(GL_UNIFORM_BUFFER, 32, numCsTex * 16, vsParams);


	//Buffer complex surface vertex data

	//Make sure at least m_csPtCount entries are left in the vertex buffers
	if ((m_curVertexBufferPos + m_csPtCount) >= VERTEX_ARRAY_SIZE) {
		FlushVertexBuffers();
	}

	//Lock vertexColor buffer
	glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
	LockVertexColorBuffer();

	//Write vertex and color
	const FGLVertex *pSrcVertexArray = m_csVertexArray;
	FGLVertexColor *pVertexColorArray = m_pVertexColorArray;
	DWORD csColor = 0xFFFFFFFF;
	u = m_csPtCount;
	do {
		pVertexColorArray->x = pSrcVertexArray->x;
		pVertexColorArray->y = pSrcVertexArray->y;
		pVertexColorArray->z = pSrcVertexArray->z;
		pVertexColorArray->color = csColor;
		pSrcVertexArray++;
		pVertexColorArray++;
	} while (--u != 0);

	//Unlock vertexColor buffer
	glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
	UnlockVertexColorBuffer();


	//Draw
	for (INT PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++) {
		glDrawArrays(GL_TRIANGLE_FAN, m_curVertexBufferPos + MultiDrawFirstArray[PolyNum], MultiDrawCountArray[PolyNum]);
	}

#ifdef UTGLR_DEBUG_WORLD_WIREFRAME
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	SetBlend(PF_Modulated);

	for (INT PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++) {
		glDrawArrays(GL_TRIANGLE_FAN, m_curVertexBufferPos + MultiDrawFirstArray[PolyNum], MultiDrawCountArray[PolyNum]);
	}

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
#endif

	//Advance vertex buffer position
	m_curVertexBufferPos += m_csPtCount;


	//Always leave texture unit 0 active
	glActiveTexture(GL_TEXTURE0);

	return;

	unguard;
}

//Must be called with (numCsTex > 0)
void UOpenGLRenderDevice::RenderSingleOrDualTextureAndDetailTexture_FP(DWORD numCsTex, DWORD BlendPolyFlags, GLuint prId) {
	guard(UOpenGLRenderDevice::RenderSingleOrDualTextureAndDetailTexture_FP);

	DWORD u;
	FLOAT vsParams[4 * 4];

	SetShaderProgram(prId);
	SetVertexArray(m_vaOneColor);

	SetBlend(BlendPolyFlags);

	u = 0;
	do {
		if (u != 0) {
			glActiveTexture(GL_TEXTURE0 + u);
		}

		SetTexture(u, *m_csTexInfo[u].Info, m_csTexInfo[u].PolyFlags, m_csTexInfo[u].PanBias);
		
		vsParams[(u * 4) + 0] = TexInfo[u].UPan;
		vsParams[(u * 4) + 1] = TexInfo[u].VPan;
		vsParams[(u * 4) + 2] = TexInfo[u].UMult;
		vsParams[(u * 4) + 3] = TexInfo[u].VMult;
	} while (++u < numCsTex);

	//Update shared uniform with complex surface texture info
	glBindBuffer(GL_UNIFORM_BUFFER, m_csVertInfo);
	glBufferSubData(GL_UNIFORM_BUFFER, 32, numCsTex * 16, vsParams);


	//Buffer complex surface vertex data

	//Make sure at least m_csPtCount entries are left in the vertex buffers
	if ((m_curVertexBufferPos + m_csPtCount) >= VERTEX_ARRAY_SIZE) {
		FlushVertexBuffers();
	}

	//Lock vertexColor buffer
	glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
	LockVertexColorBuffer();

	//Write vertex and color
	const FGLVertex *pSrcVertexArray = m_csVertexArray;
	FGLVertexColor *pVertexColorArray = m_pVertexColorArray;
	DWORD detailColor = m_detailTextureColor4ub | 0xFF000000;
	u = m_csPtCount;
	do {
		pVertexColorArray->x = pSrcVertexArray->x;
		pVertexColorArray->y = pSrcVertexArray->y;
		pVertexColorArray->z = pSrcVertexArray->z;
		pVertexColorArray->color = detailColor;
		pSrcVertexArray++;
		pVertexColorArray++;
	} while (--u != 0);

	//Unlock vertexColor buffer
	glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
	UnlockVertexColorBuffer();


	//Draw
	for (INT PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++) {
		glDrawArrays(GL_TRIANGLE_FAN, m_curVertexBufferPos + MultiDrawFirstArray[PolyNum], MultiDrawCountArray[PolyNum]);
	}

#ifdef UTGLR_DEBUG_WORLD_WIREFRAME
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	SetBlend(PF_Modulated);

	for (INT PolyNum = 0; PolyNum < m_csPolyCount; PolyNum++) {
		glDrawArrays(GL_TRIANGLE_FAN, m_curVertexBufferPos + MultiDrawFirstArray[PolyNum], MultiDrawCountArray[PolyNum]);
	}

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
#endif

	//Advance vertex buffer position
	m_curVertexBufferPos += m_csPtCount;


	//Always leave texture unit 0 active
	glActiveTexture(GL_TEXTURE0);

	return;

	unguard;
}

//Modified this routine to always set up detail texture state
//It should only be called if at least one polygon will be detail textured
void UOpenGLRenderDevice::DrawDetailTexture_FP(FTextureInfo &DetailTextureInfo) {
	INT Index = 0;
	GLuint prId = 0;

	//Setup detail texture state
	SetBlend(PF_Modulated);

	prId = m_prDetailTexture;
	if (DetailMax >= 2) prId = m_prDetailTextureTwoLayer;
	SetShaderProgram(prId);
	SetVertexArray(m_vaOneColor);

	glActiveTexture(GL_TEXTURE0);
	SetTextureNoPanBias(0, DetailTextureInfo, PF_Modulated);

	//Update shared uniform with complex surface texture info
	FLOAT vsParams[4] = { TexInfo[0].UPan, TexInfo[0].VPan, TexInfo[0].UMult, TexInfo[0].VMult };
	glBindBuffer(GL_UNIFORM_BUFFER, m_csVertInfo);
	glBufferSubData(GL_UNIFORM_BUFFER, 32, 16, vsParams);


	DWORD detailColor = m_detailTextureColor4ub | 0xFF000000;
	INT *pNumPts = &MultiDrawCountArray[0];
	DWORD *pDetailTextureIsNear = DetailTextureIsNearArray;
	DWORD csPolyCount = m_csPolyCount;
	for (DWORD PolyNum = 0; PolyNum < csPolyCount; PolyNum++, pNumPts++, pDetailTextureIsNear++) {
		DWORD NumPts = *pNumPts;
		DWORD isNearBits = *pDetailTextureIsNear;
		INT i;

		//Skip the polygon if it will not be detail textured
		if (isNearBits == 0) {
			Index += NumPts;
			continue;
		}

		//Make sure at least NumPts entries are left in the vertex buffers
		if ((m_curVertexBufferPos + NumPts) >= VERTEX_ARRAY_SIZE) {
			FlushVertexBuffers();
		}

		//Lock vertexColor buffer
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		LockVertexColorBuffer();

		//Write vertex
		const FGLVertex *pSrcVertexArray = &m_csVertexArray[Index];
		FGLVertexColor *pVertexColorArray = m_pVertexColorArray;
		for (i = 0; i < NumPts; i++) {
			pVertexColorArray[i].x = pSrcVertexArray[i].x;
			pVertexColorArray[i].y = pSrcVertexArray[i].y;
			pVertexColorArray[i].z = pSrcVertexArray[i].z;
			pVertexColorArray[i].color = detailColor;
		}

		//Unlock vertexColor buffers
		glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
		UnlockVertexColorBuffer();

		//Draw the triangles
		glDrawArrays(GL_TRIANGLE_FAN, m_curVertexBufferPos, NumPts);

		//Advance vertex buffer position
		m_curVertexBufferPos += NumPts;

		Index += NumPts;
	}


	//Clear detail texture state
	glActiveTexture(GL_TEXTURE0);

	return;
}

INT UOpenGLRenderDevice::BufferStaticComplexSurfaceGeometry_VP(const FSurfaceFacet& Facet) {
	INT numVerts = 0;

	//Buffer static geometry
	m_csPolyCount = 0;
	FGLVertex *pVertex = m_csVertexArray;
	for (FSavedPoly* Poly = Facet.Polys; Poly; Poly = Poly->Next) {
		//Skip if no points
		INT NumPts = Poly->NumPts;
		if (NumPts <= 0) {
			continue;
		}

		DWORD csPolyCount = m_csPolyCount;
		MultiDrawFirstArray[csPolyCount] = numVerts;
		MultiDrawCountArray[csPolyCount] = NumPts;
		m_csPolyCount = csPolyCount + 1;

		numVerts += NumPts;
		if (numVerts > VERTEX_ARRAY_SIZE) {
			return 0;
		}
		FTransform **pPts = &Poly->Pts[0];
		do {
			const FVector &Point = (*pPts++)->Point;

			pVertex->x = Point.X;
			pVertex->y = Point.Y;
			pVertex->z = Point.Z;
			pVertex++;
		} while (--NumPts != 0);
	}

	return numVerts;
}

DWORD UOpenGLRenderDevice::BufferDetailTextureData(FLOAT NearZ) {
	DWORD *pDetailTextureIsNear = DetailTextureIsNearArray;
	DWORD anyIsNearBits = 0;

	FGLVertex *pVertex = m_csVertexArray;
	INT *pNumPts = &MultiDrawCountArray[0];
	DWORD csPolyCount = m_csPolyCount;
	do {
		INT NumPts = *pNumPts++;
		DWORD isNear = 0;

		do {
			isNear <<= 1;
			if (pVertex->z < NearZ) {
				isNear |= 1;
			}
			pVertex++;
		} while (--NumPts != 0);

		*pDetailTextureIsNear++ = isNear;
		anyIsNearBits |= isNear;
	} while (--csPolyCount != 0);

	return anyIsNearBits;
}

#ifdef UTGLR_INCLUDE_SSE_CODE
__declspec(naked) DWORD UOpenGLRenderDevice::BufferDetailTextureData_SSE2(FLOAT NearZ) {
	__asm {
		movd xmm0, [esp+4]

		push esi
		push edi

		lea esi, [ecx]this.m_csVertexArray
		lea edx, [ecx]this.MultiDrawCountArray
		lea edi, [ecx]this.DetailTextureIsNearArray

		pxor xmm1, xmm1

		mov ecx, [ecx]this.m_csPolyCount

		poly_count_loop:
			mov eax, [edx]
			add edx, 4

			pxor xmm2, xmm2

			num_pts_loop:
				movss xmm3, [esi+8]
				add esi, TYPE FGLVertex

				pslld xmm2, 1

				cmpltss xmm3, xmm0
				psrld xmm3, 31

				por xmm2, xmm3

				dec eax
				jne num_pts_loop

			movd [edi], xmm2
			add edi, 4

			por xmm1, xmm2

			dec ecx
			jne poly_count_loop

		movd eax, xmm1

		pop edi
		pop esi

		ret 4
	}
}
#endif //UTGLR_INCLUDE_SSE_CODE

void UOpenGLRenderDevice::EndBufferingNoCheck(void) {
	switch (m_bufferedVertsType) {
	case BV_TYPE_GOURAUD_POLYS:
		EndGouraudPolygonBufferingNoCheck();
		break;

	case BV_TYPE_TILES:
		EndTileBufferingNoCheck();
		break;

	case BV_TYPE_LINES:
		EndLineBufferingNoCheck();
		break;

	case BV_TYPE_POINTS:
		EndPointBufferingNoCheck();
		break;

	default:
		;
	}

	m_bufferedVerts = 0;

	return;
}

void UOpenGLRenderDevice::EndGouraudPolygonBufferingNoCheck(void) {
	SetDefaultAAState();
	//EndGouraudPolygonBufferingNoCheck sets its own projection state
	//Vertex array state set when start buffering
	//Shader program state set when start buffering

	clock(GouraudCycles);

	//Set projection state
	SetProjectionState(m_requestNearZRangeHackProjection);

#ifdef UTGLR_DEBUG_ACTOR_WIREFRAME
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
#endif

	//Unlock vertexColor and texCoord0 buffers
	//Unlock secondary color buffer if fog
	glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
	UnlockVertexColorBuffer();
	if (m_requestedColorFlags & CF_FOG_MODE) {
		glBindBuffer(GL_ARRAY_BUFFER, m_vbSecondaryColor);
		UnlockSecondaryColorBuffer();
	}
	glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
	UnlockTexCoordBuffer(0);

	//Draw the triangles
	glDrawArrays(GL_TRIANGLES, m_curVertexBufferPos, m_bufferedVerts);

	//Advance vertex buffer position
	m_curVertexBufferPos += m_bufferedVerts;

#ifdef UTGLR_DEBUG_ACTOR_WIREFRAME
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
#endif

	unclock(GouraudCycles);
}

void UOpenGLRenderDevice::EndTileBufferingNoCheck(void) {
	if (NoAATiles) {
		SetDisabledAAState();
	}
	else {
		SetDefaultAAState();
	}
	SetDefaultProjectionState();
	//Vertex array state set when start buffering
	//Shader program state set when start buffering

	clock(TileCycles);

	//Unlock vertexColor and texCoord0 buffers
	glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
	UnlockVertexColorBuffer();
	glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
	UnlockTexCoordBuffer(0);

	//Draw the quads (stored as triangles)
	glDrawArrays(GL_TRIANGLES, m_curVertexBufferPos, m_bufferedVerts);

	//Advance vertex buffer position
	m_curVertexBufferPos += m_bufferedVerts;

	unclock(TileCycles);
}

void UOpenGLRenderDevice::EndLineBufferingNoCheck(void) {
	//AA state set when start buffering
	//Projection state set when start buffering
	//Vertex array state set when start buffering
	//Shader program state set when start buffering

	//Unlock vertexColor and texCoord0 buffers
	glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
	UnlockVertexColorBuffer();
	glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
	UnlockTexCoordBuffer(0);

	//Draw the lines
	glDrawArrays(GL_LINES, m_curVertexBufferPos, m_bufferedVerts);

	//Advance vertex buffer position
	m_curVertexBufferPos += m_bufferedVerts;
}

void UOpenGLRenderDevice::EndPointBufferingNoCheck(void) {
	//AA state set when start buffering
	//Projection state set when start buffering
	//Vertex array state set when start buffering
	//Shader program state set when start buffering

	//Unlock vertexColor and texCoord0 buffers
	glBindBuffer(GL_ARRAY_BUFFER, m_vbVertexColor);
	UnlockVertexColorBuffer();
	glBindBuffer(GL_ARRAY_BUFFER, m_vbTexCoord[0]);
	UnlockTexCoordBuffer(0);

	//Draw the points (stored as triangles)
	glDrawArrays(GL_TRIANGLES, m_curVertexBufferPos, m_bufferedVerts);

	//Advance vertex buffer position
	m_curVertexBufferPos += m_bufferedVerts;
}


// Static variables.
INT UOpenGLRenderDevice::NumDevices = 0;
INT UOpenGLRenderDevice::LockCount  = 0;
#ifdef _WIN32
HGLRC   UOpenGLRenderDevice::hCurrentRC    = NULL;
HMODULE UOpenGLRenderDevice::hModuleGlMain = NULL;
TArray<HGLRC> UOpenGLRenderDevice::AllContexts;
#endif

bool UOpenGLRenderDevice::g_gammaFirstTime = false;
bool UOpenGLRenderDevice::g_haveOriginalGammaRamp = false;
#ifdef _WIN32
UOpenGLRenderDevice::FGammaRamp UOpenGLRenderDevice::g_originalGammaRamp;
#endif

UOpenGLRenderDevice::DWORD_CTTree_t UOpenGLRenderDevice::m_sharedZeroPrefixBindTrees[NUM_CTTree_TREES];
UOpenGLRenderDevice::QWORD_CTTree_t UOpenGLRenderDevice::m_sharedNonZeroPrefixBindTrees[NUM_CTTree_TREES];
CCachedTextureChain UOpenGLRenderDevice::m_sharedNonZeroPrefixBindChain;
UOpenGLRenderDevice::QWORD_CTTree_NodePool_t UOpenGLRenderDevice::m_sharedNonZeroPrefixTexIdPool;
UOpenGLRenderDevice::TexPoolMap_t UOpenGLRenderDevice::m_sharedRGBA8TexPool;

//OpenGL 1.x function pointers for remaining subset to be used with OpenGL 3.3
#define GL1_PROC(ret, func, params) ret (STDCALL *UOpenGLRenderDevice::func)params;
#include "OpenGL1Funcs.h"
#undef GL1_PROC

//OpenGL 3.3 function pointers
#define GL3_PROC(ret, func, params) ret (STDCALL *UOpenGLRenderDevice::func)params;
#include "OpenGL3Funcs.h"
#undef GL3_PROC

//OpenGL extension function pointers
#define GL_EXT_NAME(name) bool UOpenGLRenderDevice::SUPPORTS##name = 0;
#define GL_EXT_PROC(ext, ret, func, params) ret (STDCALL *UOpenGLRenderDevice::func)params;
#include "OpenGLExtFuncs.h"
#undef GL_EXT_NAME
#undef GL_EXT_PROC

/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
